/*
 *  $Id: netlogreader.h,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: NetLogger File Reader
 *
 */


#ifndef NETLOGREADER_H
#define NETLOGREADER_H


#include "tdsystem.h"
#include "timedthread.h"
#include "networkmonitorinterface.h"
#include "streammonitorreport.h"


namespace Coral {


/**
 * This class implements a reader for NetworkMonitorReports. The reports will
 * be read from a file written by NetLogWriter.
 *
 * @short   Network Log Reader
 * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
 * @version 1.0
 */       
class NetLogReader
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor for a new NetLogReader.
     *
     * @param name Name of log file.
     */
   NetLogReader(const char* name);

   /**
     * Destructor.
     */
   ~NetLogReader();


   // ====== Status functions ===============================================
   /**
     * Check, if NetLogReader is ready.
     *
     * @return true, if ready; false, if not.
     */  
   inline bool ready() const;

   /**
     * Check, if NetLogReader has reached end of file.
     *
     * @return true, if end of file is reached; false, if not.
     */  
   inline bool eof() const;


   /**
     * Get time stamp of file (seconds since 01-Jan-1970).
     *
     * @return Time stamp.
     */
   inline card64 getTimeStamp(); 

   /**
     * Get interval of the entries in file.
     *
     * @return Interval in microseconds.
     */
   inline card64 getInterval();
 

   // ====== Report functions ===============================================
   /**
     * Read next report from file.
     *
     * @param report Address of NetworkMonitorReport to store the report.
     * @return true, if there was a report; false otherwise.
     */
   bool readNextReport(NetworkMonitorReport* report);

   /**
     * Read next stream report from file.
     *
     * @param report Address of StreamReport to store the report.
     * @return true, if there was a stream report; false otherwise.
     */
   bool readNextStreamReport(StreamReport* report);

   /**
     * Reset report file position to first report.
     */
   inline void reset();


   // ====== Private data ===================================================
   private:
   FILE*  FileDescriptor;
   card64 TimeStamp;
   card64 Interval;
   card32 Flags;
   card32 StreamReportCount;
};


}


#include "netlogreader.icc"


#endif
