/*
 *  $Id: netlogwriter.cc,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: NetLogger File Writer
 *
 */


#include "tdsystem.h"
#include "netlogwriter.h"
#include "timedthread.h"
#include "networkmonitorinterface.h"
#include "streammonitorreport.h"
#include "netlogfileheader.h"
#include "tools.h"


namespace Coral {


// ###### Constructor #######################################################
NetLogWriter::NetLogWriter(const char*              name,
                           const card64             interval,
                           NetworkMonitorInterface* monitor,
                           StreamMonitorReport*     smReport)
   : TimedThread(interval)
{
   // ====== Initialization =================================================
   Monitor        = monitor;
   SMReport       = smReport;
   Count          = 0;
   FileDescriptor = fopen(name,"w");
   if(FileDescriptor == NULL) {
      cerr << "ERROR: NetLogWriter::NetLogWriter() - Unable to write log file <" << name << ">!" << endl;
      return;
   }

   // ====== Write header into file =========================================
   TimeStamp     = getMicroTime();
   Interval      = interval;

   cardinal flags = NLFH_Default;
   if(SMReport != NULL) {
      flags |= NLFH_ContainsStreams;
   }

   NetLogFileHeader nlfh;
   strncpy((char*)&nlfh.ID,NetLogFileHeaderID,4);
   nlfh.Version   = translate16(NetLogFileHeaderVersion);
   nlfh.EntrySize = translate16(sizeof(NetworkMonitorReport));
   nlfh.TimeStamp = translate64(TimeStamp);
   nlfh.Interval  = translate64(Interval);
   nlfh.Flags     = translate32(flags);
   if(fwrite((void *)&nlfh,sizeof(nlfh),1,FileDescriptor) != 1) {
      cerr << "ERROR: NetLogWriter::NetLogWriter() - Write error!" << endl;
      fclose(FileDescriptor);
      FileDescriptor = NULL;
      return;
   }
}


// ###### Destructor ########################################################
NetLogWriter::~NetLogWriter()
{
   stop();
   if(FileDescriptor) {
      fclose(FileDescriptor);
      FileDescriptor = NULL;
   }
}


// ###### TimedThread's timerEvent() implementation #########################
void NetLogWriter::timerEvent()
{
   if(FileDescriptor != NULL) {
      synchronized();

      // Get report from NetworkMonitor
      NetworkMonitorReport report;
      Monitor->lock();
      Monitor->moveReport(&report);

      // Translate report and write it to file
      report.translate();
      if(fwrite((void*)&report,sizeof(report),1,FileDescriptor) != 1) {
         Monitor->unlock();
         cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
         ::exit(1);
      }

      // card64 t1 = getMicroTime();

      if(SMReport != NULL) {
         const cardinal count = SMReport->getCount();
         card32 data = translate32(count);
         if(fwrite((void*)&data,sizeof(card32),1,FileDescriptor) != 1) {
            Monitor->unlock();
            cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
            ::exit(1);
         }
         for(cardinal i = 0;i < count;i++) {
            StreamReport* sr = SMReport->getReport(i);
            sr->translate();
            if(fwrite((void*)sr,sizeof(StreamReport),1,FileDescriptor) != 1) {
               Monitor->unlock();
               cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
               ::exit(1);
            }
         }
         SMReport->reset();
      }

      // card64 t2 = getMicroTime();
      // cerr << (t2 - t1) << endl;

      Monitor->unlock();
      Count++;
      unsynchronized();
   }
}


}
