/*
 *  $Id: networkmonitor.cc,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Network Monitor
 *
 */


#include "tdsystem.h"
#include "networkmonitor.h"
#include "streammonitorreport.h"
#include "thread.h"
#include "tdsocket.h"
#include "tools.h"

#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "tdin6.h"

extern "C" {
#include <pcap.h>
}


// Debug mode: Print all UDP and TCP connections
// #define DEBUG
// Print unknown protocol's number
// #define PRINT_UNKNOWN


namespace Coral {


#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86dd
#endif


// ###### Constructor #######################################################
NetworkMonitor::NetworkMonitor(StreamMonitorReport* smr = NULL, char* device = NULL)
   : Thread("NetworkMonitor")
{
   resetReport();
   
   PCapDescriptor    = NULL;
   PCapInterfaceType = (card32)-1;
   PCapHandler       = NULL;
   PCapDevice        = NULL;

   // Check user ID
   uid_t uid = geteuid();
   if(uid != 0) {
      cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Only root is allowed to capture network devices!" << endl;
      return;
   }

   // Get default network device name, if no device name is given.
   char ebuf[PCAP_ERRBUF_SIZE];
   if(device == NULL) {
      device = pcap_lookupdev(ebuf);
      if(device == NULL) {
         cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Unable to get default device name:" << endl
              << "       " << ebuf << "!" << endl;
         return;
      }
   }

   // Open device in promiscuous mode, save DefaultSnaplen bytes of each
   // packet, save error string into ebuf.
   PCapDescriptor = pcap_open_live(device,DefaultSnaplen,1,0,ebuf);
   if(PCapDescriptor == NULL) {
      cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Unable to open network device " << device << ":" << endl
           << "       " << ebuf << "!" << endl;
      return;
   }

   // Select handler for the interface type of the opened device.
   PCapInterfaceType = pcap_datalink(PCapDescriptor);
   switch(PCapInterfaceType) {
      case DLT_NULL:
         PCapHandler = &loopbackInterfaceHandler;
       break;
      case DLT_EN10MB:
         PCapHandler = &ethernetInterfaceHandler;
       break;
      case DLT_PPP:
         PCapHandler = &pppInterfaceHandler;
       break;
      case DLT_RAW:
         PCapHandler = &rawInterfaceHandler;
       break;
      default:
         cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Interface type "
              << PCapInterfaceType << " not implemented!" << endl;
         pcap_close(PCapDescriptor);
         PCapDescriptor = NULL;
       break;
   }

   // Set device name
   PCapDevice = device;
   SMReport   = smr;
}


// ###### Destructor ########################################################
NetworkMonitor::~NetworkMonitor()
{
   stop();
   if(PCapDescriptor != NULL) {
      pcap_close(PCapDescriptor);
      PCapDescriptor = NULL;
   }
}


// ###### Check, if NetworkMonitor is ready #################################
bool NetworkMonitor::ready() const
{
   return(PCapDescriptor != NULL);
}


// ###### Lock monitor to access data #######################################
void NetworkMonitor::lock()
{
   synchronized();
}


// ###### Unlock monitor ####################################################
void NetworkMonitor::unlock()
{
   unsynchronized();
}


// ###### Reset report ######################################################
void NetworkMonitor::resetReport()
{
   synchronized();
   Report.reset();
   TimeStamp = getMicroTime();
   unsynchronized();
}


// ###### Get report ########################################################
void NetworkMonitor::getReport(NetworkMonitorReport* report)
{
   synchronized();
   Report.Interval  = getMicroTime() - TimeStamp;
   Report.TimeStamp = TimeStamp;
   memcpy(report,&Report,sizeof(Report));
   unsynchronized();
}


// ###### Get report and reset ##############################################
void NetworkMonitor::moveReport(NetworkMonitorReport* report)
{
   synchronized();
   getReport(report);
   resetReport();
   unsynchronized();
}


// ###### Thread's run() implementation #####################################
void NetworkMonitor::run()
{
   if(PCapDescriptor == NULL) {
      cerr << "ERROR: NetworkMonitor::run() - Start without opened device!"
           << endl;
      ::exit(1);
   }

   // Handle packet headers on opened device
   resetReport();
   if(pcap_loop(PCapDescriptor,-1,PCapHandler,(u_char*)this) < 0) {
      cerr << "ERROR: NetworkMonitor::run() - pcap_loop failed!" << endl;
      ::exit(1);
   }
}


// ###### Handler for SCTP packets ##########################################
void NetworkMonitor::handleSCTP(const card8    protocol,
                                StreamSrcDest* stream,
                                const sctphdr* sctp,
                                const cardinal sctpLength,
                                const cardinal rawLength)
{
   Report.BytesSCTP += sctpLength;
   Report.PacketsSCTP++;
   stream->Source.Port      = sctp->src_port;
   stream->Destination.Port = sctp->dest_port;

   card64 totalPayload = 0;
   sctpchunkhdr* chunk = (sctpchunkhdr*)((long)sctp + sizeof(sctphdr));
   while((long)chunk < (long)sctp + (long)sctpLength) {
      if(chunk->chunk_id == 0) {
         const sctpdatachunkhdr* datachunk = (sctpdatachunkhdr*)chunk;
         const card16 streamID = translate16(datachunk->stream_id);
         cardinal sctpBytes = (cardinal)translate16(chunk->chunk_length) -
                                 sizeof(sctpdatachunkhdr);
         Report.BytesSCTPData += sctpBytes;
         totalPayload         += sctpBytes;

#ifdef DEBUG
         char str[64];
         snprintf((char*)&str,sizeof(str),
                  "$%02x/$%05x",stream->TrafficClass,stream->FlowLabel);
         cout << "SCTP " << stream->Source << " -> " << stream->Destination << "  " << str << endl;
#endif

         if(SMReport != NULL) {
            SMReport->update(protocol,IPPROTO_SCTP,
                             stream->Source,stream->Destination,stream->FlowLabel,
                             stream->TrafficClass,streamID,
                             sctpBytes + sizeof(sctpdatachunkhdr),sctpBytes);
         }
      }

      if(chunk->chunk_length < sizeof(sctpchunkhdr)) {
         break;
      }
      chunk = (sctpchunkhdr*)((long)chunk + (long)chunk->chunk_length);
   }

   if(SMReport != NULL) {
      SMReport->update(protocol,IPPROTO_SCTP,
                       stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,(card32)-1,
                       sctpLength,totalPayload);
   }
}


// ###### Handler for TCP packets ###########################################
void NetworkMonitor::handleTCP(const card8    protocol,
                               StreamSrcDest* stream,
                               const tcphdr*  tcp,
                               const cardinal tcpLength,
                               const cardinal rawLength)
{
   Report.BytesTCP     += tcpLength;
   Report.PacketsTCP++;
#if (SYSTEM == OS_FreeBSD)
   const cardinal tcpBytes = (tcpLength - (4 * tcp->th_off));
   Report.BytesTCPData += tcpBytes;
   stream->Source.Port      = tcp->th_sport;
   stream->Destination.Port = tcp->th_dport;
#else
   const cardinal tcpBytes = (tcpLength - (4 * tcp->doff));
   Report.BytesTCPData += tcpBytes;
   stream->Source.Port      = tcp->source;
   stream->Destination.Port = tcp->dest;
#endif


#ifdef DEBUG
   char str[64];
   snprintf((char*)&str,sizeof(str),,
            "$%02x/$%05x",stream->TrafficClass,stream->FlowLabel);
   cout << "TCP " << stream->Source << " -> " << stream->Destination << "  " << str << endl;
#endif

   if(SMReport != NULL) {
      SMReport->update(protocol,IPPROTO_TCP,
                       stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,0x0000,rawLength,tcpBytes);
   }
}


// ###### Handler for UDP packets ###########################################
void NetworkMonitor::handleUDP(const card8    protocol,
                               StreamSrcDest* stream,
                               const udphdr*  udp,
                               const cardinal udpLength,
                               const cardinal rawLength)
{
   Report.BytesUDP     += udpLength;
   Report.BytesUDPData += (udpLength - sizeof(udphdr));
   Report.PacketsUDP++;

#if (SYSTEM == OS_FreeBSD)
   stream->Source.Port      = udp->uh_sport;
   stream->Destination.Port = udp->uh_dport;
#else
   stream->Source.Port      = udp->source;
   stream->Destination.Port = udp->dest;
#endif


#ifdef DEBUG
   char str[64];
   snprintf((char*)&str,sizeof(str),
           "$%02x/$%05x",stream->TrafficClass,stream->FlowLabel);
   cout << "UDP " << stream->Source << " -> " << stream->Destination << "  " << str << endl;
#endif

   if(SMReport != NULL) {
      SMReport->update(protocol,IPPROTO_UDP,
                       stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,0x0000,rawLength,udpLength - sizeof(udphdr));
   }
}


// ###### Handler for ICMP4 packets ##########################################
void NetworkMonitor::handleICMP4(const card8    protocol,
                                 StreamSrcDest* stream,
                                 const icmp*    icmp,
                                 const cardinal icmpLength,
                                 const cardinal rawLength)
{
   Report.BytesICMP4 += icmpLength;
   Report.PacketsICMP4++;
   if(SMReport != NULL) {
      SMReport->update(protocol,IPPROTO_ICMP,
                       stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,0x0000,rawLength,icmpLength);
   }
}


// ###### Handler for ICMP6 packets ##########################################
void NetworkMonitor::handleICMP6(const card8      protocol,
                                 StreamSrcDest*   stream,
                                 const icmp6_hdr* icmp,
                                 const cardinal   icmpLength,
                                 const cardinal   rawLength)
{
   Report.BytesICMP6 += icmpLength;
   Report.PacketsICMP6++;
   if(SMReport != NULL) {
      SMReport->update(protocol,IPPROTO_ICMPV6,
                       stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,0x0000,rawLength,icmpLength);
   }
}


// ###### Handler for IPv4 packets ##########################################
void NetworkMonitor::ip4PacketHandler(const ip* ip, const cardinal length)
{
   // ====== Update byte and packet counter =================================
   Report.BytesIPv4 += (card64)length;
   Report.PacketsIPv4++;

   // ====== Set from and to addresses ======================================
   StreamSrcDest stream;
   stream.Source.Port      = 0; // In IP we do not know about ports!
   stream.Destination.Port = 0;
   for(cardinal i = 0;i < 5;i++) {
      stream.Source.Host[i]      = 0;
      stream.Destination.Host[i] = 0;
   }
   stream.Source.Host[5]      = 0xffff;
   stream.Destination.Host[5] = 0xffff;
   memcpy((void*)&stream.Source.Host[6],(void*)&ip->ip_src,4);
   memcpy((void*)&stream.Destination.Host[6],(void*)&ip->ip_dst,4);
   stream.FlowLabel    = 0;
   stream.TrafficClass = (ip->ip_tos & 0x3f);

   // ====== Update this packet's QoS group's byte and packet counter =======
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(stream.TrafficClass);
   Report.ClassBytes[tosIndex] += (card64)length;
   Report.ClassPackets[tosIndex]++;

   // ====== Handle protocols ===============================================
   const card8 protocol = ip->ip_p;
   switch(protocol) {
      case IPPROTO_UDP:
        {
         const udphdr* udp = (udphdr*)((char*)ip + sizeof(struct ip));
         const cardinal udplength = length - sizeof(struct ip);
         handleUDP(4,&stream,udp,udplength,length);
        }
       break;
      case IPPROTO_TCP:
        {
         const tcphdr* tcp = (tcphdr*)((char*)ip + sizeof(struct ip));
         const cardinal tcplength = length - sizeof(struct ip);
         handleTCP(4,&stream,tcp,tcplength,length);
        }
       break;
      case IPPROTO_SCTP:
        {
         const sctphdr* sctp = (sctphdr*)((char*)ip + sizeof(struct ip));
         const cardinal sctplength = length - sizeof(struct ip);
         handleSCTP(4,&stream,sctp,sctplength,length);
        }
       break;
      case IPPROTO_ICMP:
        const icmp*    icmppkt    = (icmp*)((char*)ip + sizeof(struct ip));
        const cardinal icmplength = length - sizeof(struct ip);
        handleICMP4(4,&stream,icmppkt,icmplength,length);
       break;
   }
}


// ###### Handler for IPv6 packets ##########################################
void NetworkMonitor::ip6PacketHandler(const ip6_hdr* ip, const cardinal length)
{
   // ====== Update byte and packet counter =================================
   Report.BytesIPv6 += (card64)length;
   Report.PacketsIPv6++;

   // ====== Set from and to addresses ======================================
   StreamSrcDest stream;
   stream.Source.Port      = 0; // In IP we do not know about ports!
   stream.Destination.Port = 0;
   memcpy((void*)&stream.Source.Host,&ip->ip6_src,16);
   memcpy((void*)&stream.Destination.Host,&ip->ip6_dst,16);
   stream.FlowLabel    = ntohl(ip->ip6_flow) & IPV6_FLOWINFO_FLOWLABEL;
   stream.TrafficClass = ((ntohl(ip->ip6_flow) & IPV6_FLOWINFO_PRIORITY) >> 20) & 0x3f;

   // ====== Update this packet's QoS group's byte and packet counter =======
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(stream.TrafficClass);
   Report.ClassBytes[tosIndex] += (card64)length;
   Report.ClassPackets[tosIndex]++;

   // ====== Handle protocols ===============================================
   const card8 protocol = ip->ip6_nxt;
   switch(protocol) {
      case IPPROTO_UDP:
        {
         const udphdr* udp = (udphdr*)((char*)ip + sizeof(ip6_hdr));
         const cardinal udplength = length - sizeof(ip6_hdr);
         handleUDP(6,&stream,udp,udplength,length);
        }
       break;
      case IPPROTO_TCP:
        {
         const tcphdr* tcp = (tcphdr*)((char*)ip + sizeof(ip6_hdr));
         const cardinal tcplength = length - sizeof(ip6_hdr);
         handleTCP(6,&stream,tcp,tcplength,length);
        }
       break;
      case IPPROTO_SCTP:
        {
         const sctphdr* sctp = (sctphdr*)((char*)ip + sizeof(struct ip6_hdr));
         const cardinal sctplength = length - sizeof(struct ip6_hdr);
         handleSCTP(6,&stream,sctp,sctplength,length);
        }
       break;
      case IPPROTO_ICMPV6:
        const icmp6_hdr* icmp = (icmp6_hdr*)((char*)ip + sizeof(ip6_hdr));
        const cardinal icmplength = length - sizeof(ip6_hdr);
        handleICMP6(6,&stream,icmp,icmplength,length);
       break;
   }
}


// ###### Handler for loopback frames #######################################
#define NULL_HDRLEN 4
void NetworkMonitor::loopbackInterfaceHandler(u_char*            userData,
                                              const pcap_pkthdr* pcapheader,
                                              const u_char*      packet)
{
   NetworkMonitor* monitor = (NetworkMonitor*)userData;
   const char*  payload    = (char*)packet + 4;
   const cardinal length     = pcapheader->len - NULL_HDRLEN;
   cardinal protocol;
   memcpy((char*)&protocol,packet,sizeof(protocol));
   protocol = ntohl(protocol) & 0xffff;

   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;

   switch(protocol) {
      case ETHERTYPE_IP:
         monitor->ip4PacketHandler((const ip*)payload,length);
        break;
      case ETHERTYPE_IPV6:
         monitor->ip6PacketHandler((const ip6_hdr*)payload,length);
        break;
      default:
#ifdef PRINT_UNKNOWN
         printf("Unknown protocol %04x\n",protocol);
#endif
        break;
   }

   monitor->unsynchronized();
}


// ###### Handler for ethernet frames #######################################
void NetworkMonitor::ethernetInterfaceHandler(u_char*            userData,
                                              const pcap_pkthdr* pcapheader,
                                              const u_char*      packet)
{
   NetworkMonitor*     monitor  = (NetworkMonitor*)userData;
   const char*         payload  = (char*)packet + sizeof(struct ether_header);
   const cardinal      length   = pcapheader->len - sizeof(struct ether_header);
   const ether_header* ethernet = (struct ether_header*)packet;
   const card16        protocol = ntohs(ethernet->ether_type);

   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;

   switch(protocol) {
      case ETHERTYPE_IP:
         monitor->ip4PacketHandler((const ip*)payload,length);
        break;
      case ETHERTYPE_IPV6:
         monitor->ip6PacketHandler((const ip6_hdr*)payload,length);
        break;
      case ETHERTYPE_ARP:
       break;
      default:
#ifdef PRINT_UNKNOWN
         printf("Unknown protocol %04x\n",protocol);
#endif
        break;
   }

   monitor->unsynchronized();
}


#define PPP_HDRLEN	4	/* length of PPP header */

#define PPP_ADDRESS	0xff	/* The address byte value */
#define PPP_CONTROL	0x03	/* The control byte value */

/* Protocol numbers */
#define PPP_IP		0x0021	/* Raw IP */
#define PPP_OSI		0x0023	/* OSI Network Layer */
#define PPP_NS		0x0025	/* Xerox NS IDP */
#define PPP_DECNET	0x0027	/* DECnet Phase IV */
#define PPP_APPLE	0x0029	/* Appletalk */
#define PPP_IPX		0x002b	/* Novell IPX */
#define PPP_VJC		0x002d	/* Van Jacobson Compressed TCP/IP */
#define PPP_VJNC	0x002f	/* Van Jacobson Uncompressed TCP/IP */
#define PPP_BRPDU	0x0031	/* Bridging PDU */
#define PPP_STII	0x0033	/* Stream Protocol (ST-II) */
#define PPP_VINES	0x0035	/* Banyan Vines */
#define PPP_IPV6	0x0057	/* IPv6 */

#define PPP_HELLO	0x0201	/* 802.1d Hello Packets */
#define PPP_LUXCOM	0x0231	/* Luxcom */
#define PPP_SNS		0x0233	/* Sigma Network Systems */

#define PPP_IPCP	0x8021	/* IP Control Protocol */
#define PPP_OSICP	0x8023	/* OSI Network Layer Control Protocol */
#define PPP_NSCP	0x8025	/* Xerox NS IDP Control Protocol */
#define PPP_DECNETCP	0x8027	/* DECnet Control Protocol */
#define PPP_APPLECP	0x8029	/* Appletalk Control Protocol */
#define PPP_IPXCP	0x802b	/* Novell IPX Control Protocol */
#define PPP_STIICP	0x8033	/* Strean Protocol Control Protocol */
#define PPP_VINESCP	0x8035	/* Banyan Vines Control Protocol */
#define PPP_IPV6CP	0x8057	/* IPv6 Control Protocol */
#define PPP_CCP		0x80fd	/* Compress Control Protocol */

#define PPP_LCP		0xc021	/* Link Control Protocol */
#define PPP_PAP		0xc023	/* Password Authentication Protocol */
#define PPP_LQM		0xc025	/* Link Quality Monitoring */
#define PPP_CHAP	0xc223	/* Challenge Handshake Authentication Protocol */
#define PPP_BACP	0xc02b	/* Bandwidth Allocation Control Protocol */
#define PPP_BAP		0xc02d	/* BAP */
#define PPP_MP		0xc03d	/* Multi-Link */


// ###### Handler for PPP frames ############################################
void NetworkMonitor::pppInterfaceHandler(u_char*            userData,
                                         const pcap_pkthdr* pcapheader,
                                         const u_char*      packet)
{
   NetworkMonitor*  monitor  = (NetworkMonitor*)userData;
   cardinal length           = pcapheader->len - sizeof(PPP_HDRLEN);
   u_char* p = (u_char*)packet;
   card16 protocol;


   // ====== Scan address and control field =================================
   if(length < 2) {
      return;
   }
   if((*p == PPP_ADDRESS) && (*(p + 1) == PPP_CONTROL)) {
      p      += 2;                     // ACFC not used
      length -= 2;
   }
   if (length < 2) {
      return;
   }
   if(*p % 2) {
      protocol = *p;                   // PFC is used
      p++;
      length--;
   } else {
      protocol = translate16(*((card16*)p));
      p += 2;
      length -= 2;
   }
   u_char* payload = p;


   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;

   switch(protocol) {
      case PPP_IP:
         monitor->ip4PacketHandler((const ip*)payload,length);
        break;
      case PPP_IPV6:
         monitor->ip6PacketHandler((const ip6_hdr*)payload,length);
        break;
      case ETHERTYPE_ARP:
       break;
      default:
#ifdef PRINT_UNKNOWN
         printf("Unknown protocol %04x\n",protocol);
#endif
        break;
   }

   monitor->unsynchronized();
}


// ###### Handler for raw IPv4 frames #######################################
void NetworkMonitor::rawInterfaceHandler(u_char*            userData,
                                         const pcap_pkthdr* pcapheader,
                                         const u_char*      packet)
{
   NetworkMonitor* monitor = (NetworkMonitor*)userData;
   const cardinal length   = pcapheader->len;

   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;
   monitor->ip4PacketHandler((const ip*)packet,length);
   monitor->unsynchronized();
}


}
