/*
 *  $Id: networkmonitor.h,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Network Monitor
 *
 */


#ifndef NETWORKMONITOR_H
#define NETWORKMONITOR_H


#include "tdsystem.h"
#include "networkmonitorinterface.h"
#include "thread.h"
#include "internetaddress.h"
#include "streamsrcdest.h"
#include "streammonitorreport.h"


#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/icmp6.h>

extern "C" {
#include <pcap.h>
}


namespace Coral {


/**
  * This class implements a network monitor, which sums bytes and packets
  * transmitted over a given network interface (eth0, lo, ...). The interface
  * will be set to promiscuous mode. Note: root permissions are required
  * for monitoring a network interface!
  *
  * @short   Network Monitor
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
class NetworkMonitor : virtual public NetworkMonitorInterface,
                       public Thread
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor for new NetworkMonitor. The new monitors's thread has to
     * be started by calling start()!
     *
     * @param device Device name (eth0, lo, ...). Default: NULL (try automatic detection).
     *
     * @see Thread#start
     */
   NetworkMonitor(StreamMonitorReport* smr = NULL, char* device = NULL);

   /**
     * Destructor.
     */
   ~NetworkMonitor();


   // ====== Status functions ===============================================
   /**
     * ready() Implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#ready
     */
   bool ready() const;

   /**
     * Get device name, NetworkMonitor is using. This can be used to get the
     * name of the automatically chosen device name (name = NULL in
     * contructor).
     *
     * @return Device name.
     */
   inline const char* getDevice() const;


   // ====== Lock functions =================================================
   /**
     * lock() implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#lock
     */ 
   void lock();

   /**
     * unlock() implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#unlock
     */ 
   void unlock();


   // ====== Report functions ===============================================
   /**
     * resetReport() Implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#resetReport
     */
   void resetReport();

   /**
     * getReport() Implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#getReport
     */
   void getReport(NetworkMonitorReport* report);

   /**
     * moveReport() Implementation of NetworkMonitorInterface.
     *
     * @see NetworkMonitorInterface#moveReport
     */
   void moveReport(NetworkMonitorReport* report);


   // ====== Private data ===================================================
   private:
   struct sctphdr {
      card16 src_port;
      card16 dest_port;
      card32 verification_tag;
      card32 checksum;
   };
   struct sctpchunkhdr {
      card8  chunk_id;
      card8  chunk_flags;
      card16 chunk_length;
   };
   struct sctpdatachunkhdr {
      sctpchunkhdr header;
      card32       tsn;
      card16       stream_id;
      card16       stream_sn;
      card32       protocolId;
   };

   void run();

   void handleSCTP(const card8    protocol,
                   StreamSrcDest* stream,
                   const sctphdr* tcp,
                   const cardinal sctpLength,
                   const cardinal rawLength);
   void handleTCP(const card8    protocol,
                  StreamSrcDest* stream,
                  const tcphdr*  tcp,
                  const cardinal tcpLength,
                  const cardinal rawLength);
   void handleUDP(const card8    protocol,
                  StreamSrcDest* stream,
                  const udphdr*  udp,
                  const cardinal udpLength,
                  const cardinal rawLength);
   void handleICMP4(const card8    protocol,
                    StreamSrcDest* stream,
                    const icmp*    icmp,
                    const cardinal icmpLength,
                    const cardinal rawLength);
   void handleICMP6(const card8      protocol,
                    StreamSrcDest*   stream,
                    const icmp6_hdr* icmp,
                    const cardinal   icmpLength,
                    const cardinal   rawLength);

   void ip4PacketHandler(const ip* ip, const cardinal length);
   void ip6PacketHandler(const ip6_hdr* ip, const cardinal length);


   static void loopbackInterfaceHandler(u_char*            userData,
                                        const pcap_pkthdr* pcapheader,
                                        const u_char*      packet);

   static void ethernetInterfaceHandler(u_char*            userData,
                                        const pcap_pkthdr* pcapheader,
                                        const u_char*      packet);

   static void pppInterfaceHandler(u_char*            userData,
                                   const pcap_pkthdr* pcapheader,
                                   const u_char*      packet);

   static void rawInterfaceHandler(u_char*            userData,
                                   const pcap_pkthdr* pcapheader,
                                   const u_char*      packet);


   static const cardinal DefaultSnaplen = 78; // Length of saved portion of packet


   card64                TimeStamp;
   NetworkMonitorReport  Report;

   pcap_t*               PCapDescriptor;
   char*                 PCapDevice;
   pcap_handler          PCapHandler;
   cardinal              PCapInterfaceType;
   StreamMonitorReport*  SMReport;
};


}


#include "networkmonitor.icc"


#endif
