/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: Module for handling heartbeats between daemon and UAs 
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include "glib.h"
#include <time.h>
#include "pacemaker.h"
#include "timer_event.h"
#include "sctpdutils.h"

unsigned int heartbeat_interval=5000;
unsigned int heartbeat_timeout=10000;
GSList *heartbeatList=NULL;

int insertHeartbeat(unsigned short ua_id,unsigned int timer_id) {

  static unsigned int heartbeat_id=1;
  Heartbeat *heartbeat;
  time_t curTime;

  if ((heartbeat=(Heartbeat *)malloc(sizeof(Heartbeat)))!=NULL) {
    /* Fill in the Heartbeat data struct */
    heartbeat->heartbeat_id = heartbeat_id++;
    heartbeat->ua_id = ua_id;
    time(&curTime);
    heartbeat->time_sent = curTime;
    heartbeat->timer_id = timer_id;

    /* Insert pointer to Timer Event data structure into heartbeatList */
    heartbeatList = g_slist_append(heartbeatList,heartbeat);
    
#ifdef SHOW_HEARTBEATS
    fprintf(stdout,"Heartbeat (id = %d) Inserted\n",(heartbeat_id -1));
    fflush(stdout);
#endif    

    return (heartbeat_id-1);

  }

  return 0;

}


int removeHeartbeat(unsigned int heartbeat_id) {

  Heartbeat *heartbeat=NULL;  

  /* Find the Heartbeat first */
  if (findHeartbeat(heartbeat_id,&heartbeat)) {

    /* Remove the event */
    if (heartbeat!=NULL) {

      heartbeatList = g_slist_remove(heartbeatList,heartbeat);

      /* Free up the memory used by the data struct */
      free (heartbeat);
      
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Heartbeat (id = %d) removed from state\n",heartbeat_id);
      fflush(stdout);
#endif
      

    return 1;
    }
    else {
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Null value returned for this heartbeat (heartbeat_id=%d)\n",heartbeat_id);
      fflush(stdout);
#endif
    }
  }
  else {
#ifdef SHOW_HEARTBEATS
    fprintf(stdout,"Unable to find Heartbeat Event for this heartbeat (heartbeat_id=%d)\n",
	    heartbeat_id);
    fflush(stdout);
#endif
  }

  return 0;

}


int findHeartbeat (unsigned int heartbeat_id,Heartbeat **heartbeat) {


  Heartbeat *hb;
  GSList *node;

  /* Search */

  for (node = heartbeatList; node ; node = node->next) {

    hb = (Heartbeat *)node->data;

    if (hb->heartbeat_id == heartbeat_id) {
      *heartbeat = hb;
      return 1;
    }

  }

  return 0;

}


int findHeartbeatForUA(unsigned short ua_id) {

  Heartbeat *hb;
  GSList *node;

  /* Search */  
  for (node = heartbeatList; node ; node = node->next) {

    hb = (Heartbeat *)node->data;

    if (hb->ua_id == ua_id) {
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Heartbeat found\n");
      fflush(stdout);
#endif
      return hb->heartbeat_id;
    }
    else {
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Heartbeat not found\n");
      fflush(stdout);
#endif
    }

  }

  return 0;


}


void handle_heartbeat_ack (unsigned int heartbeat_id) {

  Heartbeat *heartbeat;
  unsigned int timer_id;
  TimerEvent *tEvent;
  time_t curTime;

#ifdef SHOW_HEARTBEATS
  fprintf(stdout,"Heartbeat ACK received\n");
  fflush(stdout);
#endif

  /* Locate Heartbeat from List */
  if (findHeartbeat(heartbeat_id, &heartbeat)) {

    timer_id = heartbeat->timer_id;

    /* Stop and remove the timer */
    if (stopTimerEvent(timer_id))
      removeTimerEvent(timer_id);
    else {

#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Unable to stop timer, id = %d\n",timer_id);
      fflush(stdout);
#endif

    }

    /* Remove the heartbeat from List */
    removeHeartbeat(heartbeat_id);

    /* Start the Heartbeat Interval Timer again */
    tEvent = malloc(sizeof(TimerEvent));
    time(&curTime);
    tEvent->ua_id=heartbeat->ua_id;
    tEvent->time_expired = curTime + heartbeat_interval;
    tEvent->timer_type = HEARTBEAT_INT_TIMER;
    insertTimerEvent(tEvent);

  }
  else {

#ifdef SHOW_HEARTBEATS
    fprintf(stdout,"Unable to locate this heartbeat, id = %d\n",heartbeat_id);
    fflush(stdout);
#endif

  }


}


void handle_heartbeat_expired (unsigned int timer_id) {

  TimerEvent *tEvent=NULL;

#ifdef SHOW_HEARTBEATS
  fprintf(stdout,"Heartbeat expired\n");
#endif

  if (findTimerEvent(timer_id,&tEvent)) {

    /* Terminate the UA */
    fprintf(stdout,"UA with ID = %d considered dead, terminating...\n",
	    tEvent->ua_id);
    /* This is not a good way, since i'm using an implicit declaration here*/
    /* TO DO: Find a way to clean up */
    /* handle_unregister_UA(tEvent->ua_id); */
    handle_sctpd_stopLibrary(tEvent->ua_id);

  }
  else {
#ifdef SHOW_HEARTBEATS
    fprintf(stdout,"Unable to find Timer Event for id = %d\n",timer_id);
#endif
  }
  
  fflush(stdout);

}


void handle_heartbeat_int_expired (unsigned int timer_id) {

  TimerEvent *tEvent;
  unsigned short ua_id;
  time_t curTime;
  unsigned int new_timer_id;

#ifdef SHOW_HEARTBEATS
  fprintf(stdout,"Heartbeat Interval expired\n");
  fflush(stdout);
#endif

  /* Find the offending timer event */
  findTimerEvent(timer_id,&tEvent);

  ua_id = tEvent->ua_id;

  /* Remove Timer Event */
  removeTimerEvent(timer_id);
  
  /* Start the Heartbeat Timer */
  tEvent = (TimerEvent *)malloc(sizeof(TimerEvent));
  time(&curTime);
  tEvent->ua_id=ua_id;
  tEvent->time_expired = curTime + heartbeat_timeout;
  tEvent->timer_type = HEARTBEAT_TIMER;
  new_timer_id = insertTimerEvent(tEvent);
  
  /* Send a heartbeat to the UA */
  send_Heartbeat_To_UA(ua_id,new_timer_id);

}



int send_Heartbeat_To_UA(unsigned short ua_id,unsigned int timer_id) {

  unsigned int heartbeat_id;
  if ((heartbeat_id = insertHeartbeat(ua_id,timer_id))!=0) {

#ifdef SHOW_HEARTBEATS    
    fprintf(stdout,"Sending out Heartbeat with ID = %d\n",heartbeat_id);
    fflush(stdout);
#endif

    return send_heartbeat(ua_id,heartbeat_id);
  }
  return 0;

}


int startPacemaker (unsigned short ua_id) {

  TimerEvent *tEvent;
  time_t curTime;

  /* Set the Heartbeat Interval timer running for this UA */
  tEvent = malloc(sizeof(TimerEvent));
  time(&curTime);
  tEvent->ua_id=ua_id;
  tEvent->time_expired = curTime + heartbeat_interval;
  tEvent->timer_type = HEARTBEAT_INT_TIMER;
  insertTimerEvent(tEvent);


  return 1;

}


int cancelTimersForUA(unsigned short ua_id) {

  int timer_id = 0;

  /* Find the timer_id */
  while ((timer_id = findTimerEventForUA(ua_id))!=0) {
    if (removeTimerEvent(timer_id)) {
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Timer %d successfully removed \n",timer_id);
      fflush(stdout);
#endif
    }
    else {
#ifdef SHOW_HEARTBEATS
      fprintf(stdout,"Timer not removed \n");    
      fflush(stdout);
#endif
    }
  }

  return 0;

}


/* Set the heartbeat interval */
void setHeartbeatInt(unsigned int t) {

  heartbeat_interval = t;

}


/* Set the heartbeat timeout */
void setHeartbeatTimeout(unsigned int t) {

  heartbeat_timeout = t;
  
}

