/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: Module contains functions to handle timer events
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 *
 */


#include <stdlib.h>
#include "timer_event.h"
#include <glib.h>
#include "sctp.h"
#include <stdio.h>
#include "../sctpdlib/sctpudputils.h"


/* Internal state list of timer events */
GSList *timerList = NULL;


/* Functions to send notifications of timer expiry to daemon ----------------------*/
int notify_HB_Timer_Expired (unsigned int timer_id) {

  fnCall fc;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

  /* create a UDP mesg for the notification */
  create_func_call (&fc,HEARTBEAT_EXPIRED);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(timer_id),&timer_id);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  return 1;      

}


int notify_HB_INT_Timer_Expired (unsigned int timer_id) {

  fnCall fc;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

  /* create a UDP mesg for the notification */
  create_func_call (&fc,HEARTBEAT_INT_EXPIRED);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(timer_id),&timer_id);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  return 1;      

}

/* --------------------------------------------------------------------------------- */

/* Callback function supplied to the SCTP library, to be called
   when the timer expired. */
void timer_expired_callback (unsigned int timer_id, void * param1, 
			     void * param2) {

  TimerEvent *tEvent;

#ifdef SHOW_HEARTBEATS 
  fprintf(stdout,"Timer Expired (ID=%d)\n",timer_id);
  fflush(stdout);
#endif

  /* Check the type of the event that has expired and sent the 
     respective notfication to the daemon */
  if (findTimerEvent(timer_id,&tEvent)) {
    
    if (tEvent!=NULL) {

      switch (tEvent->timer_type) {

      case HEARTBEAT_TIMER:

	/* Heartbeat ACK has not been received before timeout */
	notify_HB_Timer_Expired(tEvent->timer_id);

	break;

      case HEARTBEAT_INT_TIMER:

	/* Time to send out a heartbeat */
	notify_HB_INT_Timer_Expired(tEvent->timer_id);

	break;

      }

    }
    
  }

}


int insertTimerEvent(TimerEvent *tEvent) {

  time_t currentTime;
  unsigned int timerInterval,timer_id;

  /* Compute the difference in time between now and the expected expiry time */
  time(&currentTime);
  if (currentTime<tEvent->time_expired) {

    timerInterval = tEvent->time_expired - currentTime ;

    /* Start the timer based on the timer interval */
    timer_id = sctp_startTimer(timerInterval,timer_expired_callback,NULL,NULL);

    /* Put the timer id into the TimerEvent data struct */
    tEvent->timer_id = timer_id;

    /* Insert pointer to Timer Event data structure into timerList */
    timerList = g_slist_append(timerList,tEvent);
    
#ifdef SHOW_HEARTBEATS 
    fprintf(stdout,"Timer Event (id = %d) Inserted\n",timer_id);
    fflush(stdout);
#endif 

    return timer_id;

  }  

  return 0;

}


int removeTimerEvent(unsigned int timer_id) {

  TimerEvent *tEvent=NULL;

  /* Find the Timer Event first */
  if (findTimerEvent(timer_id,&tEvent)) {
    /* Remove the event */
    if (tEvent!=NULL) {
      timerList = g_slist_remove(timerList,tEvent);
      /* Free up the memory used by the data struct */
      free (tEvent);

      /* Terminate the timer from SCTP lib*/
      sctp_stopTimer(timer_id);

#ifdef SHOW_HEARTBEATS      
      fprintf(stdout,"Timer Event (id=%d) Removed\n",timer_id);
      fflush(stdout);
#endif      
    return 1;
    }
    else {

#ifdef SHOW_HEARTBEATS      
      fprintf(stdout,"Null value returned for this timer (timer_id=%d)\n",timer_id);
      fflush(stdout);
#endif
    }
  }
  else {

#ifdef SHOW_HEARTBEATS      
    fprintf(stdout,"Unable to find Timer Event for this timer (timer_id=%d)\n",timer_id);
    fflush(stdout);
#endif

  }

  return 0;

}


int findTimerEvent (unsigned int timer_id,TimerEvent **tEvent) {

  TimerEvent *timerevent;
  GSList *node;

  /* Search */

  for (node = timerList; node ; node = node->next) {

    timerevent = (TimerEvent *)node->data;

    if (timerevent->timer_id == timer_id) {
      *tEvent = timerevent;
      return 1;
    }

  }

  return 0;

}


int findTimerEventForUA (unsigned ua_id){

  TimerEvent *timerevent;
  GSList *node;

  /* Search */

  for (node = timerList; node ; node = node->next) {

    timerevent = (TimerEvent *)node->data;

    if (timerevent->ua_id == ua_id) {
      return timerevent->timer_id;
    }

  }

  return 0;



}



int stopTimerEvent(unsigned int timer_id) {

  if (sctp_stopTimer(timer_id)==0) 
    return 1;
  else
    return 0;

}
















