/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Daemon-based echo server with GUI and basic Assoc / Path status display
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 *
 */


#include <stdio.h>
#include <stdlib.h>         /* for atoi() under Linux */

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <signal.h>

#include "../sctpdlib/ua.h"
#include "sctp_info_utils.h"

#include <glib.h>
#include <gtk/gtk.h>

/* DEFINITION OF CONSTANTS ------------------------------------------ */
#define DEFAULT_X_SIZE    500
#define DEFAULT_Y_SIZE    800

#define UDP_Notif_Port    4234
#define UDP_Resp_Port   4235
#define SCTP_App_Port     7
#define MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES 5

pid_t echo_tool_pid;

/* NETWORK VARIABLES ------------------------------------------------ */
static int udpfd;                    // file descriptor to the UDP

/* GTK VARIABLES ---------------------------------------------------- */
GtkWidget *mainWindow;
GtkWidget *assocListBox;
GtkWidget *assocDetailsTree;
GtkWidget *detailsNode;
GtkWidget *statusWindow;
GtkWidget *dataWindow;

/* Other variables */
static unsigned char localAddressList[MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES][SCTP_MAX_IP_LEN];
static unsigned short noOfLocalAddresses = 0;
static int unknownCommand = 0;
static char tempstr[255];


/* EXIT FUNCTIONS */
void cleanUpandExit() {

  gtk_main_quit();

  /* Stop the sctpdlib */
  sctpd_stopLibrary();



}


void CloseMonitor (GtkWidget * window, gpointer data) {

  cleanUpandExit();

}


/* Arguments Functions ------------------------------------------------------- */
void printUsage(void) {

  printf("Usage:   echo_server_monitor [options] -s source_addr_1 -s source_addr_2 ...\n");

}

void getArgs(int argc, char **argv) {

    int c;
    extern char *optarg;
    extern int optind;

    while ((c = getopt(argc, argv, "s:")) != -1)
    {
        switch (c) {
        case 's':
            if ((noOfLocalAddresses < MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES) &&
                (strlen(optarg) < SCTP_MAX_IP_LEN  )) {
                strcpy(localAddressList[noOfLocalAddresses], optarg);
                noOfLocalAddresses++;
            }
            break;
        default:
            unknownCommand = 1;
            break;
        }
    }
}

void checkArgs(void) {
    int abortProgram;
    int printUsageInfo;
    
    abortProgram = 0;
    printUsageInfo = 0;
    
    if (noOfLocalAddresses == 0) {
        printf("Error:   At least one sourceaddress must be specified.\n");
        abortProgram = 1;
        printUsageInfo = 1;
    }
    if (unknownCommand ==1) {
         printf("Error:   Unkown options in command.\n");
         printUsageInfo = 1;
    }
    
    if (printUsageInfo == 1)
        printUsage();
    if (abortProgram == 1)
        exit(-1);
}


/* define functions to handle the GUI ------------------------------- */
void displayStatus(char *status) {

  gtk_container_add(GTK_CONTAINER(statusWindow),gtk_list_item_new_with_label(status));
  gtk_widget_show_all(statusWindow);

}

void displayData(char *data) {

  gtk_container_add(GTK_CONTAINER(dataWindow),gtk_list_item_new_with_label(data));
  gtk_widget_show_all(dataWindow);

}


// Update the display in the assocDetailsTree UI
int updateAssocDetails(int assocID) {

  int i;

  assocInfo *associnfo;
  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;

  GtkWidget *detailsSubtree;
  GtkWidget *tnode;
  GtkWidget *pathsNode;
  GtkWidget *pathsSubtree;
  GtkWidget *pathNode;
  GtkWidget *pathSubtree;

  gchar tstr[50];              // temp variable for formatting output


  // Retrieve information from internal storage
  associnfo = get_assocInfo(assocID);
  if (associnfo==NULL)
    return -1;
 
  assocStatus = associnfo->assocStatus;
 
  // Clear the current displayed tree.
  gtk_tree_clear_items(GTK_TREE(assocDetailsTree),0,1);

  // create details node
  sprintf(tstr,"Details (Assoc ID=%d):",assocID);
  detailsNode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(assocDetailsTree),detailsNode);
  gtk_widget_show(detailsNode);

  // create details sub tree
  detailsSubtree = gtk_tree_new();
  gtk_tree_item_set_subtree(GTK_TREE_ITEM(detailsNode),detailsSubtree);

  // Go through the details of the Association Status to display
  // State
  sprintf(tstr,"State : %d",assocStatus.state);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);


  // Number of Addresses
  sprintf(tstr,"No. Of Addresses : %d",assocStatus.numberOfAddresses);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // Primary Destination Address
  sprintf(tstr,"Primary Destination Address : %s",assocStatus.primaryDestinationAddress);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // outStreams
  sprintf(tstr,"No. of Outgoing streams : %d",assocStatus.outStreams);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // inStreams
  sprintf(tstr,"No. Of Incoming Streams : %d",assocStatus.inStreams);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // primaryAddressIndex
  sprintf(tstr,"Primary Address Index : %d",assocStatus.primaryAddressIndex);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // currentReceiverWindowSize
  sprintf(tstr,"Current Receiver Window Size : %d",assocStatus.currentReceiverWindowSize);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // rtoInitial
  sprintf(tstr,"Initial Round Trip Timeout : %d",assocStatus.rtoInitial);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // rtoMin
  sprintf(tstr,"Minimun Round Trip Timeout : %d",assocStatus.rtoMin);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // validCookieLife
  sprintf(tstr,"Cookie Lifetime : %d",assocStatus.validCookieLife);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // assocMaxRetransmits
  sprintf(tstr,"Maximum retransmission per association : %d",assocStatus.assocMaxRetransmits);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // pathMaxRetransmits
  sprintf(tstr,"Maximum retransmission per path : %d",assocStatus.pathMaxRetransmits);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // maxInitRetransmits
  sprintf(tstr,"Maximum Initial Retransmission : %d",assocStatus.maxInitRetransmits);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);

  // myRwnd
  sprintf(tstr,"Local Receiver Window : %d",assocStatus.myRwnd);
  tnode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
  gtk_widget_show(tnode);


  // PATHS INFO -----------------------------------------------------------------------

  // create paths node
  strcpy(tstr,"Paths");
  pathsNode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_append(GTK_TREE(detailsSubtree),pathsNode);
  gtk_widget_show(pathsNode);

  // create paths sub tree
  pathsSubtree = gtk_tree_new();
  gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathsNode),pathsSubtree);

  for (i=0;i<associnfo->no_paths;i++)
  {

    // get the pathStatus
    pathStatus = associnfo->paths[i].pathStatus;

    sprintf(tstr,"Paths ID=%d",associnfo->paths[i].pathID);
    pathNode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathsSubtree),pathNode);
    gtk_widget_show(pathNode);

    // create path sub tree
    pathSubtree = gtk_tree_new();
    gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathNode),pathSubtree);


    // destination address
    sprintf(tstr,"Destination Address : %s",pathStatus.destinationAddress);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // state
    switch (pathStatus.state) {
    case 0:
      strcpy(tstr,"State : ACTIVE");
      break;
    case 1:
      strcpy(tstr,"State : INACTIVE");
      break;
    default:
      strcpy(tstr,"State : UNKNOWN");
      break;
    }
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // srtt
    sprintf(tstr,"Smooth Round Trip Time : %d",pathStatus.srtt);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // rto
    sprintf(tstr,"Round Trip Time : %d",pathStatus.rto);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // rttvar
    sprintf(tstr,"Round Trip Time Variation : %d",pathStatus.rttvar);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // heartbeatInterval
    sprintf(tstr,"Heartbeat Interval : %d",pathStatus.heartbeatIntervall);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // cwnd
    sprintf(tstr,"Congestion Window Size : %d",pathStatus.cwnd);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);


    // cwnd2
    sprintf(tstr,"Congestion Window size 2 : %d",pathStatus.cwnd2);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // Slow start Threshold
    sprintf(tstr,"Slow start threshold : %d",pathStatus.ssthresh);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // MTU
    sprintf(tstr,"MTU : %d",pathStatus.mtu);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);

    // ipTos
    sprintf(tstr,"IP Type Of Service : %d",pathStatus.ipTos);
    tnode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(pathSubtree),tnode);
    gtk_widget_show(tnode);


  }
  
  gtk_widget_show_all(pathsNode);
  gtk_tree_item_expand(GTK_TREE_ITEM(detailsNode));

  return 1;
}


int updateAssocList(int assocID,SCTP_AssociationStatus assocStatus) {

  gchar status[20];
  gchar *values[3];

  switch (assocStatus.state) {
  case SCTP_CLOSED :
    strcpy(status,"CLOSED");
    break;
  case SCTP_COOKIE_WAIT :
    strcpy(status,"COOKIE_WAIT");
    break;
  case SCTP_COOKIE_ECHOED :
    strcpy(status,"COOKIE_ECHOED");
    break;
  case SCTP_ESTABLISHED :
    strcpy(status,"ESTABLISHED");
    break;
  case SCTP_SHUTDOWN_PENDING :
    strcpy(status,"SHUTDOWN_PENDING");
    break;
  case SCTP_SHUTDOWN_RECEIVED :
    strcpy(status,"SHUTDOWN_RECEIVED");
    break;
  case SCTP_SHUTDOWN_SENT :
    strcpy(status,"SHUTDOWN_SENT");
    break;
  case SCTP_SHUTDOWNACK_SENT :
    strcpy(status,"SHUTDOWNACK_SENT");
    break;

  }

  values[0]=malloc(sizeof(assocID)*sizeof(gchar));
  sprintf(values[0] ,"%d",assocID);
  values[1]=malloc(sizeof(assocStatus.primaryDestinationAddress)*sizeof(gchar));
  strcpy(values[1],assocStatus.primaryDestinationAddress);
  values[2]=malloc(sizeof(status)*sizeof(gchar));
  sprintf(values[2],"%s",status);

  /* append to the list and update the row information into the assocInfoArr  */
  update_assocRow(assocID, gtk_clist_append (GTK_CLIST(assocListBox),values));

  free(values[0]);
  free(values[1]);
  free(values[2]);

  /*updateAssocDetails(assocID); */

  return 1;
}


int removeAssocFromList(assocID) {

  int row;

  /* Look for the row of this assocID */
  row =  get_assocRow(assocID);

  if (row!=-1) {
    gtk_clist_remove(GTK_CLIST(assocListBox),row);
    remove_assocInfo(assocID);

    // if there is no more assoc on the list, remove the details node
    if (row==0) {
      gtk_tree_clear_items(GTK_TREE(assocDetailsTree),0,1);
      return 1;
    }
  }
  else {
    sprintf(tempstr,"removeAssocFromList:Error in locating row.\nID = %d",assocID);
    displayStatus(tempstr);
  }
  return 0;
}

/* Code canabalize from book -------------------------------------------  */
/*
 * CloseDialog
 *
 * Close the dialog window.  The dialog handle is passed
 * in as the data.
 */
void CloseDialog (GtkWidget *widget, gpointer data) {

    /* --- Close it --- */
    gtk_widget_destroy (GTK_WIDGET (data));

}

/*
 * ClosingDialog
 *
 * Calls when window is about to close.  Returns FALSE
 * to let it close.  Releases the grab which made the dialog
 * window modal.
 */
void ClosingDialog (GtkWidget *widget, gpointer data) {
    gtk_grab_remove (GTK_WIDGET (widget));
}



/*
 * Popup
 *
 * Display a popup dialog window with a message and an
 * "Ok" button
 */
void Popup (char *szMessage) {
    static GtkWidget *label;
    GtkWidget *button;
    GtkWidget *dialog_window;

    /* --- Create a dialog window --- */
    dialog_window = gtk_dialog_new ();

    /* --- Trap the window close signal to release the grab --- */
    gtk_signal_connect (GTK_OBJECT (dialog_window), "destroy",
                     GTK_SIGNAL_FUNC (ClosingDialog),
                     &dialog_window);

    /* --- Add a title to the window --- */
    gtk_window_set_title (GTK_WINDOW (dialog_window), "Popup");

    /* --- Create a small border --- */
    gtk_container_border_width (GTK_CONTAINER (dialog_window), 5);

    /* -------------------------- */
    /* --- Create the message --- */
    /* -------------------------- */

    /* --- Create the message in a label --- */
    label = gtk_label_new (szMessage);

    /* --- Put some room around the label --- */
    gtk_misc_set_padding (GTK_MISC (label), 10, 10);

    /* --- Add the label to the dialog box --- */
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog_window)->vbox),
                        label, TRUE, TRUE, 0);

    /* --- Make the label visible --- */
    gtk_widget_show (label);

    /* ------------------- */
    /* --- "Ok" button --- */
    /* ------------------- */

    /* --- Create the "Ok" button --- */
    button = gtk_button_new_with_label ("Ok");

    /* --- Need to close the window if they press "Ok" --- */
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
                     GTK_SIGNAL_FUNC (CloseDialog),
                     dialog_window);

    /* --- Allow it to be the default button --- */
    GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

    /* --- Add the button to the dialog box --- */
    gtk_box_pack_start (GTK_BOX (
                  GTK_DIALOG (dialog_window)->action_area),
                 button, TRUE, TRUE, 0);

    /* --- Make the button the default button --- */
    gtk_widget_grab_default (button);

    /* --- Make the button visible --- */
    gtk_widget_show (button);

    /* --- Make the dialog box visible --- */
    gtk_widget_show (dialog_window);

    gtk_grab_add (dialog_window);
}

/* -- End of canabalization ---------------------------------------------- */


/* GUI Callbacks Handler --------------------------------------- */

static void clist_select_row(GtkCList * clist,gint row, gint column, 
			     GdkEventButton * event , gpointer user_data) {

  int assocID;
  char * assocIDStr;
  assocInfo *associnfo;

  /* check what is the ID of the association selected... */
  
  gtk_clist_get_text(GTK_CLIST(clist),row,0,&assocIDStr);
  assocID = atoi(assocIDStr);
  associnfo = get_assocInfo(assocID);

  if (associnfo !=NULL){
    
    updateAssocDetails(assocID);
    
  }
  else {
    sprintf(tempstr,"AssocID not found!\nValue of Association ID selected = %d\n",assocID);
    displayStatus(tempstr);
  }

}


void menuExitHandler(GtkMenuItem * menuitem,gpointer data) {

  cleanUpandExit();

}



/* UDP Callback Handlers --------------------------------------------------------------- */

void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,unsigned short noOfInStreams, 
			   unsigned short noOfOutStreams, void* dummy) {

  int i;
  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;
  char tstr[255];

  displayStatus("Reveived a CommunicationUpNotif from SCTP layer");
  
  if (sctpd_getAssocStatus(assocID,&assocStatus)==0) {

    displayStatus("Received Association Status information");
 
    /* update local storage for assocIDs. */
    insert_assocID(assocID);
    displayStatus("Reveived a response for sctp_getAssocStatus from SCTP layer");

    if (update_assocInfo(assocID,assocStatus)) {
      updateAssocList(assocID,assocStatus);

      /* Get the path information */
      sprintf (tstr,"No. of paths in this association = %d\n", assocStatus.numberOfAddresses);
      displayStatus(tstr);

      for (i=0; i < assocStatus.numberOfAddresses; i++){   
	if (sctpd_getPathStatus(assocID, i, &pathStatus)==0) {
	  
	  displayStatus("Received Path Status information");

	  /* update local storage for pathIDs. */
	  insert_pathID(assocID,i);

	  if (update_pathInfo(assocID,i,pathStatus)) {
	    displayStatus("Internal list updated with new path status info");
	  }
	  updateAssocDetails(assocID);
	}
	else {
	    displayStatus("Unable to obtail Path Status Information");
	}
      }
    }
  }
  else {
    displayStatus("Unable to obtain Association Status Information");
  }

  return NULL;

}


void dataArriveNotif(unsigned int assocID, unsigned int streamID, unsigned int len,
                     unsigned int protoID, unsigned int unordered, void* ulpDataPtr) {

  int result,length;
  char tstr[255],data[255];

  
  sprintf(tstr,"Data Arrived from assocID=%d, stream = %d , size = %d",assocID,streamID,len);
  displayStatus(tstr);

  /* Read the data */
  bzero(data,255);
  if (sctpd_receive(assocID,streamID,data,&length,SCTP_MSG_DEFAULT)==0) {

    sprintf(tstr,"Data read from buffer, size = %d",length);
    displayStatus(tstr);

    /* Hack : Remove the trailing \n from the data as it affects the display... */
    bzero(tstr,255);
  
    if (data[len-1]=='\n')
      strncpy(tstr,data,len-1);
    else
      strncpy(tstr,data,len);
    
    displayData(tstr);

    /* Since this sctp_monitor is suppose to act like a echo server, it will bounce the data
       back to the sender */
  
    result = sctpd_send(assocID,
			streamID,
			data, len,
			protoID,
			SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, SCTP_INFINITE_LIFETIME, 
			unordered, SCTP_BUNDLING_DISABLED);
  
  }
  else {
    sprintf(tstr,"Uable to read data, len = %d",length);
    displayStatus(tstr);
  }

}


void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr) {

  displayStatus("Shutdown completed message received!\nRemoving entry from association list");

  removeAssocFromList(assocID);

}


void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, 
			      unsigned short newState, void* ulpDataPtr) {

  displayStatus("Network status changed Notification message received!");
  
  /* Alert user */
  Popup("Network Status Changed");

  /* TO DO: 1. Get new path_status info
            2. Update display */



}


void communicationLostNotif(unsigned int assocID, unsigned short status, void* ulpDataPtr) {

  char tstr[255];

  sprintf(tstr,"Communication Lost for association (ID=%d)",assocID);
  displayStatus(tstr);


}


void restartNotif(unsigned int assocID, void* ulpDataPtr) {

  char tstr[255];

  sprintf(tstr,"Restart for association (ID=%d)",assocID);
  displayStatus(tstr);


}



void udp_recv_callback (gpointer data,gint source,GdkInputCondition condition)
{

  char recvbuff[256];
  socklen_t len;
  struct sockaddr_in taddr;
  ssize_t mesglen;

  mesglen = recvfrom(source,recvbuff,sizeof(recvbuff),0,(struct sockaddr *) &taddr,&len);

  if (mesglen>0) {
    dispatchUAMesg(recvbuff);
  }


}



/* connect to the UDP port */
int initializeUDPListening(int port) {

  int sockflags;
  struct sockaddr_in uaddr;
  struct sockaddr_in paddr;

  udpfd = socket(AF_INET, SOCK_DGRAM,0);

  bzero (&uaddr,sizeof(uaddr));
  bzero (&paddr,sizeof(paddr));

  uaddr.sin_family = AF_INET;
  uaddr.sin_port = htons(port);
  uaddr.sin_addr.s_addr=htonl(INADDR_ANY);
  
  if ((bind(udpfd, (struct sockaddr *) &uaddr, sizeof(uaddr)))==0) {
    /* set socket as non-blocking */
    sockflags = fcntl (udpfd,F_GETFL,0);
    fcntl (udpfd,F_SETFL, sockflags | O_NONBLOCK);

    /* register the fd with the event handler of GDK */
    gdk_input_add(udpfd,GDK_INPUT_READ,udp_recv_callback,NULL);
    sprintf(tempstr,"Sucessfully binded to the application UDP port (%d)",port);
    displayStatus(tempstr);
    return 1;
  }

  else {
    displayStatus("unable to bind to UDP port");
  }

  return 0;

}



void BuildInterface() {

  GtkWidget *mainHBox;
  GtkWidget *assocListScrollWin;  
  GtkWidget *assocDetailsScrollWin;
  GtkWidget *hpaned1;
  GtkWidget *hpaned2;
  GtkWidget *menubar;
  GtkWidget *actionMenu;
  GtkWidget *actionMenuItem;
  GtkWidget *menuitem;
  GtkWidget *statusScrollWin;  
  GtkWidget *dataScrollWin;  

  gchar *titles[3] = {"AssocID","Destination Address","Status"};

  /* create and set the parameters of the window */
  mainWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position(GTK_WINDOW(mainWindow), GTK_WIN_POS_CENTER);
  gtk_window_set_default_size (GTK_WINDOW(mainWindow), DEFAULT_X_SIZE,DEFAULT_Y_SIZE);
  gtk_window_set_title (GTK_WINDOW(mainWindow), "SCTP Echo Server with Monitor");

  /* connect the signal handlers for the window */
  gtk_signal_connect (GTK_OBJECT(mainWindow), "destroy", GTK_SIGNAL_FUNC(CloseMonitor), NULL);

  /* create the main packing box */
  mainHBox = gtk_vbox_new (FALSE,5);
  gtk_container_add (GTK_CONTAINER(mainWindow), mainHBox);



  /* MENU */
  menubar=gtk_menu_bar_new();
 
  actionMenuItem = gtk_menu_item_new_with_label("Action");
  gtk_menu_bar_append(GTK_MENU_BAR(menubar),actionMenuItem);

  /* Action Menu */
  actionMenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(actionMenuItem),actionMenu);
  menuitem =  gtk_menu_item_new_with_label("Exit");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(menuExitHandler),NULL);
  

  gtk_box_pack_start (GTK_BOX(mainHBox), menubar, FALSE, FALSE, 0);

  /* Pane to house the assoc List and assoc Details */
  hpaned1 = gtk_vpaned_new();
  
  /* Association List display ------------------------------------------------- */
  /* create assocList to hold the association list  */
  assocListBox = gtk_clist_new_with_titles(3,titles); 
  /* create scroll window to place the clist */
  assocListScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					       GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_container_add(GTK_CONTAINER(assocListScrollWin),assocListBox);
  /* attach signal handler for selection */
  gtk_signal_connect (GTK_OBJECT(assocListBox),"select_row",GTK_SIGNAL_FUNC(clist_select_row),NULL);

  gtk_paned_add1(GTK_PANED(hpaned1),assocListScrollWin);


  /* Association Details Tree display ----------------------------------------- */
  /* create the tree to hold the details */
  assocDetailsTree = gtk_tree_new();
  /* create scroll window to place the tree in */
  assocDetailsScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					       GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(assocDetailsScrollWin),
					assocDetailsTree);
  gtk_paned_add2(GTK_PANED(hpaned1),assocDetailsScrollWin);
  gtk_paned_set_position(GTK_PANED(hpaned1),100);
  gtk_box_pack_start (GTK_BOX(mainHBox), hpaned1, TRUE, TRUE, 0);
  gtk_widget_show(hpaned1);


  /* Pane to house the data Window and Status Window*/
  hpaned2 = gtk_vpaned_new();

  /* Data Window */
  
  dataWindow = gtk_list_new();   
  dataScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					       GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(dataScrollWin),dataWindow);
  gtk_paned_add1(GTK_PANED(hpaned2),dataScrollWin); 

  /* Status Window */
  
  statusWindow = gtk_list_new();   
  statusScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					       GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(statusScrollWin),statusWindow);
  gtk_paned_add2(GTK_PANED(hpaned2),statusScrollWin);

  gtk_paned_set_position(GTK_PANED(hpaned2),150);
  gtk_box_pack_start (GTK_BOX(mainHBox), hpaned2, TRUE, TRUE, 0);
 

  /* display the window on the screen */
  gtk_widget_show_all(mainWindow);  



}



gint main (gint argc, gchar * argv[]) {

  UDP_UA_CB udp_callbacks;
 
  /* Trap Ctrl-C */
  signal(SIGINT, cleanUpandExit);

  /* handle all command line options */
  getArgs(argc, argv);
  checkArgs();

  gtk_init (&argc, &argv);

  BuildInterface();

  udp_callbacks.communicationUpNotif = &communicationUpNotif;
  udp_callbacks.shutdownCompleteNotif = &shutdownCompleteNotif;
  udp_callbacks.networkStatusChangeNotif = &networkStatusChangeNotif;
  udp_callbacks.dataArriveNotif = &dataArriveNotif;
  udp_callbacks.restartNotif = &restartNotif;

  initializeUDPListening(UDP_Notif_Port);

  /* Initialize sctpdlib */
  if (sctpd_initLibrary(UDP_Notif_Port,UDP_Resp_Port)==1) {

    /* Call sctpd_registerInstance */
    if (sctpd_registerInstance(SCTP_App_Port,10,10,noOfLocalAddresses,
			       localAddressList,udp_callbacks)) {

      /* Start the GTK event loop */
      gtk_main();
    }
    else {
      printf("Unable to register an instance with sctpd.\n");
      exit(1);
    }
  }
  else {
    printf("Unable to initialize the sctpdlib.\n");
    exit(1);
  }

  exit (0);
}



