/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "ua.h"

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>         /* for atoi() under Linux */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#define ECHO_PORT                             7
#define MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES    10
#define MAXIMUM_NUMBER_OF_ASSOCIATIONS        5
#define MAXIMUM_NUMBER_OF_IN_STREAMS         17
#define MAXIMUM_NUMBER_OF_OUT_STREAMS        17
#define MAXIMUM_PAYLOAD_LENGTH             8192
#define UDP_Notif_Port  9000
#define UDP_Resp_Port   9002

#define min(x,y)            (x)<(y)?(x):(y)

struct ulp_data {
    int maximumStreamID;
    unsigned int assocID;
    unsigned long nrOfReceivedChunks;
    unsigned long nrOfReceivedBytes;
};

static struct ulp_data ulpData[MAXIMUM_NUMBER_OF_ASSOCIATIONS];
static unsigned char localAddressList[MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES][SCTP_MAX_IP_LEN];
static unsigned char destinationAddress[SCTP_MAX_IP_LEN];

static unsigned short noOfLocalAddresses      = 0;
static unsigned short numberOfInitialPackets  = 0;
static unsigned short chunkLength             = 512;
static unsigned char  tosByte                 = 0x10;  /* IPTOS_LOWDELAY */
static unsigned int doMeasurements            = 0;
static unsigned int doAllMeasurements         = 0;
static int verbose                            = 0;
static int vverbose                           = 0;
static int unknownCommand                     = 0;
static unsigned int deltaT                    = 1000;
static int sendToAll                          = 0;                 
static int startAssociation                   = 0;
static int sendUnordered                      = 0;

static unsigned short sctpInstance=0;

void printUsage(void) {

  printf("usage:   echo_tool [options] -s source_addr_1 -s source_addr_2 ...\n");
  printf("options:\n");
  printf("-b                  send back incoming data on all existing associations\n");
  printf("-d destination_addr establish a association with the specified address\n");   
  printf("-l length           number of bytes of the payload when generating traffic (default 512)\n");   
  printf("-m                  print number of received bytes and chunks per period (see -p)\n");
  printf("-M                  print number of received bytes, chunks per period (see -p) and flow control info\n");
  printf("-n number           number of packets initially send out (default 0)\n");   
  printf("-p period           period for the measurements in milliseconds (default 1000)\n");   
  printf("-t byte             TOS byte used by all assiciations (default 0x10)\n");
  printf("-u                  inject the initial packets unordered\n");
  printf("-v                  verbose mode\n");   
  printf("-V                  very verbose mode\n");   
}

void getArgs(int argc, char **argv){

  int c;
  extern char *optarg;
  extern int optind;
  
  while ((c = getopt(argc, argv, "bd:l:mMn:p:s:t:uvV")) != -1){
    switch (c) {
    case 'b':
      sendToAll = 1;
      break;
    case 'd':
      if (strlen(optarg) < SCTP_MAX_IP_LEN) {
	strcpy(destinationAddress, optarg);
	startAssociation = 1;
      }
            break;
        case 'l':
            chunkLength = min(atoi(optarg),MAXIMUM_PAYLOAD_LENGTH);
            break;
        case 'm':
            doMeasurements = 1;
            break;
        case 'M':
            doMeasurements = 1;
            doAllMeasurements = 1;
            break;
        case 'n':
            numberOfInitialPackets = atoi(optarg);
            break;
        case 'p':
            deltaT = atoi(optarg);
            break;
        case 's':
            if ((noOfLocalAddresses < MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES) &&
                (strlen(optarg) < SCTP_MAX_IP_LEN  )) {
                strcpy(localAddressList[noOfLocalAddresses], optarg);
                noOfLocalAddresses++;
            }
            break;  
        case 't':
            tosByte = (unsigned char) atoi(optarg);
            break;
        case 'u':
            sendUnordered = 1;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'V':
            verbose = 1;
            vverbose = 1;
            break;
        default:
            unknownCommand = 1;
            break;
        }
    }
}


void checkArgs(void) {

    int abortProgram;
    int printUsageInfo;
    
    abortProgram = 0;
    printUsageInfo = 0;
    
    if (noOfLocalAddresses == 0) {
        printf("Error:   At least one sourceaddress must be specified.\n");
        abortProgram = 1;
        printUsageInfo = 1;
    }
    if (unknownCommand ==1) {
         printf("Error:   Unkown options in command.\n");
         printUsageInfo = 1;
    }
    
    if (printUsageInfo == 1)
        printUsage();
    if (abortProgram == 1)
        exit(-1);
}


/* Exit function */
void cleanUpAndExit() {

  /* Unregister Instance with sctpd */
  sctpd_unregisterInstance(sctpInstance);

  /* Stop the sctpdlib */
  sctpd_stopLibrary();
 
  exit(0);
}



void dataArriveNotif(unsigned int assocID, unsigned int streamID, unsigned int len,
                     unsigned int protoID, unsigned int unordered, void* ulpDataPtr){

    unsigned char chunk[MAXIMUM_PAYLOAD_LENGTH];
    int length, index=0, result;
 
    if (vverbose) {  
      fprintf(stdout, "%-8x: Data arrived (%u bytes on stream %u, %s)\n",
                      assocID, len, streamID, (unordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered");
      fflush(stdout);
    }
    /* read it */
    length = sizeof(chunk);
    sctpd_receive(assocID, streamID, chunk, &length,SCTP_MSG_DEFAULT);

    /* update counter */
    ((struct ulp_data *) ulpDataPtr)->nrOfReceivedChunks += 1;
    ((struct ulp_data *) ulpDataPtr)->nrOfReceivedBytes  += length;

    /* and send it */
    if (sendToAll) {
        for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
            if (ulpData[index].maximumStreamID != -1) {
	      result = sctpd_send(ulpData[index].assocID,
				    min(streamID, ulpData[index].maximumStreamID),
				    chunk, length,
				    protoID,
				    SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, SCTP_INFINITE_LIFETIME, 
				    unordered, SCTP_BUNDLING_DISABLED);
                if (vverbose) {
                    fprintf(stdout, "%-8x: Data sent (%u bytes on stream %u, %s) Result: %d\n",
                                    ulpData[index].assocID, len, min(streamID, ulpData[index].maximumStreamID),
                                    (unordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered", result);
                    fflush(stdout);
                }
            }
        }
    } else {
        result = sctpd_send(assocID,
			    min(streamID, ((struct ulp_data *) ulpDataPtr)->maximumStreamID),
			    chunk, length,
			    protoID,
			    SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, SCTP_INFINITE_LIFETIME, 
			    unordered, SCTP_BUNDLING_DISABLED);
        if (vverbose) {
            fprintf(stdout, "%-8x: Data sent (%u bytes on stream %u, %s) Result: %d\n",
                            ulpData[index].assocID, len, min(streamID, ((struct ulp_data *) ulpDataPtr)->maximumStreamID),
                            (unordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered", result);
            fflush(stdout);
        }
    }
}

void sendFailureNotif(unsigned int assocID,
                      unsigned char *unsent_data, 
		      unsigned int dataLength, 
		      unsigned int *context, void* dummy){

  if (verbose) {  
    fprintf(stdout, "%-8x: Send failure\n", assocID);
    fflush(stdout);
  }
}

void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, 
			      unsigned short newState, void* ulpDataPtr){

    SCTP_AssociationStatus assocStatus;
    SCTP_PathStatus pathStatus;
    unsigned short pathID;
    
    if (verbose) {
        sctpd_getPathStatus(assocID, affectedPathID, &pathStatus);
        fprintf(stdout, "%-8x: Network status change: path %u (towards %s) is now %s\n", 
                        assocID, affectedPathID,
                        pathStatus.destinationAddress,
                        ((newState == SCTP_PATH_OK) ? "ACTIVE" : "INACTIVE"));
        fflush(stdout);
    }
    
    /* if the primary path has become inactive */
    if ((newState == SCTP_PATH_UNREACHABLE) &&
        (affectedPathID == sctpd_getPrimary(assocID))) {
        
        /* select a new one */
        sctpd_getAssocStatus(assocID, &assocStatus);
        for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
            sctpd_getPathStatus(assocID, pathID, &pathStatus);
            if (pathStatus.state == SCTP_PATH_OK)
                break;
        }
        
        /* and use it */
        if (pathID < assocStatus.numberOfAddresses) {
            sctpd_setPrimary(assocID, pathID);
        }
    }
}

void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,
                           unsigned short noOfInStreams, 
			   unsigned short noOfOutStreams, void* dummy){
	
    unsigned int index, packetNumber;
    unsigned char chunk[MAXIMUM_PAYLOAD_LENGTH];
    SCTP_PathStatus pathStatus;
    unsigned short pathID;
     
    if (verbose) {  
        fprintf(stdout, "%-8x: Communication up (%u paths:", assocID, noOfPaths);
        for (pathID=0; pathID < noOfPaths; pathID++){
            sctpd_getPathStatus(assocID, pathID, &pathStatus);
            fprintf(stdout, " %s", pathStatus.destinationAddress);
        }
        fprintf(stdout, ")\n");
        fprintf(stdout, "%-8x:                  %u incoming, %u outgoing streams.\n",
                assocID, noOfInStreams, noOfOutStreams);
        fflush(stdout);
    }
  
    /* look for a free ULP data */
    for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
        if (ulpData[index].maximumStreamID == -1)
            break;
    }
   
    /* if found */
    if (index < MAXIMUM_NUMBER_OF_ASSOCIATIONS) {
        /* use it */
        ulpData[index].maximumStreamID = noOfOutStreams - 1;
        ulpData[index].assocID = assocID;
    
        /* send the initial packets */
        memset(chunk, 0, sizeof(chunk));
        for(packetNumber=1; packetNumber <= numberOfInitialPackets; packetNumber++) {
            sctpd_send(assocID,
                      0,
                      chunk, chunkLength,
                      SCTP_GENERIC_PAYLOAD_PROTOCOL_ID,
                      SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, SCTP_INFINITE_LIFETIME,
                      (sendUnordered)?SCTP_UNORDERED_DELIVERY:SCTP_ORDERED_DELIVERY, 
                      SCTP_BUNDLING_DISABLED);
        }
        return &ulpData[index];       
    } else {
        /* abort assoc due to lack of resources */
        sctpd_abort(assocID);
        return NULL;
    }
}

void communicationLostNotif(unsigned int assocID, unsigned short status, void* ulpDataPtr){
	
  unsigned char buffer[MAXIMUM_PAYLOAD_LENGTH];
  unsigned int bufferLength;
  unsigned short streamID, streamSN;
  unsigned int protoID;
    
  if (verbose) {
    fprintf(stdout, "%-8x: Communication lost (status %u)\n", assocID, status);
    fflush(stdout);
  }
  
  /* retrieve data */
  bufferLength = sizeof(buffer);
  while (sctpd_receiveUnsent(assocID, buffer, &bufferLength,           
			     &streamID, &streamSN, &protoID) > 0){
    /* do something with the retrieved data */
    /* after that, reset bufferLength */
    bufferLength = sizeof(buffer);
  }
    
  bufferLength = sizeof(buffer);
  while (sctpd_receiveUnacked(assocID, buffer, &bufferLength,
			      &streamID, &streamSN, &protoID) > 0){
    /* do something with the retrieved data */
    /* after that, reset bufferLength */
    bufferLength = sizeof(buffer);
  }
                      
  /* free ULP data */
  ((struct ulp_data *) ulpDataPtr)->maximumStreamID = -1;
    
  /* delete the association */
  sctpd_deleteAssociation(assocID);

}

void communicationErrorNotif(unsigned int assocID, unsigned short status, void* dummy){

  if (verbose) {  
    fprintf(stdout, "%-8x: Communication error (status %u)\n", assocID, status);
    fflush(stdout);
  }
}

void restartNotif(unsigned int assocID, void* ulpDataPtr){

    SCTP_AssociationStatus assocStatus;
    
    if (verbose) {  
        fprintf(stdout, "%-8x: Restart\n", assocID);
        fflush(stdout);
    }
    
    sctpd_getAssocStatus(assocID, &assocStatus);
    
    /* update ULP data */
    ((struct ulp_data *) ulpDataPtr)->maximumStreamID = assocStatus.outStreams - 1;
    ((struct ulp_data *) ulpDataPtr)->assocID = assocID;

}

void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr){

  if (verbose) {  
    fprintf(stdout, "%-8x: Shutdown complete\n", assocID);
    fflush(stdout);
  }
  /* free ULP data */
  ((struct ulp_data *) ulpDataPtr)->maximumStreamID = -1;
  //  sctpd_deleteAssociation(assocID);

}


int main(int argc, char **argv){

  UDP_UA_CB echoUlp;
  SCTP_InstanceParameters instanceParameters;
  unsigned int index;

  /* Trap Ctrl-C */
  signal(SIGINT, cleanUpAndExit);
    
  /* initialize ULP data */
  for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
    ulpData[index].maximumStreamID    = -1;
    ulpData[index].nrOfReceivedChunks = 0;
    ulpData[index].nrOfReceivedBytes  = 0;        
  }

  /* initialize the echo_ulp variable */
  echoUlp.dataArriveNotif          = &dataArriveNotif;
  echoUlp.sendFailureNotif         = &sendFailureNotif;
  echoUlp.networkStatusChangeNotif = &networkStatusChangeNotif;
  echoUlp.communicationUpNotif     = &communicationUpNotif;
  echoUlp.communicationLostNotif   = &communicationLostNotif;
  echoUlp.communicationErrorNotif  = &communicationErrorNotif;
  echoUlp.restartNotif             = &restartNotif;
  echoUlp.shutdownCompleteNotif    = &shutdownCompleteNotif;

  getArgs(argc, argv);
  checkArgs();


  /* Initialize sctplib */
  if (sctpd_initLibrary(UDP_Notif_Port,UDP_Resp_Port)==1) {
    
    if ((sctpInstance=sctpd_registerInstance(ECHO_PORT,
					     MAXIMUM_NUMBER_OF_IN_STREAMS, 
					     MAXIMUM_NUMBER_OF_OUT_STREAMS,
					     noOfLocalAddresses, localAddressList,
					     echoUlp)) != 0) {

      /* set the TOS field */                  
      sctpd_getAssocDefaults(sctpInstance, &instanceParameters);
      instanceParameters.ipTos=tosByte;
      sctpd_setAssocDefaults(sctpInstance, &instanceParameters);
      
      fprintf(stdout,"Test Tool : echo started \n");
      fflush(stdout);

      sctpd_eventLoop();

    }
    
  }
    
  /* this will never be reached */
  exit(0);

}










