/*
 * terminal_monitor.c, modified from terminal.c.
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Daemon-based terminal program with GUI and Status monitoring capabilities
 * Author : Eric Yeo Siemens ICM N PG U SE D2
 */



#include "../sctpdlib/ua.h"
#include <stdio.h>
#include <stdlib.h>         /* for atoi() under Linux */
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>

#include <glib.h>
#include <gtk/gtk.h>

/* DEFINITION OF CONSTANTS --------------------------------------------------*/
#define DEFAULT_X_SIZE    600
#define DEFAULT_Y_SIZE    800

#define SCTP_App_Port     0   

pid_t echo_tool_pid;

#define MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES    10
#define MAXIMUM_NUMBER_OF_IN_STREAMS         10
#define MAXIMUM_NUMBER_OF_OUT_STREAMS        10
#define SCTP_GENERIC_PAYLOAD_PROTOCOL_ID      0
#define min(x,y)            (x)<(y)?(x):(y)

/* SCTP VARIABLES -----------------------------------------------------------*/
static unsigned short sctpInstance;
static unsigned char localAddressList[MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES][SCTP_MAX_IP_LEN];
static unsigned short noOfLocalAddresses = 0;
static unsigned short unknownCommand = 0;
static unsigned int curAssocID = 0;
static unsigned int currInstreamID = 0, currOutstreamID = 0;


/* NETWORK VARIABLES --------------------------------------------------------*/
static int udpfd;                    /* file descriptor to the UDP */
static unsigned short connected = 0, timer_id;
static unsigned int timerStarted = 0, deltaT = 0;
static int UDP_Notif_Port = 7234;
static int UDP_Resp_Port  = 7235;

/* GTK VARIABLES ------------------------------------------------------------*/
GtkWidget *mainWindow;
GtkWidget *assocListBox;
GtkWidget *assocDetailsTree;
GtkWidget *detailsNode;
GtkWidget *assocstatusWindow;
GtkWidget *mesgstatusWindow;
GtkWidget *statusBar;
GtkWidget *textbox;
GtkWidget *eventsWindow;
GtkWidget *connectDialogWin;
GtkWidget *disconnectDialogWin;
GtkWidget *ipEntry, *portEntry;
GtkWidget *textEntry, *periodEntry;
GtkWidget *InstreamNode=NULL;
GtkWidget *OutstreamNode=NULL;
GtkWidget *statenode=NULL;
GtkWidget *pathsNode=NULL;
GtkWidget *detailsSubtree=NULL;

/* The Monitor tool structure stores infomation that cannot be retrieved from the 
   SCTP library, but may be of interest to the user */
typedef struct SCTP_Monitor_ToolStatus{
     
  unsigned char remoteAddr[SCTP_MAX_IP_LEN];
  unsigned short noOfInStreams;
  unsigned short noOfOutStreams;
  short InstreamIDList[MAXIMUM_NUMBER_OF_IN_STREAMS];
  short OutstreamIDList[MAXIMUM_NUMBER_OF_OUT_STREAMS];

} SCTP_MonitorToolStatus;

SCTP_MonitorToolStatus statusUpdate;

/*---------------------------- Arguments Functions -------------------------- */
void printUsage(void) {

  printf("Usage: terminal_monitor [options] -s source_addr_1 -s source_addr_2 ...\n");
  printf("Options : \n");
  printf("r     : response port\n");
  printf("a     : app port\n");
  
}

void getArgs(int argc, char **argv) {

  int c;
  extern char *optarg;
  extern int optind;

  while ((c = getopt(argc, argv, "s:")) != -1){
    switch (c) {
    case 's':
      if ((noOfLocalAddresses < MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES) &&
	  (strlen(optarg) < SCTP_MAX_IP_LEN  )) {
	strcpy(localAddressList[noOfLocalAddresses], optarg);
	noOfLocalAddresses++;
      }
      break;
    case 'a':
      UDP_Notif_Port = atoi(optarg);
      break;
    case 'r':
      UDP_Resp_Port = atoi(optarg);
      break;
    default:
      unknownCommand = 1;
      break;
    }
  }
}

void checkArgs(void) {

  int abortProgram;
  int printUsageInfo;
    
  abortProgram = 0;
  printUsageInfo = 0;
  
  if (noOfLocalAddresses == 0) {
    printf("Error:   At least one sourceaddress must be specified.\n");
    abortProgram = 1;
    printUsageInfo = 1;
  }
  if (unknownCommand ==1) {
    printf("Error:   Unkown options in command.\n");
    printUsageInfo = 1;
  }
    
  if (printUsageInfo == 1)
    printUsage();
  if (abortProgram == 1)
    exit(-1);
}


/* Define functions to handle the GUI ----------------------------------------*/

void displayStatus(char *status){
  
  gtk_container_add(GTK_CONTAINER(mesgstatusWindow),
		    gtk_list_item_new_with_label(status));
  gtk_widget_show_all(mesgstatusWindow);

}

void displaySCTPevents(char *status){

  gtk_container_add(GTK_CONTAINER(eventsWindow),gtk_list_item_new_with_label(status));
  gtk_widget_show_all(eventsWindow);

}


/* Functions to Update the Display--------------------------------------------*/

void DisplayAssocDetails(int assocID) {

  int i;
  unsigned short pathID;
  gchar tstr[100];              /* temp variable for formatting output */
 
  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;

  GtkWidget *tnode;
  GtkWidget *pathsSubtree;
  GtkWidget *pathNode;
  GtkWidget *pathSubtree;
  GtkWidget *InstreamSubtree; 
  GtkWidget *OutstreamSubtree;
 
  if (connected) {
    if (sctpd_getAssocStatus(assocID,&assocStatus) == 0){
      displayStatus("Received Association Status information");  

      /* Clear the current displayed tree */
      gtk_tree_clear_items(GTK_TREE(assocDetailsTree),0,1);
      
      /* create details node */
      sprintf(tstr,"Details (Assoc ID=%d):",assocID);
      detailsNode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(assocDetailsTree),detailsNode);
      gtk_widget_show(detailsNode);
      
      /* create details sub tree */
      detailsSubtree = gtk_tree_new();
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(detailsNode),detailsSubtree);
      
      /* Go through the MonitorTool Status structure and display all info
	 sctpInstance ID */
      sprintf(tstr,"SCTP Instance ID : %d",sctpInstance);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode); 
      
      /* No of Local Addresses */
      sprintf(tstr,"No. of local Addresses : %d",noOfLocalAddresses);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
      
      /* Source IP addresses */
      for (i=1; i <= noOfLocalAddresses; i++){
	sprintf(tstr,"Local IP Address %d : %s",i,localAddressList[i-1]);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
	gtk_widget_show(tnode);
      }
      
      /* Source Port */
      sprintf(tstr,"Local port : %d",assocStatus.sourcePort);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* Remote Port */
      sprintf(tstr,"Remote port : %d",assocStatus.destPort);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
	
      /* Go through the details of the Association Status to display */
      /* state */
      switch (assocStatus.state) {
      case 0:
	strcpy(tstr,"State : SCTP CLOSED");
	break;
      case 1:
	strcpy(tstr,"State : SCTP COOKIE WAIT");
	break;
      case 2:
	strcpy(tstr,"State : SCTP COOKIE ECHOED");
	break;
      case 3:
	strcpy(tstr,"State : SCTP ASSOCIATION ESTABLISHED");
	break;
      case 4:
	strcpy(tstr,"State : SCTP SHUTDOWN PENDING");
	break;
      case 5:
	strcpy(tstr,"State : SCTP SHUTDOWN RECEIVED");
	break;
      case 6:
	strcpy(tstr,"State : SCTP SHUTDOWN SENT");
	break;
      case 7:
	strcpy(tstr,"State : SCTP SHUTDOWN ACK SENT");
	break;	
      }
      
      statenode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),statenode);
      gtk_widget_show(statenode);
	
      /* Number of Addresses */
      sprintf(tstr,"No. Of Destination Addresses : %d",
	      assocStatus.numberOfAddresses);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* Primary Destination Address */
      sprintf(tstr,"Primary Destination Address : %s",
	      assocStatus.primaryDestinationAddress);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* OutStreams */
      sprintf(tstr,"No. of Outgoing streams : %d",assocStatus.outStreams);
      OutstreamNode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),OutstreamNode);
      gtk_widget_show(OutstreamNode);
	
      /* Create outStreams details sub tree */
      OutstreamSubtree = gtk_tree_new();
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(OutstreamNode),OutstreamSubtree);
	
      /* Current OutStream ID */
      switch (statusUpdate.OutstreamIDList[currOutstreamID]) {
      case -1:
	strcpy(tstr,"No details on current Outstream ID - no data sent.");
	break;
      }
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(OutstreamSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* InStreams */
      sprintf(tstr,"No. Of Incoming Streams : %d",assocStatus.inStreams);
      InstreamNode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),InstreamNode);
      gtk_widget_show(InstreamNode);
	
      /* Create InStreams details sub tree */
      InstreamSubtree = gtk_tree_new();
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(InstreamNode),InstreamSubtree);
	
      /* Current InStream ID */
      switch (statusUpdate.InstreamIDList[currInstreamID]) {
      case -1:
	strcpy(tstr,"No details on current Instream ID - no data received.");
	break;
      }
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(InstreamSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* primaryAddressIndex */
      sprintf(tstr,"Primary Address Index (Primary Path ID): %d"
	      ,assocStatus.primaryAddressIndex);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* currentReceiverWindowSize */
      sprintf(tstr,"Current Receiver Window Size : %d",
	      assocStatus.currentReceiverWindowSize);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* outstandingBytes */
      sprintf(tstr,"Outstanding bytes : %d",assocStatus.outstandingBytes);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* noOfChunksInSendQueue */
      sprintf(tstr,"No. of chunks in send queue : %d",
	      assocStatus.noOfChunksInSendQueue);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* noOfChunksInRetransmissionQueue */
      sprintf(tstr,"No. of chunks in retransmission queue : %d",
	      assocStatus.noOfChunksInRetransmissionQueue);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* noOfChunksInReceptionQueue */
      sprintf(tstr,"No. of chunks in reception queue : %d",
	      assocStatus.noOfChunksInReceptionQueue);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* rtoInitial */
      sprintf(tstr,"Initial Round Trip Timeout : %d (msecs)",
	      assocStatus.rtoInitial);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* rtoMin */
      sprintf(tstr,"Minimum Round Trip Timeout : %d (msecs)",assocStatus.rtoMin);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* rtoMax */
      sprintf(tstr,"Maximum Round Trip Timeout : %d (msecs)",assocStatus.rtoMax);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* validCookieLife */
      sprintf(tstr,"Cookie Lifetime : %d (msecs)",assocStatus.validCookieLife);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* assocMaxRetransmits */
      sprintf(tstr,"Maximum retransmission per association : %d",
	      assocStatus.assocMaxRetransmits);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* pathMaxRetransmits */
      sprintf(tstr,"Maximum retransmission per path : %d",
	      assocStatus.pathMaxRetransmits);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* maxInitRetransmits */
      sprintf(tstr,"Maximum Initial Retransmission : %d",
	      assocStatus.maxInitRetransmits);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);
	
      /* myRwnd */
      sprintf(tstr,"Local Receiver Window : %d",assocStatus.myRwnd);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

      /* Delay for delay acknowledge */
      sprintf(tstr,"Delay for delay acknowledge : %d (msecs)",assocStatus.delay);
      tnode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),tnode);
      gtk_widget_show(tnode);

	
      /* PATHS INFO -----------------------------------------------------------*/
      
      /* create paths node */
      strcpy(tstr,"Paths");
      pathsNode = gtk_tree_item_new_with_label(tstr);
      gtk_tree_append(GTK_TREE(detailsSubtree),pathsNode);
      gtk_widget_show(pathsNode);
	
      /* create paths sub tree */
      pathsSubtree = gtk_tree_new();
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathsNode),pathsSubtree);

      for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
	/* get the pathStatus */
	if (sctpd_getPathStatus(assocID, pathID, &pathStatus)==0){
	  
	  /* Check if the current path is the primary path */		     
	  if (pathID == sctpd_getPrimary(assocID)){
	    sprintf(tstr,"Primary path ID=%d",pathID);
	    pathNode = gtk_tree_item_new_with_label(tstr);
	    gtk_tree_append(GTK_TREE(pathsSubtree),pathNode);
	    gtk_widget_show(pathNode);
	  }
	  
	  else {  
	    sprintf(tstr,"Path ID=%d",pathID);
	    pathNode = gtk_tree_item_new_with_label(tstr);
	    gtk_tree_append(GTK_TREE(pathsSubtree),pathNode);
	    gtk_widget_show(pathNode);
	    }
	  
	  /* create path sub tree */
	  pathSubtree = gtk_tree_new();
	  gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathNode),pathSubtree);
	
	  /* destination address */
	  sprintf(tstr,"Destination Address : %s",pathStatus.destinationAddress);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	  
	  /* state */
	  switch (pathStatus.state) {
	  case 0:
	    strcpy(tstr,"State : ACTIVE");
	    break;
	  case 1:
	    strcpy(tstr,"State : INACTIVE");
	    break;
	  default:
	    strcpy(tstr,"State : UNKNOWN");
	    break;
	  }
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  /* srtt */
	  sprintf(tstr,"Smooth Round Trip Time : %d (msecs)",pathStatus.srtt);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  /* rto */
	  sprintf(tstr,"Round Trip Time : %d (msecs)",pathStatus.rto);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	  
	  /* rttvar */
	  sprintf(tstr,"Round Trip Time Variation : %d (msecs)",pathStatus.rttvar);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	  
	  /* heartbeatInterval */
	  sprintf(tstr,"Heartbeat Interval : %d",pathStatus.heartbeatIntervall);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	  
	  /* cwnd */
	  sprintf(tstr,"Congestion Window Size : %d",pathStatus.cwnd);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  /* cwnd2 */
	  sprintf(tstr,"Congestion Window size 2 : %d",pathStatus.cwnd2);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	  
	  /* Slow start Threshold */
	  sprintf(tstr,"Slow start threshold : %d",pathStatus.ssthresh);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  /* MTU */
	  sprintf(tstr,"MTU : %d",pathStatus.mtu);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  /* ipTos */
	  sprintf(tstr,"IP Type Of Service : %x",pathStatus.ipTos);
	  tnode = gtk_tree_item_new_with_label(tstr);
	  gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	  gtk_widget_show(tnode);
	
	  gtk_tree_item_expand(GTK_TREE_ITEM(pathNode));  
	  
	}
      }
    }
    else 
      displayStatus("Receive Association Status Information failed!");
  }
  else  
    displayStatus("There is no connection!");
  
  gtk_widget_show_all(pathsNode);
  gtk_tree_item_expand(GTK_TREE_ITEM(detailsNode));
  gtk_tree_item_expand(GTK_TREE_ITEM(pathsNode));
  
}

void UpdatePathDetails (unsigned int assocID){

  GtkWidget *pathNode;
  GtkWidget *pathSubtree;
  GtkWidget *pathsSubtree;
  GtkWidget *tnode;

  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;
  unsigned short pathID;
  gchar statusInfo[100];        /* temp variable for formatting output */
  gchar tstr[50];              /* temp variable for formatting output */

  /* Delete paths sub tree */
  gtk_tree_item_remove_subtree(GTK_TREE_ITEM(pathsNode));  

  /* create paths sub tree */
  pathsSubtree = gtk_tree_new();
  gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathsNode),pathsSubtree);
  
  if (sctpd_getAssocStatus(assocID, &assocStatus)==0) {
  
    for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
      /* get the pathStatus */
      if (sctpd_getPathStatus(assocID, pathID, &pathStatus)==0) {
        	
	/* state */
	switch (pathStatus.state) {
	case 0:
	  strcpy(tstr,"(ACTIVE)");
	  break;
	case 1:
	  strcpy(tstr,"(INACTIVE)");
	  break;
	default:
	  strcpy(tstr,"(UNKNOWN)");
	  break;
	}

	/* Check if the current path is the primary path */
	if (pathID == sctpd_getPrimary(assocID)){
	  sprintf(statusInfo,"Primary path ID=%d ",pathID);
	  strcat(statusInfo,tstr);
	  pathNode = gtk_tree_item_new_with_label(statusInfo);
	  gtk_tree_append(GTK_TREE(pathsSubtree),pathNode);
	  gtk_widget_show(pathNode);
	}
	
	else {
	  sprintf(statusInfo,"Path ID=%d ",pathID);
	  strcat(statusInfo,tstr);
	  pathNode = gtk_tree_item_new_with_label(statusInfo);
	  gtk_tree_append(GTK_TREE(pathsSubtree),pathNode);
	  gtk_widget_show(pathNode);
	}
      
	/* create path sub tree */
	pathSubtree = gtk_tree_new();
	gtk_tree_item_set_subtree(GTK_TREE_ITEM(pathNode),pathSubtree);
      
	/* destination address */
	sprintf(tstr,"Destination Address : %s",pathStatus.destinationAddress);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
        
	/* srtt */
	sprintf(tstr,"Smooth Round Trip Time : %d (msecs)",pathStatus.srtt);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
	
	/* rto */
	sprintf(tstr,"Round Trip Time : %d (msecs)",pathStatus.rto);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* rttvar */
	sprintf(tstr,"Round Trip Time Variation : %d (msecs)",pathStatus.rttvar);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* heartbeatInterval */
	sprintf(tstr,"Heartbeat Interval : %d (msecs)",pathStatus.heartbeatIntervall);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* cwnd */
	sprintf(tstr,"Congestion Window Size : %d",pathStatus.cwnd);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* cwnd2 */
	sprintf(tstr,"Congestion Window size 2 : %d",pathStatus.cwnd2);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* Slow start Threshold */
	sprintf(tstr,"Slow start threshold : %d",pathStatus.ssthresh);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
    
	/* MTU */
	sprintf(tstr,"MTU : %d",pathStatus.mtu);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);

	/* ipTos */
	sprintf(tstr,"IP Type Of Service : %x",pathStatus.ipTos);
	tnode = gtk_tree_item_new_with_label(tstr);
	gtk_tree_append(GTK_TREE(pathSubtree),tnode);
	gtk_widget_show(tnode);
      
	gtk_tree_item_expand(GTK_TREE_ITEM(pathNode));  
      }
    }
  } 
  gtk_tree_item_expand(GTK_TREE_ITEM(pathsNode));  
}


void UpdateStateInfo(unsigned int assocID){

  SCTP_AssociationStatus assocStatus;
  gchar tstr[50];              /* temp variable for formatting output */
  gint position = 0;

  /* Delete state of association infomation from tree */
  gtk_tree_remove_item(GTK_TREE(detailsSubtree), statenode);
  
  if (connected){
    if (sctpd_getAssocStatus(curAssocID,&assocStatus)==0){ 
      /* Update state of association */
      switch (assocStatus.state) {
      case 0:
	strcpy(tstr,"State : SCTP CLOSED");
	break;
      case 1:
	strcpy(tstr,"State : SCTP COOKIE WAIT");
	break;
      case 2:
	strcpy(tstr,"State : SCTP COOKIE ECHOED");
	break;
      case 3:
	strcpy(tstr,"State : SCTP ASSOCIATION ESTABLISHED");
	break;
      case 4:
	strcpy(tstr,"State : SCTP SHUTDOWN PENDING");
	break;
      case 5:
	strcpy(tstr,"State : SCTP SHUTDOWN RECEIVED");
	break;
      case 6:
	strcpy(tstr,"State : SCTP SHUTDOWN SENT");
	break;
      case 7:
	strcpy(tstr,"State : SCTP SHUTDOWN ACK SENT");
	break;	
      }
    }
  }
  else
    strcpy(tstr,"ASSOCIATION NOT ESTABLISHED!");
  
  statenode = gtk_tree_item_new_with_label(tstr);
  gtk_tree_insert(GTK_TREE(detailsSubtree),statenode, position);
  gtk_widget_show(statenode);	  
} 
  
void UpdateCurInstreamID(){
 
  GtkWidget *treenode;
  GtkWidget *InstreamSubtree;
  char statusInfo[256];        /* temp variable for formatting output */

  if (InstreamNode!=NULL) {
    /* Delete InStreams details sub tree */
    gtk_tree_item_remove_subtree(GTK_TREE_ITEM(InstreamNode));
  
    /* Create the new InStreams details sub tree */
    InstreamSubtree = gtk_tree_new();
    gtk_tree_item_set_subtree(GTK_TREE_ITEM(InstreamNode),InstreamSubtree);
  
    /* Current InStream ID */
    sprintf(statusInfo,"Current Incoming stream ID : %d",
	    statusUpdate.InstreamIDList[currInstreamID]);
    treenode = gtk_tree_item_new_with_label(statusInfo);
    gtk_tree_append(GTK_TREE(InstreamSubtree),treenode);
    gtk_widget_show(treenode);
  }
}

void UpdateCurOutstreamID(){

  GtkWidget *treenode;
  GtkWidget *OutstreamSubtree;  
  char tstr[50];

  if (OutstreamNode!=NULL) {
    /* Delete outStreams details sub tree */
    gtk_tree_item_remove_subtree(GTK_TREE_ITEM(OutstreamNode));
  
    /* Create the new outStreams details sub tree */
    OutstreamSubtree = gtk_tree_new();
    gtk_tree_item_set_subtree(GTK_TREE_ITEM(OutstreamNode),OutstreamSubtree);
  
    /* Current OutStream ID */
    sprintf(tstr,"Current Outgoing stream ID : %d",
	    statusUpdate.OutstreamIDList[currOutstreamID]);
    treenode = gtk_tree_item_new_with_label(tstr);
    gtk_tree_append(GTK_TREE(OutstreamSubtree),treenode);
    gtk_widget_show(treenode);
  }     
}

/* Timer function for periodic refresh of data */
gint timeout_callback(gpointer data) {

  UpdatePathDetails(curAssocID);
  displayStatus("Path details updated");      
  
  return 1;
}


/* EXIT FUNCTION--------------------------------------------------------------*/

void cleanUpanddExit(){
    
  /* calls sctpd_shutdown() to close the association.
     waits for the receipt of the shutdown notification before terminating */

  if (connected){
    if (sctpd_shutdown(curAssocID)==0)
      displayStatus("Terminating Monitoring tool successful");

    else
      displayStatus("Terminating Monitoring tool failed");

    /* Unregister Instance with sctpd */
    sctpd_unregisterInstance(sctpInstance);
      
    /* Stop sctpdlib */
    sctpd_stopLibrary();
      
    gtk_main_quit();
  }

  else {
    /* Unregister Instance with sctpd */
    sctpd_unregisterInstance(sctpInstance);
      
    /* Stop sctpdlib */
    sctpd_stopLibrary();
    
    gtk_main_quit();
  }
}

/* Functions for handling menu items events-----------------------------------*/

void UpdateAll(GtkMenuItem * menuitem, gpointer data){

  if (connected){
    DisplayAssocDetails(curAssocID);
    UpdateCurInstreamID();
    UpdateCurOutstreamID();
    displayStatus("Association and Path details updated");      
  }
  else 
    displayStatus("There is no connection!");      
}

void UpdatePath(GtkMenuItem * menuitem, gpointer data){

  if (connected){
    UpdatePathDetails(curAssocID);
    displayStatus("Path details updated");      
  }
  else 
    displayStatus("There is no connection!"); 
}

/* Functions for handling timer events----------------------------------------*/

void startTimer (GtkMenuItem * menuitem, gpointer data){

  char tstr[50];

  if (connected){
    if (!timerStarted){
      /* Start Timer for periodic refresh */
      timer_id = gtk_timeout_add(deltaT,timeout_callback,NULL);
      timerStarted = 1;
      sprintf(tstr,"Timer Started, period = %d", deltaT);
      displayStatus(tstr);
    }
    else 
      displayStatus("Timer already started!");
  }
  else 
    displayStatus("There is no connection!"); 
}

void stopTimer (GtkMenuItem * menuitem, gpointer data){

  if (connected){
    if (timerStarted){
      /* Stop Timer */
      gtk_timeout_remove(timer_id);
      timerStarted = 0;
      displayStatus("Timer Stopped");
    }
    else 
      displayStatus("Timer not started!");
  }
  else 
    displayStatus("There is no connection!"); 
}

/* Functions for handling button clicks events--------------------------------*/

void Sendtext( GtkWidget *widget, GtkWidget *entry ){

  gchar *text;
  int result;
  static unsigned short outstreamID = 0;

  if (connected) {   
    /* Extract the text from the Entry widget */
    text = gtk_entry_get_text(GTK_ENTRY(entry));
            
    /* if text == NULL add some spaces so that strlen(text) != 0) */
    if (strlen(text) == 0)
      sprintf(text, " ");
      
    /*Increment the Stream ID by 1 everytime data is sent, 
      if stream ID = largest stream ID assigned, reset stream ID to 0 */      
    if (outstreamID == statusUpdate.noOfOutStreams)
      outstreamID = 0;
       
    /* Call the daemon to send */
    result = sctpd_send(curAssocID, outstreamID, text, strlen(text),
			SCTP_GENERIC_PAYLOAD_PROTOCOL_ID,
			SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, 
			SCTP_INFINITE_LIFETIME, SCTP_ORDERED_DELIVERY, 
			SCTP_BUNDLING_DISABLED);
      
    /* Update the Outstream ID list in Monitor tool structure */
    if (statusUpdate.OutstreamIDList[outstreamID] == -1)
      statusUpdate.OutstreamIDList[outstreamID] = outstreamID; 
      
    /* Update the current position of OutstreamIDList array */
    currOutstreamID = outstreamID;

    /* Update the display */
    UpdateCurOutstreamID();
    outstreamID ++;
      
    /* Clear the entry */
    gtk_entry_set_text (GTK_ENTRY(entry), "");

  }
  else
    displayStatus("There is no connection");

}

void Cleartext (GtkButton *button,  GtkWidget *entry){ 

  gtk_entry_set_text (GTK_ENTRY(entry), "");

}

/* Code canabalize from book -----------------------------------------------*/ 
/*
 * CloseDialog
 *
 * Close the dialog window.  The dialog handle is passed
 * in as the data.
 */
void CloseDialog (GtkWidget *widget, gpointer data){

    /* --- Close it --- */
    gtk_widget_destroy (GTK_WIDGET (data));

}

/*
 * ClosingDialog
 *
 * Calls when window is about to close.  Returns FALSE
 * to let it close.  Releases the grab which made the dialog
 * window modal.
 */
void ClosingDialog (GtkWidget *widget, gpointer data){

    gtk_grab_remove (GTK_WIDGET (widget));
}


void connectToPeer (GtkWidget *window, gpointer data) {

  char tstr[255];
  
  if (!connected)
    {
      sprintf(tstr,"Connecting to peer with IP address = %s at port = %s",
	      gtk_entry_get_text(GTK_ENTRY(ipEntry)),
	      gtk_entry_get_text(GTK_ENTRY(portEntry)));
      displayStatus(tstr);

      /* Update the MonitorTool Status Structure */
      strcpy(statusUpdate.remoteAddr, gtk_entry_get_text(GTK_ENTRY(ipEntry)) );	
  
      /* Get value of deltaT from entry box */
      deltaT = atoi(gtk_entry_get_text(GTK_ENTRY(periodEntry)));

      /* Call daemon to send */
      sctpd_associate(sctpInstance, MAXIMUM_NUMBER_OF_OUT_STREAMS, 
			  gtk_entry_get_text(GTK_ENTRY(ipEntry)), 
			  atoi(gtk_entry_get_text(GTK_ENTRY(portEntry))), 
			  NULL);

      /* Remove the dialog box */
      gtk_grab_remove (GTK_WIDGET (data));
      gtk_widget_destroy(GTK_WIDGET (data));
    }

  else 
    displayStatus("Already Connected!");

}


void disconnectFromPeer (GtkWidget *window, gpointer data) {

  if (connected && timerStarted ) {
    gtk_timeout_remove(timer_id);
    timerStarted = 0;
    if (sctpd_shutdown(curAssocID) == 0){
      displayStatus("Disconnected from peer");
    }
    else{
      displayStatus("Unable to disconect from peer!");
      UpdateStateInfo(curAssocID);
      UpdatePathDetails(curAssocID);
    }
  }
  else if (connected && !timerStarted){
    if (sctpd_shutdown(curAssocID) == 0){
      displayStatus("Disconnected from peer");
    }
    else{
      displayStatus("Unable to disconect from peer!");
      UpdateStateInfo(curAssocID);
      UpdatePathDetails(curAssocID);
    }
  }
    
  else if (!connected){
    displayStatus("Not Connected to peer!");
    UpdateStateInfo(curAssocID);
  }  

  /* Remove the dialog box */
  gtk_grab_remove (GTK_WIDGET (data));
  gtk_widget_destroy(GTK_WIDGET (data));
}

/*
 * connectDialog
 *
 * Display a connection Dialog
 */
void  connectDialog (GtkMenuItem * menuitem, gpointer data) {

  GtkWidget *label;
  GtkWidget *connectBtn, *cancelBtn;
  GtkWidget *table, *hbox, *vbox;
    
  /* --- Create a dialog window --- */
  connectDialogWin = gtk_dialog_new ();

  /* --- Trap the window close signal to release the grab --- */
  gtk_signal_connect (GTK_OBJECT (connectDialogWin), "destroy",
		      GTK_SIGNAL_FUNC (ClosingDialog),
		      &connectDialogWin);

  /* --- Add a title to the window --- */
  gtk_window_set_title (GTK_WINDOW (connectDialogWin), "Connect to:");

  /* --- Create a small border --- */
  gtk_container_border_width (GTK_CONTAINER (connectDialogWin), 5);

  /* --- Create the message in a label --- */
  label = gtk_label_new ("Enter the destination information");

  /* --- Add the label to the dialog box --- */
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (connectDialogWin)->vbox),
		      label, TRUE, TRUE, 0);

  /* Create the vbox */
  vbox = gtk_vbox_new(FALSE,8);

  /* Packing table to hold text entries and their labels */
  table = gtk_table_new(3,2,FALSE);

  /* IP Address */
  label = gtk_label_new("IP Address: ");
  gtk_table_attach(GTK_TABLE(table),label,0,1,0,1,GTK_FILL,GTK_FILL,0,0);
  ipEntry = gtk_entry_new_with_max_length ( 15 * sizeof(gchar));
  gtk_entry_set_text (GTK_ENTRY (ipEntry), "192.168.1.");
  gtk_entry_select_region (GTK_ENTRY (ipEntry),0 , 
			   GTK_ENTRY(ipEntry)->text_length);
  gtk_table_attach(GTK_TABLE(table),ipEntry,1,2,0,1,GTK_FILL,GTK_FILL,0,0);

  /* Port Number */
  label = gtk_label_new("Port:");
  gtk_table_attach(GTK_TABLE(table),label,0,1,1,2,GTK_FILL,GTK_FILL,0,0);
  portEntry = gtk_entry_new_with_max_length(5 * sizeof(gchar));
  gtk_entry_set_text (GTK_ENTRY (portEntry), "7");
  gtk_entry_select_region (GTK_ENTRY (portEntry),0 , 
			   GTK_ENTRY(portEntry)->text_length);
  gtk_table_attach(GTK_TABLE(table),portEntry,1,2,1,2,GTK_FILL,GTK_FILL,0,0);

  /* set Timer period */
  label = gtk_label_new("Period (msecs):");
  gtk_table_attach(GTK_TABLE(table),label,0,1,2,3,GTK_FILL,GTK_FILL,0,0);
  periodEntry = gtk_entry_new_with_max_length(6 * sizeof(gchar));
  gtk_entry_set_text (GTK_ENTRY (periodEntry), "2000");
  gtk_entry_select_region (GTK_ENTRY (periodEntry),0 , 
			   GTK_ENTRY(periodEntry)->text_length);
  gtk_table_attach(GTK_TABLE(table),periodEntry,1,2,2,3,GTK_FILL,GTK_FILL,0,0);

  /* Add the table to the vbox */
  gtk_box_pack_start (GTK_BOX (vbox),
		      table, TRUE, TRUE, 0);

  /* --- Create the "Connect" button --- */
  connectBtn = gtk_button_new_with_label ("Connect");

  gtk_signal_connect (GTK_OBJECT (connectBtn), "clicked",
		      GTK_SIGNAL_FUNC (connectToPeer),
		      connectDialogWin);

  /* --- Create the "Cancel" button --- */
  cancelBtn = gtk_button_new_with_label ("Cancel");

  /* --- Need to close the window if they press "cancel" --- */
  gtk_signal_connect (GTK_OBJECT (cancelBtn), "clicked",
		      GTK_SIGNAL_FUNC (CloseDialog),
		      connectDialogWin);


  /* --- Add the buttons to the hbox --- */
  hbox = gtk_hbox_new(FALSE,5);
  gtk_box_pack_start (GTK_BOX (hbox),
		      connectBtn, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox),
		      cancelBtn, TRUE, TRUE, 0);

  /* Add the hbox to the vbox */
  gtk_box_pack_start (GTK_BOX (vbox),
		      hbox, TRUE, TRUE, 0);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (connectDialogWin)->action_area),
		      vbox, TRUE, TRUE, 0);
    
  /* --- Make the dialog box visible --- */
  gtk_widget_show_all (connectDialogWin);
  gtk_grab_add (connectDialogWin);

}

/*
 * disconnectDialog
 *
 * Display a disconnection Dialog
 */
void  disconnectDialog (GtkMenuItem * menuitem, gpointer data) {

  GtkWidget *label;
  GtkWidget *disconnectBtn, *cancelBtn;
  GtkWidget *hbox;
    
  /* --- Create a dialog window --- */
  disconnectDialogWin = gtk_dialog_new ();

  /* --- Trap the window close signal to release the grab --- */
  gtk_signal_connect (GTK_OBJECT (disconnectDialogWin), "destroy",
		      GTK_SIGNAL_FUNC (ClosingDialog),
		      &disconnectDialogWin);

  /* --- Add a title to the window --- */
  gtk_window_set_title (GTK_WINDOW (disconnectDialogWin), "Connect to:");

  /* --- Create a small border --- */
  gtk_container_border_width (GTK_CONTAINER (disconnectDialogWin), 5);

  /* --- Create the message in a label --- */
  label = gtk_label_new ("Are you sure you wish to disconnect?");

  /* --- Add the label to the dialog box --- */
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (disconnectDialogWin)->vbox),
		      label, TRUE, TRUE, 0);


  /* --- Create the "Disconnect" button --- */
  disconnectBtn = gtk_button_new_with_label ("Disconnect");
  
  gtk_signal_connect (GTK_OBJECT (disconnectBtn), "clicked",
		      GTK_SIGNAL_FUNC (disconnectFromPeer),
		      disconnectDialogWin);


  /* --- Create the "Cancel" button --- */
  cancelBtn = gtk_button_new_with_label ("Cancel");

  /* --- Need to close the window if they press "cancel" --- */
  gtk_signal_connect (GTK_OBJECT (cancelBtn), "clicked",
		      GTK_SIGNAL_FUNC (CloseDialog),
		      disconnectDialogWin);


  /* --- Add the buttons to the hbox --- */
  hbox = gtk_hbox_new(FALSE,5);
  gtk_box_pack_start (GTK_BOX (hbox),
		      disconnectBtn, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox),
		      cancelBtn, TRUE, TRUE, 0);


  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (disconnectDialogWin)->action_area),
		      hbox, TRUE, TRUE, 0);
    
  /* --- Make the dialog box visible --- */
  gtk_widget_show_all (disconnectDialogWin);
  gtk_grab_add (disconnectDialogWin);

}

void menuExitHandler(GtkMenuItem * menuitem,gpointer data) {

  cleanUpanddExit();

}

/*
 * Popup
 *
 * Display a popup dialog window with a message and an
 * "Ok" button
 */
void Popup (char *szMessage) {

  static GtkWidget *label;
  GtkWidget *button;
  GtkWidget *dialog_window;
  
  /* --- Create a dialog window --- */
  dialog_window = gtk_dialog_new ();

  /* --- Trap the window close signal to release the grab --- */
  gtk_signal_connect (GTK_OBJECT (dialog_window), "destroy",
		      GTK_SIGNAL_FUNC (ClosingDialog),
		      &dialog_window);

  /* --- Add a title to the window --- */
  gtk_window_set_title (GTK_WINDOW (dialog_window), "SCTP EVENTS INITIATED!");

  /* --- Create a small border --- */
  gtk_container_border_width (GTK_CONTAINER (dialog_window), 5);

  /* -------------------------- */
  /* --- Create the message --- */
  /* -------------------------- */

  /* --- Create the message in a label --- */
  label = gtk_label_new (szMessage);

  /* --- Put some room around the label --- */
  gtk_misc_set_padding (GTK_MISC (label), 10, 10);

  /* --- Add the label to the dialog box --- */
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog_window)->vbox),
		      label, TRUE, TRUE, 0);

  /* --- Make the label visible --- */
  gtk_widget_show (label);

  /* ------------------- */
  /* --- "Ok" button --- */
  /* ------------------- */

  /* --- Create the "Ok" button --- */
  button = gtk_button_new_with_label ("Ok");

  /* --- Need to close the window if they press "Ok" --- */
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (CloseDialog),
		      dialog_window);

  /* --- Allow it to be the default button --- */
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  /* --- Add the button to the dialog box --- */
  gtk_box_pack_start (GTK_BOX (
			       GTK_DIALOG (dialog_window)->action_area),
		      button, TRUE, TRUE, 0);

  /* --- Make the button the default button --- */
  gtk_widget_grab_default (button);
  
  /* --- Make the button visible --- */
  gtk_widget_show (button);

  /* --- Make the dialog box visible --- */
  gtk_widget_show (dialog_window);

  gtk_grab_add (dialog_window);
}

/* ------------------------- End of canabalization ------------------------------ */

/* Function to update the monitor tool data structure */
void UpdateMonitorToolStruct(unsigned int assocID, unsigned short noOfInStreams, 
			     unsigned short noOfOutStreams) {

  int i;

  /* To fix current bug with assignment of no. of in and out streams by SCTP library */
  noOfOutStreams = min ( noOfOutStreams, noOfInStreams);

  /* Update the Monitor tool status structure */
  statusUpdate.noOfInStreams            = noOfInStreams;
  statusUpdate.noOfOutStreams           = noOfOutStreams;
  
  /* Initialise the Contents of InStreamList and OutStreamList to -1 */
  for (i=0; i < noOfInStreams; i++) {
    statusUpdate.InstreamIDList[i] = -1;
  }
  
  for (i=0; i < noOfOutStreams; i++) {
    statusUpdate.OutstreamIDList[i] = -1;
  }  
}

/* UDP Callback Handlers --------------------------------------------------------*/

void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,unsigned short noOfInStreams, 
			   unsigned short noOfOutStreams, void* dummy) {

  char statusInfo[256];        /* temp variable for formatting output */
  /*  SCTP_AssociationStatus assocStatus; */

  sprintf(statusInfo, "Association ID =%-8x: Communication up (%u paths)", 
	  assocID, noOfPaths);
  displaySCTPevents(statusInfo);

  connected = 1;
  curAssocID = assocID;

  /* Modify RTO Max to 10 secs so that network status change occurs faster 
  if (sctpd_getAssocStatus(assocID, &assocStatus)!=-1) {
    assocStatus.rtoMax = 10000;
    if (sctpd_setAssocStatus(assocID, &assocStatus)==0) {
      sctpd_getAssocStatus(assocID, &assocStatus);
      if (assocStatus.rtoMax == 10000) {
	sprintf(statusInfo,"Maximum Round Trip Timeout modified to %d (msecs)"
		,assocStatus.rtoMax);
	displayStatus(statusInfo);
      }
    }
    } */

  /* Update the Monitor tool data structures */
  UpdateMonitorToolStruct(assocID, noOfInStreams, noOfOutStreams);

  DisplayAssocDetails(assocID);

}
 
void dataArriveNotif(unsigned int assocID, unsigned int streamID, 
		     unsigned int len,unsigned int protoID,
		     unsigned int unordered, void* ulpDataPtr){

  char statusInfo[256];                 /* temp variable for formatting output */
  char tstr[1500];                      /* temp variable for formatting output */
  int length, i; 
  char chunk[SCTP_MAXIMUM_DATA_LENGTH + 1];
  unsigned int InstreamID_position = 0;

  sprintf(statusInfo,"Association ID =%-8x: Data arrived (%u bytes on stream %u,%s)",
	  assocID, len, streamID, 
	  (unordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered");

  displaySCTPevents(statusInfo);

  /* read it */
  length = SCTP_MAXIMUM_DATA_LENGTH;
  sctpd_receive(assocID, streamID, chunk, &length, SCTP_MSG_DEFAULT);
  chunk[length] = 0;   
  sprintf(tstr,"Received data: %s on stream %u", chunk , streamID);
  displayStatus(tstr);
  
  /* Stores the stream Id into the structure statusUpdate each time a data arrives */  
  for (i=0; i <statusUpdate.noOfInStreams; i++) {

    if (statusUpdate.InstreamIDList[i] == -1) {
      statusUpdate.InstreamIDList[i] = streamID;  
      InstreamID_position = i;
    }
    else if (statusUpdate.InstreamIDList[i] != -1 &&
	     statusUpdate.InstreamIDList[i] != streamID) {
      statusUpdate.InstreamIDList[i] = streamID;  
      InstreamID_position = i;
    }
    else if (statusUpdate.InstreamIDList[i] == streamID) {
      InstreamID_position = i;
    }
  }

  /* Update the current position of the InstreamIDList array */  
  currInstreamID = InstreamID_position;
  
  /* Update the display */
  UpdateCurInstreamID();

}

void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, 
			      unsigned short newState, void* ulpDataPtr) {

  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;
  unsigned short pathID;
  char statusInfo[256];        /* temp variable for formatting output */

  if (sctpd_getPathStatus(assocID, affectedPathID, &pathStatus)!=-1) {
    sprintf(statusInfo, "Association ID =%-8x: Network status change: path %u (towards %s) is now %s", 
	    assocID, affectedPathID, pathStatus.destinationAddress,
	    ((newState == SCTP_PATH_OK) ? "ACTIVE" : "INACTIVE"));
    
    displaySCTPevents(statusInfo);
    
    /* if the primary path has become inactive */
    if ((newState == SCTP_PATH_UNREACHABLE) &&
	(affectedPathID == sctpd_getPrimary(assocID))) {       
      /* select a new one */

      if (sctpd_getAssocStatus(assocID, &assocStatus)!=-1) {
	for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
	  sctpd_getPathStatus(assocID, pathID, &pathStatus);
	  if (pathStatus.state == SCTP_PATH_OK)
	    break;
	}
        
	/* and use it */
	if (pathID < assocStatus.numberOfAddresses) {
	  sctpd_setPrimary(assocID, pathID);
	  
	}
      }
    }
  }

  /* Update the display */
  UpdatePathDetails(assocID);

  /* Alert user */
  Popup("Network Status Changed");

}


void communicationLostNotif(unsigned int assocID, unsigned short status, void* ulpDataPtr) {

  char statusInfo[256];        /* temp variable for formatting output */

  sprintf(statusInfo,"Communication Lost for association (ID=%d)",assocID);
  displaySCTPevents(statusInfo);

  connected = 0;
  curAssocID = 0;

  /* Alert user */
  Popup("Communications Lost!");

  UpdateStateInfo(assocID);

}

void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr){
  
  displaySCTPevents("Shutdown completed message received!");
  
  connected=0;
  curAssocID = 0;

  UpdateStateInfo(assocID);

}

void communicationErrorNotif(unsigned int assocID, unsigned short status, void* dummy) {

  char statusInfo[256];        /* temp variable for formatting output */

  sprintf(statusInfo, "Association ID =%-8x: Communication error (status %u)", 
	  assocID, status);
  displaySCTPevents(statusInfo);

  UpdateStateInfo(assocID);
  UpdatePathDetails(assocID);  
}

void restartNotif(unsigned int assocID, void* ulpDataPtr) {

  char statusInfo[256];        /* temp variable for formatting output */

  sprintf(statusInfo, "Association ID =%-8x: Restart", assocID);
  displaySCTPevents(statusInfo);

  DisplayAssocDetails(assocID);
}

void sendFailureNotif(unsigned int assocID, unsigned char *unsent_data, 
		      unsigned int dataLength, unsigned int *context, void* dummy) {

  char statusInfo[256];        /* temp variable for formatting output */

  sprintf(statusInfo, "Association ID =%-8x: Send failure", assocID);
  displaySCTPevents(statusInfo);

  UpdateStateInfo(assocID);
  UpdatePathDetails(assocID);  
}


void udp_recv_callback (gpointer data,gint source,GdkInputCondition condition) {

  char recvbuff[256];
  socklen_t len;
  struct sockaddr_in taddr;
  ssize_t mesglen;

  mesglen = recvfrom(source, recvbuff, sizeof(recvbuff), 0,
		     (struct sockaddr *) &taddr, &len);

  if (mesglen>0) {
    /* call the dispatcher provided by sctpudputils which will
       call the registered UDP callback functions. */
    dispatchUAMesg(recvbuff);
  }
}


/* connect to the UDP port */
void initializeUDPListening(int port) {

  int sockflags;
  struct sockaddr_in uaddr;
  struct sockaddr_in paddr;

  udpfd = socket(AF_INET, SOCK_DGRAM,0);

  bzero (&uaddr,sizeof(uaddr));
  bzero (&paddr,sizeof(paddr));

  uaddr.sin_family = AF_INET;
  uaddr.sin_port = htons(port);
  uaddr.sin_addr.s_addr=htonl(INADDR_ANY);
  
  if ((bind(udpfd, (struct sockaddr *) &uaddr, sizeof(uaddr)))==0) {
    /* set socket as non-blocking */
    sockflags = fcntl (udpfd,F_GETFL,0);
    fcntl (udpfd,F_SETFL, sockflags | O_NONBLOCK);
      
    /* register the fd with the event handler of GTK */
    gdk_input_add(udpfd,GDK_INPUT_READ,udp_recv_callback,NULL);
      
    displayStatus("Sucessfully binded to the application UDP port");
  }
  else {
    displayStatus("unable to bind to UDP port");
  }
}

/* -----------------------End of UDP handler functions-------------------------- */

/* Building the GUI */
void BuildInterface() {

  GtkWidget *TopLevelBox;
  GtkWidget *MenuBox;
  GtkWidget *MainHBox;
  GtkWidget *SubHBox;
  GtkWidget *frame;
  GtkWidget *menubar;
  GtkWidget *actionMenu;
  GtkWidget *actionMenuItem;
  GtkWidget *menuitem;
  GtkWidget *assocDetailsScrollWin;
  GtkWidget *mesgstatusScrollWin; 
  GtkWidget *eventsScrollWin;
  GtkWidget *textlabel;
  GtkWidget *textentry;
  GtkWidget *buttonbox;
  GtkWidget *sendbutton;
  GtkWidget *clearbutton;
  GtkWidget *connectbutton;
  GtkWidget *disconnectbutton;

  /* Create and set the parameters of the window */
  mainWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position(GTK_WINDOW(mainWindow), GTK_WIN_POS_CENTER);
  gtk_window_set_default_size (GTK_WINDOW(mainWindow), DEFAULT_X_SIZE,DEFAULT_Y_SIZE);
  gtk_window_set_title (GTK_WINDOW(mainWindow), "SCTP Terminal Monitor");
  gtk_container_border_width (GTK_CONTAINER (mainWindow), 3);

  /* Connect the signal handlers for the window */
  gtk_signal_connect (GTK_OBJECT(mainWindow),"destroy",GTK_SIGNAL_FUNC(cleanUpanddExit),NULL);

  /* Create the TOP Level packing box */
  TopLevelBox = gtk_vbox_new (FALSE, 10);
  gtk_container_add (GTK_CONTAINER(mainWindow), TopLevelBox);

  /* Create the Main packing box */
  MainHBox = gtk_vbox_new (TRUE, 10);
  gtk_box_pack_end (GTK_BOX(TopLevelBox), MainHBox, TRUE, TRUE, 0);

  /* Create a Sub packing box */
  SubHBox = gtk_vbox_new (FALSE, 10);
  gtk_box_pack_end (GTK_BOX(MainHBox), SubHBox, TRUE, TRUE, 0);
  
  /* Create menu packing box and menu bar */
  MenuBox = gtk_vbox_new (FALSE, 10);
  menubar=gtk_menu_bar_new();
  gtk_box_pack_start (GTK_BOX(MenuBox), menubar, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX(TopLevelBox), MenuBox, FALSE, FALSE, 0);

  /* Create Action Menu----------------------------------------------------------------*/
  actionMenuItem = gtk_menu_item_new_with_label("Action");
  gtk_menu_bar_append(GTK_MENU_BAR(menubar),actionMenuItem);

  /* Action Menu */
  actionMenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(actionMenuItem),actionMenu);

  /* Action Menu Item -- Connect */
  menuitem =  gtk_menu_item_new_with_label("Connect");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(connectDialog),NULL);  
  /* Action Menu Item -- Start Timer */
  menuitem =  gtk_menu_item_new_with_label("Start Timer");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(startTimer),NULL);  

  /* Action Menu Item -- Stop Timer */
  menuitem =  gtk_menu_item_new_with_label("Stop Timer");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(stopTimer),NULL);  

  /* Action Menu Item -- Update Association and path details */
  menuitem =  gtk_menu_item_new_with_label("Update All");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(UpdateAll),NULL);  

  /* Action Menu Item -- Update path details */
  menuitem =  gtk_menu_item_new_with_label("Update Path Details");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(UpdatePath),NULL);  

  /* Action Menu Item -- Disconnect */
  menuitem =  gtk_menu_item_new_with_label("Disconnect");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(disconnectDialog),NULL);  

  /* Action Menu item -- Exit */
  menuitem =  gtk_menu_item_new_with_label("Exit");
  gtk_menu_append(GTK_MENU(actionMenu),menuitem);
  gtk_signal_connect(GTK_OBJECT(menuitem),"activate",GTK_SIGNAL_FUNC(cleanUpanddExit),NULL);
  
/* Association Details Tree display ----------------------------------------------------*/
  /* create the tree to hold the details */
  assocDetailsTree = gtk_tree_new();
  /* create scroll window to place the tree in */
  assocDetailsScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
						  GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(assocDetailsScrollWin),
					assocDetailsTree);

  /* Create a frame to hold SCTP Events Window and pack them accordingly */
  frame = gtk_frame_new ("SCTP Monitoring Details Window");
  gtk_container_add (GTK_CONTAINER(frame), assocDetailsScrollWin);
  gtk_box_pack_start (GTK_BOX(MainHBox), frame, TRUE, TRUE, 0);
  
/* Message Status Window----------------------------------------------------------------*/  
  /* create MesgInfoList to hold the path list */
  mesgstatusWindow = gtk_list_new(); 
  /* create scroll window to place the clist */
  mesgstatusScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					    GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(mesgstatusScrollWin),mesgstatusWindow);

  /* Create a frame to hold the SCTP Monitoring Tools Data and pack them accordingly */
  frame = gtk_frame_new ("Message Window - Scroll down for lastest message");
  gtk_container_add (GTK_CONTAINER(frame), mesgstatusScrollWin);
  gtk_box_pack_start (GTK_BOX(SubHBox), frame, TRUE, TRUE, 0);
 
/* SCTP Events status window-----------------------------------------------------------*/
  /* Create SCTP Events status window to hold the SCTP Events list */
  eventsWindow = gtk_list_new();

  /* create scroll window to place the clist */
  eventsScrollWin = gtk_scrolled_window_new(GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)),
					    GTK_ADJUSTMENT(gtk_adjustment_new (0,0,0,0,0,0)));
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(eventsScrollWin),eventsWindow);

  /* Create a frame to hold SCTP Events Window and pack them accordingly */
  frame = gtk_frame_new ("SCTP Events Window");
  gtk_container_add (GTK_CONTAINER(frame), eventsScrollWin);
  gtk_box_pack_start (GTK_BOX(SubHBox), frame, TRUE, TRUE, 0);
 
/* Textbox and Buttons-----------------------------------------------------------------*/
  /* Create the textbox */
  textbox = gtk_hbox_new ( FALSE, 10 );

  textlabel = gtk_label_new ( "Enter text here:" );
  gtk_box_pack_start( GTK_BOX ( textbox ), textlabel, FALSE, FALSE, 0 );

  textentry = gtk_entry_new (); 
  gtk_box_pack_start( GTK_BOX ( textbox ), textentry, TRUE, TRUE, 0 );
  gtk_box_pack_start (GTK_BOX(SubHBox), textbox, FALSE, FALSE, 0);

  /* Create the buttons */
  buttonbox = gtk_hbox_new ( FALSE, 10 );
  gtk_box_pack_start (GTK_BOX(SubHBox), buttonbox, FALSE, FALSE, 0);

  connectbutton = gtk_button_new_with_label ( "Connect" );
  gtk_box_pack_start( GTK_BOX ( buttonbox ), connectbutton, TRUE, TRUE, 0 );
  
  disconnectbutton = gtk_button_new_with_label ( "Disconnect" );
  gtk_box_pack_start( GTK_BOX ( buttonbox ), disconnectbutton, TRUE, TRUE, 0 );

  sendbutton = gtk_button_new_with_label ( "Send" );
  gtk_box_pack_start( GTK_BOX ( buttonbox ), sendbutton, TRUE, TRUE, 0 );

  clearbutton = gtk_button_new_with_label ( "Clear" );
  gtk_box_pack_start( GTK_BOX ( buttonbox ), clearbutton, TRUE, TRUE, 0 );
  
  /* Connect the buttons and "Enterkey" to signal functions */
  gtk_signal_connect (GTK_OBJECT(connectbutton), "clicked", 
		      GTK_SIGNAL_FUNC(connectDialog),NULL); 
  gtk_signal_connect (GTK_OBJECT(disconnectbutton), "clicked", 
		      GTK_SIGNAL_FUNC(disconnectDialog),NULL); 
  gtk_signal_connect (GTK_OBJECT(sendbutton), "clicked", 
		      GTK_SIGNAL_FUNC(Sendtext),textentry);
  gtk_signal_connect (GTK_OBJECT(clearbutton), "clicked", 
		      GTK_SIGNAL_FUNC(Cleartext),textentry);
  gtk_signal_connect (GTK_OBJECT(textentry), "activate",
		      GTK_SIGNAL_FUNC(Sendtext), textentry);

  /* Display the window on the screen */
  gtk_widget_show_all(mainWindow);  

}

gint main (gint argc, gchar * argv[]) {

  UDP_UA_CB udp_callbacks;
  char tstr[255];

  /* Trap Ctrl-C */
  signal(SIGINT, cleanUpanddExit);

  /* handle all command line options */
  getArgs(argc, argv);
  checkArgs();

  /* Required GTK initialization call */
  gtk_init (&argc, &argv);

  /* Build the user interface */
  BuildInterface();

  /* Register the UDP callbacks with the dispatch mechanism */
  udp_callbacks.communicationUpNotif     = &communicationUpNotif;
  udp_callbacks.networkStatusChangeNotif = &networkStatusChangeNotif;
  udp_callbacks.dataArriveNotif          = &dataArriveNotif;
  udp_callbacks.communicationLostNotif   = &communicationLostNotif;
  udp_callbacks.communicationErrorNotif  = &communicationErrorNotif;
  udp_callbacks.restartNotif             = &restartNotif;
  udp_callbacks.shutdownCompleteNotif    = &shutdownCompleteNotif;  
  
  /* Start listening on the UDP port on which the daemon sends notification */
  initializeUDPListening(UDP_Notif_Port);

  /* Initialize sctpdlib */
  if (sctpd_initLibrary(UDP_Notif_Port,UDP_Resp_Port)==1) {
    
    /* Send a registration message to the SCTPd */
    sctpInstance = sctpd_registerInstance(SCTP_App_Port,
					  MAXIMUM_NUMBER_OF_IN_STREAMS, 
					  MAXIMUM_NUMBER_OF_OUT_STREAMS,
					  noOfLocalAddresses,
					  localAddressList,udp_callbacks);

    sprintf(tstr,"Instance ID = %d",sctpInstance);
    displayStatus(tstr);
  
    if (sctpInstance!=0) {    
      /* Start the GTK loop */
      gtk_main();
    }

    else {
      printf("Unable to register an instance with sctpd\n");
      exit(1);
    }
  }
  else {
    printf("Unable to initialize sctpdlib\n");
    exit(1);
  }
  
  return 0;
}











