/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: interface for a User Adaptation Layer with SCTP demon program
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include "sctp.h"
#include "sctpudputils.h"

static unsigned char *localhostAddr          = "127.0.0.1";
static int response_port_fd;
static int notif_port_fd;

static unsigned short resp_port =0;
static unsigned short notif_port=0;

/* Private utility function */

unsigned short getRespPort() {
  return resp_port;
}

unsigned short getNotifPort() {
  return notif_port;
}

int initLocalUDPPort(int port) {

  struct sockaddr_in listenaddr;
  int sockfd=0;

  sockfd = socket(AF_INET, SOCK_DGRAM,0);
  bzero (&listenaddr,sizeof(listenaddr));

  listenaddr.sin_family = AF_INET;
  listenaddr.sin_port = htons(port);

  if (inet_aton(localhostAddr, &listenaddr.sin_addr)) {
    if ((bind(sockfd, (struct sockaddr *) &listenaddr, sizeof(listenaddr)))==0) {
      return sockfd;
    }
  }

  return 0;

}


/* Alarm signal handler : Doesn't do anything, just interrupt the process*/
static void sig_alarm(int signo) {
  return;
}


/* 
 * Waits for the response port for the return message from the 
 * sctpd on the return of a function call 
*/

int wait_for_response(fnCall *fc) {

  int len,n;
  struct sockaddr_in srcaddr;
  unsigned char temparr[MAX_UDP_DATA_LENGTH];
    
  bzero (&srcaddr,sizeof(srcaddr));

  /* set the alarm signal handler */
  signal(SIGALRM,sig_alarm);
   
  len = sizeof(srcaddr);	

#ifdef DEBUG
  fprintf(stdout,"Waiting for response\n");
  fflush(stdout);
#endif

  /* Set alarm for five seconds */
  alarm(5);

  n=recvfrom(response_port_fd,temparr,MAX_UDP_DATA_LENGTH,0, (struct sockaddr *) &srcaddr, &len);

  if (n>0) {
    unpack_function_call(temparr,fc);       	     	    
  }
  else if (n==-1) {
    fprintf(stdout,"Error in receiving response : %s\n",strerror(errno));
    fflush(stdout);
    return 0;
  }

  return 1;

}



/* Private function to handle the heartbeats */
void handle_heartbeat(unsigned int heartbeat_id) {

  unsigned char udpresparr[MAX_UDP_DATA_LENGTH];
  fnCall fr;

#ifdef SHOW_HEARTBEATS
  fprintf(stdout,"Heartbeat Received, ID = %d\n",heartbeat_id);
  fflush(stdout);
#endif

  create_func_call (&fr,UDP_HEARTBEAT_ACK);
  add_func_arg(&fr,ARG_TYPE_INT,sizeof(heartbeat_id),&heartbeat_id);  

  pack_function_call(&fr,udpresparr);
  sendnRawData(sctpdPort,udpresparr,MAX_UDP_DATA_LENGTH);      

}


/* IMPLEMENTATION OF API FOR UA */

/* Implementor note: The first value of the return structure should always be the
   return value of the native function call if available */

int sctpd_eventLoop(){

  int n=0;
  socklen_t len;
  struct sockaddr_in srcaddr;
  static int udpfd;
  unsigned char udparr[MAX_UDP_DATA_LENGTH];
  unsigned short udp_notif_port=0;
  
  if ((udp_notif_port=getNotifPort())!=0) {

    len = sizeof(srcaddr);
    
    if ((udpfd=initLocalUDPPort(udp_notif_port))>0) {
      while (1) {
	/* Wait for incoming notifications */
	n=recvfrom(udpfd,udparr,MAX_UDP_DATA_LENGTH,0, 
		   (struct sockaddr *) &srcaddr, &len);	
	if (n>0) {
	  dispatchUAMesg(udparr);
	}
	else if (n==-1) {
	  fprintf(stdout,"Error in receiving notifications : %s\n",
		  strerror(errno));
	  fflush(stdout);
	  break;
	}
      }
    }
    else {  
      return -1;
    }  
  } 
  else {
    fprintf(stdout,"sctpdlib not initialized\n");
    fflush(stdout);
  }
    
  return 0;
}


unsigned short sctpd_initLibrary(int udp_port,int udp_resp_port) {

  UDP_UA_CB udpcb;
  unsigned short retValue;
  unsigned char udpresparr[MAX_UDP_DATA_LENGTH];
  fnCall fr,funcresponse;
  short pid;
  int len=0;

  pid = getpid();

  notif_port= udp_port;
  resp_port = udp_resp_port;

  /* Register the response port  - done now so that response for registration
   * with daemon can be received.
   * TO DO : Close port if unsucessful 
   */

   response_port_fd = initLocalUDPPort(udp_resp_port) ;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTPD_INIT_LIBRARY message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fr,FN_SCTPD_INITLIBRARY);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(pid),&pid);  
  add_func_arg(&fr,ARG_TYPE_INT,sizeof(udp_port),&udp_port);
  add_func_arg(&fr,ARG_TYPE_INT,sizeof(udp_resp_port),&udp_resp_port);
  len = pack_function_call(&fr,udpresparr);
  sendnRawData(sctpdPort,udpresparr,len);      

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    if (funcresponse.fs.func_id== RESP_SCTPD_INITLIBRARY) {
#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTPD_INITLIBRARY message\n");
      fflush(stdout);
#endif
      /* Get the return value from the daemon */
      retValue = *(unsigned short *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);

      /* Register the Heartbeat handling function provided by the library */
      udpcb.handle_heartbeat = &handle_heartbeat;
      register_UA_UDP_Callbacks2(udpcb);

      return retValue;
    }
    free_func_args(&funcresponse);
  }
  return 0;
}


int sctpd_stopLibrary(){

  unsigned char udpresparr[MAX_UDP_DATA_LENGTH];
  fnCall fr;
  short pid;
  int len=0;

  pid = getpid();

#ifdef DEBUG
  fprintf(stdout,"Sending SCTPD_STOPLIBRARY message to SCTPd\n");
  fflush(stdout);
#endif

  create_func_call (&fr,FN_SCTPD_STOPLIBRARY);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(pid),&pid);  

  len = pack_function_call(&fr,udpresparr);
  sendnRawData(sctpdPort,udpresparr,len);      

  /* TO DO: WAIT FOR RESPONSE */

  /* Close the UDP ports */
  close(response_port_fd);

  return 1;

}



unsigned short sctpd_registerInstance(unsigned short localPort,
				      unsigned short noOfInStreams,
				      unsigned short noOfOutStreams,
				      unsigned int   noOfLocalAddresses,
				      unsigned char  localAddressList[][SCTP_MAX_IP_LEN],
				      UDP_UA_CB ULPcallbackFunctions) {
  int pid,i=0;
  unsigned short retValue;
  unsigned char udpresparr[MAX_UDP_DATA_LENGTH];
  fnCall fr,funcresponse;
  int len=0;

  pid = getpid();

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_REGISTER_INSTANCE message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fr,FN_SCTP_REGISTER_INSTANCE);
  add_func_arg(&fr,ARG_TYPE_INT,sizeof(pid),&pid);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(localPort),&localPort);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(noOfInStreams),&noOfInStreams);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(noOfOutStreams),&noOfOutStreams);
  add_func_arg(&fr,ARG_TYPE_INT,sizeof(noOfLocalAddresses),&noOfLocalAddresses);

  /* Add the addresses one by one into the data structure */
  for (i=0;i<noOfLocalAddresses;i++) {
    add_func_arg(&fr,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,localAddressList[i]);
  }

  len = pack_function_call(&fr,udpresparr);
  sendnRawData(sctpdPort,udpresparr,len);      

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */    
    if (funcresponse.fs.func_id==RESP_SCTP_REGISTER_INSTANCE) {
#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_REGISTER_INSTANCE message\n");
      fflush(stdout);
#endif
      /* Get the return value from the daemon */
      retValue = *(unsigned short *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);

      /* NOTE: This is a hack right now, as this means all the instances uses the same
       * callbacks. once there is a new set of callbacks registered for a new instance,
       * every previously registered instances use this new set 
       * TO DO: Implement a internal array of sets of handler pointers, each for each 
       * instance.
       */
      /* Insert the heartbeat handler again */
      ULPcallbackFunctions.handle_heartbeat = &handle_heartbeat;

      /* Register the user-supplied callback functions */
      register_UA_UDP_Callbacks2(ULPcallbackFunctions);

      return retValue;
    }

    free_func_args(&funcresponse);

  }
  return 0;

}


int sctpd_unregisterInstance(unsigned short instance_name) {

  unsigned char udpresparr[MAX_UDP_DATA_LENGTH];
  int retValue;
  fnCall fr,funcresponse;
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending Unregister Instance message to SCTPd\n");
  fflush(stdout);
#endif

  create_func_call (&fr,FN_SCTP_UNREGISTER_INSTANCE);
  add_func_arg(&fr,ARG_TYPE_SHORT,sizeof(instance_name),&instance_name);  

  len = pack_function_call(&fr,udpresparr);
  sendnRawData(sctpdPort,udpresparr,len);      

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    if (funcresponse.fs.func_id== RESP_SCTP_UNREGISTER_INSTANCE) {
#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_UNREGISTER_INSTANCE MESG\n");
      fflush(stdout);
#endif   
      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);
      return retValue;
    }
    free_func_args(&funcresponse);

  }
  return -1;      /* -1 = error ; 0 = success */

}


unsigned int sctpd_associate(unsigned short SCTP_InstanceName,
			     unsigned short noOfOutStreams,
			     unsigned char  *destinationAddress,
			     unsigned short destinationPort,
			     void* ulp_data) {

  int retValue;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

  fprintf(stdout,"Destination address = %s\n",destinationAddress);
  fprintf(stdout,"Destination port = %d\n",destinationPort);
  fflush(stdout);

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_ASSOCIATE message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_ASSOCIATE);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(SCTP_InstanceName),&SCTP_InstanceName);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(noOfOutStreams),&noOfOutStreams);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,destinationAddress);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(destinationPort),&destinationPort);  
  add_func_arg(&fc,ARG_TYPE_VOID_PTR,sizeof(ulp_data),&ulp_data);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_ASSOCIATE) {

      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);
#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_ASSOCIATE message (%d)\n",retValue);
      fflush(stdout);
#endif
      return retValue;
    }

    free_func_args(&funcresponse);
  }

  return 0;      /* 0 = error in this case */

}



int sctpd_shutdown(unsigned int associationID) {

  int retValue;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

  create_func_call (&fc,FN_SCTP_SHUTDOWN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

#ifdef DEBUG
  fprintf(stdout,"SCTP_SHUTDOWN message sent to sctpd for assocID = %d\n",associationID);
  fflush(stdout);
#endif

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SHUTDOWN) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SHUTDOWN MESG\n");
      fflush(stdout);
#endif
      
      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);
      return retValue;
    }
    else {
      printf("Wrong response received\n");
      free_func_args(&funcresponse);
    }

  }
  return 1;      /* 1 = error ; 0 = success */

}


int sctpd_abort(unsigned int associationID) {

  int retValue;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_ABORT message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_ABORT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    
    if (funcresponse.fs.func_id== RESP_SCTP_ABORT) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_ABORT\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);
      return retValue;
    }
    free_func_args(&funcresponse);
  }
  return 1;     /* 1 = error ; 0 = success */
  

}



int sctpd_send(unsigned int associationID, 
		   unsigned short streamID,
		   unsigned char *buffer, 
		   unsigned int length, 
		   unsigned int protocolId,
		   short path_id,  
		   void * context, 
		   unsigned int lifetime,
		   int unorderedDelivery,						
		   int dontBundle) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SEND message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_SEND);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(streamID),&streamID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,length,buffer);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(length),&length);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(protocolId),&protocolId);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);
  add_func_arg(&fc,ARG_TYPE_VOID_PTR,sizeof(context),&context);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(lifetime),&lifetime);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(unorderedDelivery),&unorderedDelivery);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(dontBundle),&dontBundle);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SEND) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SEND MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    else {
      printf("Wrong response received\n");
      free_func_args(&funcresponse);
    }
  }
  return -1;

}



unsigned short sctpd_receive(unsigned int   associationID,
			     unsigned short streamID,
			     unsigned char  *buffer,
			     unsigned int   *length,
			     unsigned int flags) {



  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_RECEIVE message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_RECEIVE);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(streamID),&streamID);
  /* Skip the buffer and length field */
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(flags),&flags);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_RECEIVE) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_RECEIVE MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	*length = *(int *)funcresponse.dataptr[4];
	memcpy(buffer,funcresponse.dataptr[3],*length);
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }

    free_func_args(&funcresponse);
    
  }
  return -1;



}



int sctpd_getAssocDefaults(unsigned short SCTP_InstanceName, SCTP_InstanceParameters* params) {


  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_ASSOC_DEFAULTS message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_ASSOC_DEFAULTS);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(SCTP_InstanceName),&SCTP_InstanceName);

  len = pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_ASSOC_DEFAULTS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_ASSOC_DEFAULTS\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	*params = *(SCTP_InstanceParameters *)funcresponse.dataptr[1];
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    else {
      printf("Wrong response received\n");
      free_func_args(&funcresponse);
    }
  }
  return -1;


}


int sctpd_setAssocDefaults(unsigned short SCTP_InstanceName, 
			       SCTP_InstanceParameters* params) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_ASSOC_DEFAULTS message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_SET_ASSOC_DEFAULTS);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(SCTP_InstanceName),&SCTP_InstanceName);
  add_func_arg(&fc,ARG_TYPE_INSTANCE_PARAMS,sizeof(*params),params);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SET_ASSOC_DEFAULTS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_ASSOC_DEFAULTS MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	*params = *(SCTP_InstanceParameters *)funcresponse.dataptr[1];
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


int sctpd_getAssocStatus(unsigned int associationID, 
			     SCTP_AssociationStatus* status) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending sctp_getAssocStatus message to SCTPd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_ASSOC_STATUS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    if (funcresponse.fs.func_id== RESP_SCTP_GET_ASSOC_STATUS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_ASSOC_STATUS MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	*status = *(SCTP_AssociationStatus *)funcresponse.dataptr[1];
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


int sctpd_setAssocStatus(unsigned int associationID, 
			     SCTP_AssociationStatus* new_status) {


  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_ASSOC_STATUS message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_SET_ASSOC_STATUS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_ASSOC_STATUS,sizeof(*new_status),new_status);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)){

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SET_ASSOC_STATUS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_ASSOC_STATUS MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


int sctpd_getPathStatus(unsigned int associationID, short path_id, 
			    SCTP_PathStatus* status) {


  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_PATH_STATUS message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_PATH_STATUS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_PATH_STATUS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_PATH_STATUS MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {
	*status = *(SCTP_PathStatus *)funcresponse.dataptr[1];
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


int sctpd_setPathStatus(unsigned int associationID, short path_id, 
			    SCTP_PathStatus *new_status) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_PATH_STATUS message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_SET_PATH_STATUS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);
  add_func_arg(&fc,ARG_TYPE_PATH_STATUS,sizeof(*new_status),new_status);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SET_PATH_STATUS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_PATH_STATUS MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


short sctpd_setPrimary(unsigned int associationID, short path_id) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_PRIMARY message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_SET_PRIMARY);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    
    if (funcresponse.fs.func_id== RESP_SCTP_SET_PRIMARY) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_PRIMARY MESG\n");
      fflush(stdout);
#endif

      if (*(short *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


short sctpd_getPrimary(unsigned int associationID) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  short primary;
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_PRIMARY message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_PRIMARY);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_PRIMARY) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_PRIMARY MESG\n");
      fflush(stdout);
#endif

      primary = *(short *)funcresponse.dataptr[0];
      if (primary!=-1) {      
	free_func_args(&funcresponse);
	return primary;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}


unsigned int sctpd_getSrttReport(unsigned int associationID, short path_id) {


  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  unsigned int srtt;
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_SRTT_REPORT message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_SRTT_REPORT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_SRTT_REPORT) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_SRTT_REPORT MESG\n");
      fflush(stdout);
#endif

      srtt = *(unsigned int *)funcresponse.dataptr[0];
      if (srtt!=0) {      
	free_func_args(&funcresponse);
	return srtt;
      }
      free_func_args(&funcresponse);
      return 0;
    }
    free_func_args(&funcresponse);
  }
  return 0;
}

int sctpd_changeHeartBeat(unsigned int associationID,
			      short path_id, int heartbeatON, unsigned int timeIntervall) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_CHANGE_HEARTBEAT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(heartbeatON),&heartbeatON);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(timeIntervall),&timeIntervall);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_CHANGE_HEARTBEAT) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_CHANGE_HEARTBEAT MESG\n");
      fflush(stdout);
#endif

      if (*(unsigned int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return 0;
    }
    free_func_args(&funcresponse);
  }
  return 0;

}


int sctpd_requestHeartbeat(unsigned int associationID, short path_id) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_REQUEST_HEARTBEAT message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_REQUEST_HEARTBEAT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(path_id),&path_id);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_REQUEST_HEARTBEAT) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_REQUEST_HEARTBEAT MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return 1;
    }
    free_func_args(&funcresponse);
  }
  return 1;

}


int sctpd_setFailureThreshold(unsigned int associationID,
				  unsigned short pathMaxRetransmissions) {


  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_FAILURE_THRES message to sctpd\n");
  fflush(stdout);
#endif
  
  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_SET_FAILURE_THRES);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(pathMaxRetransmissions),&pathMaxRetransmissions);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SET_FAILURE_THRES) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_FAILURE_THRES MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      free_func_args(&funcresponse);
      return 1;
    }
    free_func_args(&funcresponse);
  }
  return 1;

}


int sctpd_deleteAssociation(unsigned int associationID) {

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_DELETE_ASSOC message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_DELETE_ASSOC);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_DELETE_ASSOC) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_DELETE_ASSOC MESG\n");
      fflush(stdout);
#endif

      if (*(int *)funcresponse.dataptr[0]==0) {      
	free_func_args(&funcresponse);
	return 0;
      }
      else if (*(int *)funcresponse.dataptr[0]==1) {      
	free_func_args(&funcresponse);
	return 1;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;


}


int sctpd_getInstanceID(unsigned int associationID, 
			unsigned short* instanceID){

  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_INSTANCE_ID message to sctpd\n");
  fflush(stdout);
#endif

  /* Generate and send out request */
  create_func_call (&fc,FN_SCTP_GET_INSTANCE_ID);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */
    if (funcresponse.fs.func_id== RESP_SCTP_GET_INSTANCE_ID) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_INSTANCE_ID MESG\n");
      fflush(stdout);
#endif
      if (*(int *)funcresponse.dataptr[0]==0) {
	*instanceID=*(unsigned short*)funcresponse.dataptr[1];
	
	free_func_args(&funcresponse);
	return 0;
      }
      else if (*(int *)funcresponse.dataptr[0]==1) {      
	free_func_args(&funcresponse);
	return 1;
      }
      free_func_args(&funcresponse);
      return -1;
    }
    free_func_args(&funcresponse);
  }
  return -1;

}



int sctpd_receiveUnsent(unsigned int associationID, unsigned char *buffer, 
			unsigned int *length,unsigned short *streamID, 
			unsigned short *streamSN,unsigned int* protocolId) {

  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_RECEIVE_UNSENT message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_RECEIVE_UNSENT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  /* Skip the buffer field */
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*length),length);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamID),streamID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamSN),streamSN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*protocolId),protocolId);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_RECEIVE_UNSENT) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_RECEIVE_UNSENT MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      *length = *(int *)funcresponse.dataptr[3];
      memcpy(buffer,funcresponse.dataptr[2],*length);
      *streamID = *(unsigned short *)funcresponse.dataptr[4];
      *streamSN = *(unsigned short *)funcresponse.dataptr[5];
      *protocolId = *(unsigned int *)funcresponse.dataptr[6];
      
      free_func_args(&funcresponse);	                    
    
    }
  }

  return retValue;

}



int sctpd_receiveUnacked(unsigned int associationID, unsigned char *buffer, 
			 unsigned int *length,unsigned short *streamID, 
			 unsigned short *streamSN,unsigned int* protocolId) {
  
  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_RECEIVE_UNACK message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_RECEIVE_UNACK);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  /* Skip the buffer field */
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*length),length);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamID),streamID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamSN),streamSN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*protocolId),protocolId);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_RECEIVE_UNACK) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_RECEIVE_UNACK MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      *length = *(int *)funcresponse.dataptr[3];
      memcpy(buffer,funcresponse.dataptr[2],*length);
      *streamID = *(unsigned short *)funcresponse.dataptr[4];
      *streamSN = *(unsigned short *)funcresponse.dataptr[5];
      *protocolId = *(unsigned int *)funcresponse.dataptr[6];

      free_func_args(&funcresponse);          
    }    
  }
  return retValue;
}


int sctpd_addIPAddress(unsigned int associationID,
		       unsigned char address[SCTP_MAX_IP_LEN],
		       unsigned int *correlationId) {

  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_ADD_IP_ADDRESS message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_ADD_IP_ADDRESS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,address);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*correlationId),correlationId);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_ADD_IP_ADDRESS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_ADD_IP_ADDRESS MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      /* IP Addr field not returned */
      *correlationId = *(unsigned int *)funcresponse.dataptr[1];

      free_func_args(&funcresponse);          
    }    
  }
  return retValue;
}


int sctpd_deleteIPAddress(unsigned int associationID, 
			  unsigned char address[SCTP_MAX_IP_LEN],
			  unsigned int* correlationId) {


  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_DELETE_IP_ADDRESS message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_DELETE_IP_ADDRESS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,address);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*correlationId),correlationId);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_DELETE_IP_ADDRESS) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_DELETE_IP_ADDRESS MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      /* IP Addr field not returned */
      *correlationId = *(unsigned int *)funcresponse.dataptr[1];

      free_func_args(&funcresponse);          
    }    
  }
  return retValue;

}


int sctpd_setRemotePrimary(unsigned int associationID, 
			   unsigned char address[SCTP_MAX_IP_LEN]) {

  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_SET_REMOTE_PRIMARY message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_SET_REMOTE_PRIMARY);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,address);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_SET_REMOTE_PRIMARY) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_SET_REMOTE_PRIMARY MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);          
    }    
  }
  return retValue;

}


int sctpd_getNumberOfLocalAddresses(unsigned int associationID) {

  int retValue=0;
  int len=0;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  
#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_NO_LOCAL_ADDR message to sctpd\n");
  fflush(stdout);
#endif

  create_func_call (&fc,FN_SCTP_GET_NO_LOCAL_ADDR);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_NO_LOCAL_ADDR) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_NO_LOCAL_ADDR MESG\n");
      fflush(stdout);
#endif

      retValue = *(int *)funcresponse.dataptr[0];
      free_func_args(&funcresponse);          
    }    
  }
  return retValue;

}


unsigned char* sctpd_getLocalAddress(unsigned int associationID, int index) {

  char *retValue;
  fnCall fc,funcresponse;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  int len=0;

#ifdef DEBUG
  fprintf(stdout,"Sending SCTP_GET_LOCAL_ADDR message to sctpd\n");
  fflush(stdout);
#endif

  retValue = (char *)malloc(sizeof(char) * SCTP_MAX_IP_LEN);

  create_func_call (&fc,FN_SCTP_GET_LOCAL_ADDR);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(index),&index);

  len=pack_function_call(&fc,udpdataarr);
  sendnRawData(sctpdPort,udpdataarr,len);

  /* Wait for reply */
  if (wait_for_response(&funcresponse)) {

    /* Process and return */

    if (funcresponse.fs.func_id== RESP_SCTP_GET_LOCAL_ADDR) {

#ifdef DEBUG    
      fprintf(stdout,"Received response to SCTP_GET_LOCAL_ADDR MESG\n");
      fflush(stdout);
#endif

      
      strcpy(retValue,(char *)funcresponse.dataptr[0]);
      free_func_args(&funcresponse);          
    }    
  }
  return retValue;


}





