/*
 *  $Id: sctpsocket.cc,v 1.40 2002/01/07 12:08:14 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: SCTP Socket
 *
 */



#include "tdsystem.h"
#include "sctpsocket.h"
#include "sctpsocketmaster.h"


namespace Coral {


/*
#define PRINT_BIND
#define PRINT_UNBIND
#define PRINT_ACCEPT
#define PRINT_ASSOCIATE
#define PRINT_NEW_ASSOCIATIONS
#define PRINT_SHUTDOWNS

#define PRINT_AUTOCLOSE_TIMEOUT
#define PRINT_AUTOCLOSE_CHECK

#define PRINT_RECVWAIT
#define PRINT_ASSOCSEARCH
// #define PRINT_ASSOC_USECOUNT
*/


// ###### Constructor #######################################################
SCTPSocket::SCTPSocket(const cardinal flags)
{
   CorrelationID       = 0;
   AutoCloseTimeout    = 30000000;
   InstanceName        = 0;
   ConnectionRequests  = NULL;
   Flags               = flags;
   NotificationFlags   = 0;
   DefaultTrafficClass = 0x00;

   EstablishCondition.setName("SCTPSocket::EstablishCondition");
   ReadUpdateCondition.setName("SCTPSocket::ReadUpdateCondition");
   WriteUpdateCondition.setName("SCTPSocket::WriteUpdateCondition");
   ExceptUpdateCondition.setName("SCTPSocket::ExceptUpdateCondition");
   GlobalQueue.getUpdateCondition()->setName("SCTPSocket::GlobalQueueCondition");

   EstablishCondition.addParent(&ExceptUpdateCondition);
   GlobalQueue.getUpdateCondition()->addParent(&ReadUpdateCondition);
}


// ###### Destructor ########################################################
SCTPSocket::~SCTPSocket()
{
   unbind();
}


// ###### Get association for association ID ################################
SCTPAssociation* SCTPSocket::getAssociationForAssociationID(const unsigned int assocID,
                                                            const bool activeOnly)
{
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      AssociationList.find(assocID);
   if(iterator != AssociationList.end()) {
      if(!((iterator->second->IsShuttingDown) && (activeOnly))) {
         return(iterator->second);
      }
   }
   return(NULL);
}


// ###### Get local addresses ###############################################
bool SCTPSocket::getLocalAddresses(SocketAddress**& addressArray)
{
   // ====== Initialize =====================================================
   unsigned int i;
   bool         result = true;
   addressArray = SocketAddress::newAddressList(NoOfLocalAddresses);
   if(addressArray == NULL) {
      return(false);
   }

   // ====== Get local addresses ============================================
   SCTPSocketMaster::MasterInstance.lock();
   for(i = 0;i < NoOfLocalAddresses;i++) {
      addressArray[i] = SocketAddress::createSocketAddress(0,(char*)LocalAddressList[i],LocalPort);
      if(addressArray[i] == NULL) {
#ifndef DISABLE_WARNINGS
         cerr << "WARNING: SCTPSocket::getLocalAddresses() - Bad address "
              << LocalAddressList[i] << ", port " <<LocalPort << "!" << endl;
#endif
         SocketAddress::deleteAddressList(addressArray);
         result = false;
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();

   return(result);
}


// ###### Get remote addresses for given association ########################
bool SCTPSocket::getRemoteAddresses(SocketAddress**& addressArray,
                                    unsigned int     assocID)
{
   SCTPSocketMaster::MasterInstance.lock();

   // ====== Try to find association in association list ====================
   SCTPAssociation* association = getAssociationForAssociationID(assocID,false);
   if(association == NULL) {
      // ====== Try to find association in AutoConnect list =================
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         ConnectionlessAssociationList.find(assocID);
      if(iterator != ConnectionlessAssociationList.end()) {
         association = iterator->second;
      }
   }

   bool ok = false;
   if(association != NULL) {
      ok = association->getRemoteAddresses(addressArray);
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(ok);
}


// ###### Bind socket to local address(es) and port #########################
int SCTPSocket::bind(const unsigned short  localPort,
                     const unsigned short  noOfInStreams,
                     const unsigned short  noOfOutStreams,
                     const SocketAddress** localAddressList)
{
   if(SCTPSocketMaster::InitializationResult != 0) {
#ifndef DISABLE_WARNINGS
      cerr << "WARNING: SCTPSocket::bind() - SCTP is not initialized!" << endl;
#endif
      return(-EPROTONOSUPPORT);
   }

   SCTPSocketMaster::MasterInstance.lock();
   if(!SCTPSocketMaster::MasterInstance.running()) {
      // Start SCTPSocketMaster thread.
      if(SCTPSocketMaster::MasterInstance.start() == false) {
#ifndef DISABLE_WARNINGS
         cerr << "WARNING: SCTPSocket::bind() - Unable to start master thread!" << endl;
#endif
         SCTPSocketMaster::MasterInstance.unlock();
         return(-EPROTONOSUPPORT);
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();

   unsigned int noOfLocalAddresses = 0;
   while(localAddressList[noOfLocalAddresses] != NULL) {
      noOfLocalAddresses++;
   }
   SCTPSocketMaster::MasterInstance.lock();

   // ====== Initialize =====================================================
   unbind();
   LocalPort          = localPort;
   NoOfInStreams      = noOfInStreams;
   NoOfOutStreams     = noOfOutStreams;
   NoOfLocalAddresses = noOfLocalAddresses;
   NoOfLocalAddresses = 0;
   CorrelationID      = 0;


   // ====== Initialize local addresses =====================================
#ifdef PRINT_BIND
   cout << "Binding to {";
   for(unsigned int i = 0;i < noOfLocalAddresses;i++) {
      cout << " " << localAddressList[i]->getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy) << ":" << localPort << " ";
   }
   cout << "}." << endl;
#endif
   for(unsigned int i = 0;i < min(noOfLocalAddresses,(unsigned int)SCTP_MAX_NUM_ADDRESSES);i++) {
      snprintf((char*)&LocalAddressList[i],SCTP_MAX_IP_LEN,"%s",localAddressList[i]->getAddressString(SocketAddress::PF_HidePort|SocketAddress::PF_Address|SocketAddress::PF_Legacy).getData());
      NoOfLocalAddresses++;
   }
   if(NoOfLocalAddresses < 1) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPSocket::bind() - No local addresses!" << endl;
#endif
      SCTPSocketMaster::MasterInstance.unlock();
      return(-EINVAL);
   }


   // ====== Register SCTP instance =========================================
   InstanceName = sctp_registerInstance(LocalPort,NoOfInStreams,NoOfOutStreams,
                                        NoOfLocalAddresses,LocalAddressList,
                                        SCTPSocketMaster::Callbacks);
   if(InstanceName == 0) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPSocket::bind() - sctp_registerInstance() failed!" << endl;
#endif
      SCTPSocketMaster::MasterInstance.unlock();
      return(-EFAULT);
   }


   // ====== Add socket to global list ======================================
   SCTPSocketMaster::SocketList.insert(pair<unsigned short, SCTPSocket*>(InstanceName,this));


   SCTPSocketMaster::MasterInstance.unlock();
   return(0);
}


// ###### Release socket binding ############################################
void SCTPSocket::unbind(bool sendAbort)
{
   if(InstanceName != 0) {
      SCTPSocketMaster::MasterInstance.lock();
#ifdef PRINT_UNBIND
      cout << "Unbinding...";
#endif

      // ====== Delete all associations made by sendTo() ====================
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         ConnectionlessAssociationList.begin();
      while(iterator != ConnectionlessAssociationList.end()) {
         SCTPAssociation* association = iterator->second;
         ConnectionlessAssociationList.erase(iterator);
         if(sendAbort) {
            association->abort();
         }
         delete association;
         iterator = ConnectionlessAssociationList.begin();
      }

      // ====== Mark SCTP instance for unregistering ========================
      SCTPSocketMaster::delayedDeleteSocket(InstanceName);

      // ====== Remove socket from global list ==============================
      multimap<unsigned short, SCTPSocket*>::iterator socketIterator =
         SCTPSocketMaster::SocketList.find(InstanceName);
      if(socketIterator != SCTPSocketMaster::SocketList.end()) {
         SCTPSocketMaster::SocketList.erase(socketIterator);
      }
      else {
#ifndef DISABLE_WARNINGS
         cerr << "INTERNAL ERROR: SCTPSocket::unbind() - Erase failed!" << endl;
         ::exit(1);
#endif
      }

      // ====== Remove incoming associations ================================
      while(ConnectionRequests != NULL) {
         SCTPAssociation* association   = ConnectionRequests->Association;
         IncomingConnection* oldRequest = ConnectionRequests;
         ConnectionRequests = oldRequest->NextConnection;
         delete association;
         delete oldRequest;
      }

      SCTPSocketMaster::MasterInstance.unlock();

      GlobalQueue.flush();
      InstanceName  = 0;
      CorrelationID = 0;
      Flags &= ~SSF_Listening;

#ifdef PRINT_UNBIND
      cout << "Unbind complete." << endl;
#endif
   }
}


// ###### Set listen mode ###################################################
void SCTPSocket::listen(const unsigned int backlog)
{
   SCTPSocketMaster::MasterInstance.lock();
   Flags |= SSF_Listening;
   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Accept new association ############################################
SCTPAssociation* SCTPSocket::accept(SocketAddress*** addressArray,
                                    const bool       wait)
{
   if(addressArray != NULL) {
      *addressArray = NULL;
   }
   SCTPSocketMaster::MasterInstance.lock();
   if(!(Flags & SSF_Listening)) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPSocket::accept() - Socket is not in server mode, call listen() first!" << endl;
#endif
      return(NULL);
   }


   // ====== Get new association=============================================
   while(ConnectionRequests == NULL) {
      SCTPSocketMaster::MasterInstance.unlock();

      if(!wait) {
         return(NULL);
      }
      while(EstablishCondition.timedWait(100000) == false) {
         checkAutoConnect();
      }

      SCTPSocketMaster::MasterInstance.lock();
   }
   if(ConnectionRequests == NULL) {
      SCTPSocketMaster::MasterInstance.unlock();
      return(NULL);
   }


   // ====== Initialize address array =======================================
   if(addressArray != NULL) {
      *addressArray = SocketAddress::newAddressList(ConnectionRequests->Notification.RemoteAddresses);
      if(*addressArray == NULL) {
#ifndef DISABLE_WARNINGS
         cerr << "ERROR: SCTPSocket::accept() - Out of memory!" << endl;
#endif
      }
      else {
         unsigned int i;
         for(i = 0;i < ConnectionRequests->Notification.RemoteAddresses;i++) {
            (*addressArray)[i] = SocketAddress::createSocketAddress(
                                    0,
                                    (char*)&ConnectionRequests->Notification.RemoteAddress[i],
                                    ConnectionRequests->Notification.RemotePort);
            if((*addressArray)[i] == NULL) {
#ifndef DISABLE_WARNINGS
               cerr << "WARNING: SCTPSocket::accept() - Bad address "
                    << ConnectionRequests->Notification.RemoteAddress[i] << ", port " << ConnectionRequests->Notification.RemotePort << "!" << endl;
#endif
               SocketAddress::deleteAddressList(*addressArray);
            }
         }
      }
   }

#ifdef PRINT_ACCEPT
   cout << "Accepted association #" << ConnectionRequests->Association->AssociationID << " from {";
   for(unsigned int i = 0;i < ConnectionRequests->Notification.RemoteAddresses;i++) {
      InternetAddress address(String((char*)&ConnectionRequests->Notification.RemoteAddress[i]),ConnectionRequests->Notification.RemotePort);
      cout << " " << address.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy) << " ";
   }
   cout << "}." << endl;
#endif


   // ====== Remove association from incoming list ==========================
   SCTPAssociation* association   = ConnectionRequests->Association;
   IncomingConnection* oldRequest = ConnectionRequests;
   ConnectionRequests = oldRequest->NextConnection;
   delete oldRequest;


   SCTPSocketMaster::MasterInstance.unlock();
   return(association);
}


// ###### Establish new association #########################################
SCTPAssociation* SCTPSocket::associate(const unsigned short noOfOutStreams,
                                       const SocketAddress& destinationAddress,
                                       const bool           blocking)
{
   // ====== Establish new association ======================================
   SCTPSocketMaster::MasterInstance.lock();
   const unsigned int assocID = sctp_associate(InstanceName,
                                               (noOfOutStreams < 1) ? 1 : noOfOutStreams,
                                               (unsigned char*)destinationAddress.getAddressString(SocketAddress::PF_HidePort|SocketAddress::PF_Address|SocketAddress::PF_Legacy).getData(),
                                               destinationAddress.getPort(),
                                               NULL);
#ifdef PRINT_ASSOCIATE
   cout << "Association to " << destinationAddress.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy) << " => ID #"
        << assocID << "." << endl;
#endif


   // ====== Create SCTPAssociation object ==================================
   SCTPAssociation* association = NULL;
   if(assocID != 0) {
      association = new SCTPAssociation(this,assocID,NotificationFlags);
      if(association == NULL) {
         sctp_abort(assocID);
         sctp_deleteAssociation(assocID);
#ifndef DISABLE_WARNINGS
         cerr << "ERROR: SCTPSocket::associate() - Out of memory!" << endl;
#endif
      }
      else {
#ifdef PRINT_ASSOC_USECOUNT
         cout << "Associate: UseCount increment for A" << association->getID() << ": "
              << association->UseCount << " -> ";
#endif
         association->UseCount++;
#ifdef PRINT_ASSOC_USECOUNT
         cout << association->UseCount << endl;
#endif
         association->setTrafficClass(DefaultTrafficClass);
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();

   // ====== Wait for peer's connection up notification =====================
   if(association != NULL) {
      if(blocking) {
#ifdef PRINT_ASSOCIATE
         cout << "Waiting for establishment of association #" << assocID << "..." << endl;
#endif
         while(association->EstablishCondition.timedWait(100000) == false) {
            checkAutoConnect();
         }
         if(!association->CommunicationUpNotification) {
#ifdef PRINT_ASSOCIATE
            cout << "Association #" << assocID << " failed!" << endl;
#endif
            delete association;
            association = NULL;
         }
         else {
            association->setTrafficClass(DefaultTrafficClass);
         }
      }
   }

#ifdef PRINT_ASSOCIATE
   if(association != NULL) {
      cout << "Association #" << assocID << " established." << endl;
   }
#endif

   // ====== Decrement use count ============================================
   SCTPSocketMaster::MasterInstance.lock();
   if(association != NULL) {
#ifdef PRINT_ASSOC_USECOUNT
         cout << "Associate: UseCount decrement for A" << association->getID() << ": "
              << association->UseCount << " -> ";
#endif
      association->UseCount--;
#ifdef PRINT_ASSOC_USECOUNT
      cout << association->UseCount << endl;
#endif
   }
   SCTPSocketMaster::MasterInstance.unlock();

   return(association);
}


// ###### Get error code for given association ID ###########################
int SCTPSocket::getErrorCode(const unsigned int assocID)
{
   SCTPAssociation* association = getAssociationForAssociationID(assocID,false);
   if(association != NULL) {
      if(association->ShutdownCompleteNotification) {
         return(-ESHUTDOWN);
      }
      else if(association->CommunicationLostNotification) {
         return(-ECONNABORTED);
      }
   }
   return(0);
}


// ###### Internal receive implementation ###################################
int SCTPSocket::internalReceive(SCTPNotificationQueue& queue,
                                char*                  buffer,
                                size_t&                bufferSize,
                                int&                   flags,
                                unsigned int&          assocID,
                                unsigned short&        streamID,
                                unsigned int&          protoID,
                                SCTPNotification&      notification,
                                const unsigned int     notificationFlags)
{
   // ====== Check parameters ===============================================
   if(flags & MSG_WAITALL) {
#ifndef DISABLE_WARNINGS
      cerr << "NOT IMPLEMENTED: SCTPSocket::internalReceive() - MSG_WAITALL flag set!" << endl;
#endif
      return(-EINVAL);
   }
   else if(flags & MSG_ADDR_OVER) {
#ifndef DISABLE_WARNINGS
      cerr << "NOT IMPLEMENTED: SCTPSocket::internalReceive() - MSG_ADDR_OVER flag set!" << endl;
#endif
      return(-EINVAL);
   }
   if(bufferSize == 0) {
#ifndef DISABLE_WARNINGS
      cerr << "WARNING: SCTPSocket::internalReceive() - Data buffer size is zero!" << endl;
#endif
      return(-EINVAL);
   }


   // ====== Get next data or notification from queue =====================
#ifdef PRINT_RECVWAIT
   cout << "Waiting... ";
   cout.flush();
#endif
   SCTPSocketMaster::MasterInstance.lock();
   bool received = queue.getNotification(notification);
   while(received == false) {
      // ====== No chunk available -> wait for chunk ======================
      SCTPSocketMaster::MasterInstance.unlock();
      int errorCode = getErrorCode(assocID);
      if(errorCode != 0) {
         bufferSize = 0;
         return(errorCode);
      }
      if(flags & MSG_DONTWAIT) {
         bufferSize = 0;
         return(-EAGAIN);
      }
      while(queue.waitForChunk(100000) == false) {
         checkAutoConnect();
      }
      errorCode = getErrorCode(assocID);
      if(errorCode != 0) {
         bufferSize = 0;
         return(errorCode);
      }
      SCTPSocketMaster::MasterInstance.lock();
      received = queue.getNotification(notification);
   }
#ifdef PRINT_RECVWAIT
   cout << "Wakeup!" << endl;
#endif


   // ====== Read data ======================================================
   // If MSG_NOTIFICATION is set, notifications are received!
   const bool receiveNotifications = (flags & MSG_NOTIFICATION);
   bool updatedNotification = false;
   int result               = 0;
   sctp_notification_header* header = &notification.Content.sn_notification_header;
   if(header->snh_type == SCTP_DATA_ARRIVE) {
      flags &= ~MSG_NOTIFICATION;
      sctp_data_arrive* sda = &notification.Content.sn_data_arrive;
      if(sda->sda_bytes_arrived > 0) {
         assocID  = sda->sda_assoc_id;
         streamID = sda->sda_stream;
         protoID  = sda->sda_ppid;
         if(sda->sda_flags & SCTP_ARRIVE_UNORDERED) {
            flags |= MSG_UNORDERED;
         }
         size_t receivedBytes = min((size_t) sda->sda_bytes_arrived, (size_t) bufferSize);
         const int ok = sctp_receive(assocID, streamID,
                                     (unsigned char*)buffer,
                                     (unsigned int*)&receivedBytes,
                                     (flags & MSG_PEEK) ? SCTP_MSG_PEEK : SCTP_MSG_DEFAULT);
         if(ok == 0) {
            bufferSize = receivedBytes;

            // ====== Peek mode: Restore chunk arrival information ===================
            if(flags & MSG_PEEK) {
               queue.updateNotification(notification);
               updatedNotification = true;
            }
            else {
               sda->sda_bytes_arrived -= receivedBytes;
               if(sda->sda_bytes_arrived > 0) {
                  queue.updateNotification(notification);
                  updatedNotification = true;
               }
               else {
                  flags |= MSG_EOR;
               }
            }
         }
         else {
#ifndef DISABLE_WARNINGS
            cerr << "WARNING: SCTPSocket::internalReceive() - sctp_receive() failed!" << endl;
#endif
            result = -ECONNABORTED;
         }
      }
      else {
         bufferSize = 0;
      }
   }

   // ====== Handle notification ============================================
   else {
      switch(header->snh_type) {
         case SCTP_ASSOC_CHANGE:
            assocID = notification.Content.sn_assoc_change.sac_assoc_id;
          break;
         case SCTP_REMOTE_ERROR:
            assocID = notification.Content.sn_remote_error.sre_assoc_id;
          break;
         case SCTP_SEND_FAILED:
            assocID = notification.Content.sn_send_failed.ssf_assoc_id;
          break;
         case SCTP_SHUTDOWN_EVENT:
            assocID = notification.Content.sn_shutdown_event.sse_assoc_id;
          break;
      }

      // ====== Copy notification ===========================================
      if((receiveNotifications) &&
         (((header->snh_type == SCTP_ASSOC_CHANGE)     && (notificationFlags & SCTP_RECVASSOCEVNT)) ||
          ((header->snh_type == SCTP_PEER_ADDR_CHANGE) && (notificationFlags & SCTP_RECVPADDREVNT)) ||
          ((header->snh_type == SCTP_REMOTE_ERROR)     && (notificationFlags & SCTP_RECVPEERERR))   ||
          ((header->snh_type == SCTP_SEND_FAILED)      && (notificationFlags & SCTP_RECVSENDFAILEVNT)) ||
          ((header->snh_type == SCTP_SHUTDOWN_EVENT)   && (notificationFlags & SCTP_RECVSHUTDOWNEVNT)))) {
         const cardinal toCopy = min((cardinal)header->snh_length,(cardinal)bufferSize);
         const char* from = (char*)&notification.Content;
         memcpy(buffer,&from[notification.ContentPosition],toCopy);
         bufferSize = toCopy;
         notification.ContentPosition += toCopy;
         if(notification.ContentPosition < header->snh_length) {
            queue.updateNotification(notification);
            updatedNotification = true;
            flags |= MSG_NOTIFICATION;
         }
         else {
            flags |= (MSG_EOR|MSG_NOTIFICATION);
         }
      }
      else {
         result = -EAGAIN;
      }
   }


   // ====== Drop notification, if not updated ==============================
   if(!updatedNotification) {
      queue.dropNotification();
      SCTPAssociation* association = getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         association->LastUsage = getMicroTime();
         if(association->UseCount > 0) {
#ifdef PRINT_ASSOC_USECOUNT
            cout << "Receive: UseCount decrement for A" << association->getID() << ": "
                 << association->UseCount << " -> ";
#endif
            association->UseCount--;
#ifdef PRINT_ASSOC_USECOUNT
            cout << association->UseCount << endl;
#endif
         }
#ifndef DISABLE_WARNINGS
         else {
            cerr << "INTERNAL ERROR: SCTPSocket::internalReceive() - Too many association usecount decrements!" << endl;
            exit(1);
         }
#endif
      }
   }


   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Internal send implementation ######################################
int SCTPSocket::internalSend(const char*            buffer,
                             const size_t           length,
                             const int              flags,
                             const unsigned int     assocID,
                             const unsigned short   streamID,
                             const unsigned int     protoID,
                             Condition*             waitCondition)
{
   // ====== Check error code ===============================================
   const int errorCode = getErrorCode(assocID);
   if(errorCode != 0) {
      return(errorCode);
   }


   // ====== Do send ========================================================
   int result = 0;
   do {
      SCTPSocketMaster::MasterInstance.lock();
      result = sctp_send(
                  assocID, streamID,
                  (unsigned char*)buffer, length,
                  protoID,
                  SCTP_USE_PRIMARY,
                  SCTP_NO_CONTEXT,
                  SCTP_INFINITE_LIFETIME,
                  ((flags & MSG_UNORDERED) ? SCTP_UNORDERED_DELIVERY : SCTP_ORDERED_DELIVERY),
                  ((flags & MSG_UNBUNDLED) ? SCTP_BUNDLING_DISABLED : SCTP_BUNDLING_ENABLED));
      if((result == SCTP_QUEUE_EXCEEDED) && (!(flags & MSG_DONTWAIT))) {
         if(waitCondition != NULL) {
            SCTPSocketMaster::MasterInstance.unlock();
            waitCondition->timedWait(100000);
            SCTPSocketMaster::MasterInstance.lock();
         }
      }
      SCTPSocketMaster::MasterInstance.unlock();
   }
   while((!(flags & MSG_DONTWAIT)) && (result == SCTP_QUEUE_EXCEEDED));

   return((result == 0) ? (int)length : -EIO);
}


// ###### Receive ###########################################################
int SCTPSocket::receive(char*           buffer,
                        size_t&         bufferSize,
                        int&            flags,
                        unsigned int&   assocID,
                        unsigned short& streamID,
                        unsigned int&   protoID)
{
   SCTPNotification notification;
   return(receiveFrom(buffer,bufferSize,
                      flags,
                      assocID,streamID,protoID,
                      NULL,notification));
}


// ###### Check, if socket has data for given flags #########################
bool SCTPSocket::hasData(const int flags)
{
   if(!(Flags & SSF_GlobalQueue)) {
#ifndef DISABLE_WARNINGS
      cerr << "WARNING: SCTPSocket::receiveFrom() - No global queue!" << endl;
#endif
      return(-EBADF);
   }

   SCTPSocketMaster::MasterInstance.lock();
   const bool result = GlobalQueue.hasData(flags, NotificationFlags);
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Receive ###########################################################
int SCTPSocket::receiveFrom(char*              buffer,
                            size_t&            bufferSize,
                            int&               flags,
                            unsigned int&      assocID,
                            unsigned short&    streamID,
                            unsigned int&      protoID,
                            SocketAddress***   addressArray,
                            SCTPNotification&  notification)
{
   // ====== Receive ========================================================
   if(!(Flags & SSF_GlobalQueue)) {
#ifndef DISABLE_WARNINGS
      cerr << "WARNING: SCTPSocket::receiveFrom() - No global queue!" << endl;
#endif
      return(-EBADF);
   }
   assocID = 0;
   const int result = internalReceive(
                         GlobalQueue,
                         buffer, bufferSize,
                         flags,
                         assocID, streamID, protoID,
                         notification,
                         NotificationFlags);

   // ====== Extract sender's addresses =====================================
   if((result == 0) && (addressArray != NULL)) {
      unsigned int i;
      *addressArray = SocketAddress::newAddressList(notification.RemoteAddresses);
      if(*addressArray != NULL) {
         for(i = 0;i < notification.RemoteAddresses;i++) {
            (*addressArray)[i] = SocketAddress::createSocketAddress(
                                    0,
                                    (char*)&notification.RemoteAddress[i],
                                    notification.RemotePort);
            if(addressArray[i] == NULL) {
#ifndef DISABLE_WARNINGS
               cerr << "WARNING: SCTPSocket::receiveFrom() - Bad address \""
                    << (char*)&notification.RemoteAddress[i] << "\"!" << endl;
#endif
            }
         }
      }
   }

   // ====== Check, if association has to be closed =========================
   checkAutoConnect();

   return(result);
}


// ###### Send ##############################################################
int SCTPSocket::sendTo(const char*          buffer,
                       const size_t         length,
                       const int            flags,
                       const unsigned short streamID,
                       const unsigned int   protoID,
                       const SocketAddress& destinationAddress,
                       const cardinal       noOfOutgoingStreams)
{
   SCTPSocketMaster::MasterInstance.lock();

   // ====== Check for already created association ==========================
   SCTPAssociation* association = NULL;
   if(Flags & SSF_AutoConnect) {
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         ConnectionlessAssociationList.begin();
      while(iterator != ConnectionlessAssociationList.end()) {
         SCTP_Association_Status status;
         if(sctp_getAssocStatus(iterator->second->AssociationID,&status) == 0) {
#ifdef PRINT_ASSOCSEARCH
            cout << "CL "
                 << destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy)
                 << " == "
                 << String((const char*)&status.primaryDestinationAddress)
                 << "?" << endl;
#endif
            if( (!iterator->second->IsShuttingDown)               &&
                (destinationAddress.getPort() == status.destPort) &&
                (destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy) == String((const char*)&status.primaryDestinationAddress)) ) {
               association = iterator->second;
               break;
            }
         }
         iterator++;
      }
   }
   if(association == NULL) {
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         AssociationList.begin();
      while(iterator != AssociationList.end()) {
         SCTP_Association_Status status;
         if(sctp_getAssocStatus(iterator->second->AssociationID,&status) == 0) {
#ifdef PRINT_ASSOCSEARCH
            cout << "CO "
                 << destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy)
                 << " == "
                 << String((const char*)&status.primaryDestinationAddress)
                 << "?" << endl;
#endif
            if( (!iterator->second->IsShuttingDown)               &&
                (destinationAddress.getPort() == status.destPort) &&
                (destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy) == String((const char*)&status.primaryDestinationAddress)) ) {
               association = iterator->second;
               break;
            }
         }
         iterator++;
      }
   }
   if(association != NULL) {
#ifdef PRINT_ASSOC_USECOUNT
      cout << "Send: UseCount increment for A" << association->getID() << ": "
           << association->UseCount << " -> ";
#endif
      association->UseCount++;
#ifdef PRINT_ASSOC_USECOUNT
      cout << association->UseCount << endl;
#endif
   }

   SCTPSocketMaster::MasterInstance.unlock();


   // ====== Create new association =========================================
   if((Flags & SSF_AutoConnect) && (association == NULL)) {
#ifdef PRINT_NEW_ASSOCIATIONS
      cout << "AutoConnect: New outgoing association to "
           << destinationAddress.getAddressString(InternetAddress::PF_Address|InternetAddress::PF_Legacy)
           << "..." << endl;
#endif
      association = associate(noOfOutgoingStreams,destinationAddress,
                              (flags & MSG_DONTWAIT) ? false : true);
      if(association != NULL) {
#ifdef PRINT_NEW_ASSOCIATIONS
         cout << "AutoConnect: New outgoing association to "
                 << destinationAddress.getAddressString(InternetAddress::PF_Address|InternetAddress::PF_Legacy)
                 << ", #" << association->getID() << " established!" << endl;
#endif
         SCTPSocketMaster::MasterInstance.lock();
#ifdef PRINT_ASSOC_USECOUNT
         cout << "AutoConnect: UseCount increment for A" << association->getID() << ": "
              << association->UseCount << " -> ";
#endif
         association->UseCount++;
#ifdef PRINT_ASSOC_USECOUNT
         cout << association->UseCount << endl;
#endif
         ConnectionlessAssociationList.insert(pair<unsigned int, SCTPAssociation*>(association->getID(),association));
         SCTPSocketMaster::MasterInstance.unlock();
      }
#ifdef PRINT_NEW_ASSOCIATIONS
      else {
         cout << "AutoConnect: New outgoing association to "
              << destinationAddress.getAddressString(InternetAddress::PF_Address|InternetAddress::PF_Legacy)
              << " failed!" << endl;
      }
#endif
   }


   // ====== Send data ======================================================
   int result;
   if(association != NULL) {
      result = association->send(buffer,length,flags,streamID,protoID);

      // ====== Remove association, if SHUTDOWN flag is set =================
      if((flags & MSG_EOF) || (flags & MSG_ABORT)) {
         if(flags & MSG_ABORT) {
            association->abort();
         }
         else if(flags & MSG_EOF) {
            association->shutdown();
         }
         if(Flags & SSF_AutoConnect) {
#ifdef PRINT_SHUTDOWNS
            cout << "AutoConnect: Shutdown of outgoing association to " << destinationAddress << "..." << endl;
#endif
            SCTPSocketMaster::MasterInstance.lock();
            multimap<unsigned int, SCTPAssociation*>::iterator iterator =
               ConnectionlessAssociationList.find(association->getID());
            if(iterator != ConnectionlessAssociationList.end()) {
               ConnectionlessAssociationList.erase(iterator);
            }
            SCTPSocketMaster::MasterInstance.unlock();
            delete association;
#ifdef PRINT_SHUTDOWNS
            cout << "AutoConnect: Shutdown of outgoing association to " << destinationAddress << " completed!" << endl;
#endif
         }
         checkAutoConnect();
      }
   }
   else {
      result = -EIO;
   }


   SCTPSocketMaster::MasterInstance.lock();
   if(association != NULL) {
      association->LastUsage = getMicroTime();
      if(association->UseCount > 0) {
#ifdef PRINT_ASSOC_USECOUNT
      cout << "Send: UseCount decrement for A" << association->getID() << ": "
           << association->UseCount << " -> ";
#endif
         association->UseCount--;
#ifdef PRINT_ASSOC_USECOUNT
      cout << association->UseCount << endl;
#endif
      }
#ifndef DISABLE_WARNINGS
      else {
         cerr << "INTERNAL ERROR: SCTPSocket::sendTo() - Too many association usecount decrements!" << endl;
         exit(1);
      }
#endif
   }
   SCTPSocketMaster::MasterInstance.unlock();


   return(result);
}


// ###### Peel UDP-like association off #####################################
SCTPAssociation* SCTPSocket::peelOff(const SocketAddress& destinationAddress)
{
   SCTPAssociation* association = NULL;
   SCTPSocketMaster::MasterInstance.lock();

   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.begin();
   while(iterator != ConnectionlessAssociationList.end()) {
      SCTP_Association_Status status;
      if(sctp_getAssocStatus(iterator->second->AssociationID,&status) == 0) {
#ifdef PRINT_ASSOCSEARCH
         cout << "CL "
              << destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy)
              << " == "
              << String((const char*)&status.primaryDestinationAddress)
              << "?" << endl;
#endif
         if( (!iterator->second->IsShuttingDown)               &&
             (destinationAddress.getPort() == status.destPort) &&
             (destinationAddress.getAddressString(InternetAddress::PF_HidePort|InternetAddress::PF_Address|InternetAddress::PF_Legacy) == String((const char*)&status.primaryDestinationAddress)) ) {
            ConnectionlessAssociationList.erase(iterator);
            association = iterator->second;
            break;
         }
      }
      iterator++;
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(association);
}


// ###### Peel UDP-like association off #####################################
SCTPAssociation* SCTPSocket::peelOff(const unsigned int assocID)
{
   SCTPAssociation* association = NULL;

   SCTPSocketMaster::MasterInstance.lock();
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.find(assocID);
   if( (iterator != ConnectionlessAssociationList.end()) &&
       (!iterator->second->IsShuttingDown) ) {
      association = iterator->second;
      ConnectionlessAssociationList.erase(iterator);
   }
   SCTPSocketMaster::MasterInstance.unlock();

   return(association);
}


// ###### Get instance parameters ###########################################
void SCTPSocket::getInstanceParameters(SCTP_Instance_Parameters& instanceParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_getAssocDefaults(InstanceName,&instanceParameters);
   SCTPSocketMaster::MasterInstance.unlock();
#ifndef DISABLE_WARNINGS
   if(ok != 0) {
      cerr << "INTERNAL ERROR: SCTPSocket::getInstanceParameters() - "
              "sctp_getAssocDefaults() failed!" << endl;
      exit(1);
   }
#endif
}


// ###### Set instance parameters ###########################################
bool SCTPSocket::setInstanceParameters(const SCTP_Instance_Parameters& instanceParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_setAssocDefaults(InstanceName,
                                        (SCTP_Instance_Parameters*)&instanceParameters);
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok == 0);
}


// ###### Get association parameters ########################################
bool SCTPSocket::getAssociationParameters(const unsigned int       assocID,
                                          SCTP_Association_Status& associationParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_getAssocStatus(assocID,&associationParameters);
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok == 0);
}


// ###### Set association parameters ########################################
bool SCTPSocket::setAssociationParameters(const unsigned int             assocID,
                                          const SCTP_Association_Status& associationParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_setAssocStatus(InstanceName,
                                      (SCTP_Association_Status*)&associationParameters);
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok == 0);
}


// ###### Get path parameters ###############################################
bool SCTPSocket::getPathParameters(const unsigned int assocID,
                                   const int          pathIndex,
                                   SCTP_PathStatus&   pathParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_getPathStatus(assocID,
                                     (pathIndex < 0) ? sctp_getPrimary(assocID) : (short)pathIndex,
                                     &pathParameters);
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok == 0);
}


// ###### Set path parameters ###############################################
bool SCTPSocket::setPathParameters(const unsigned int     assocID,
                                   const int              pathIndex,
                                   const SCTP_PathStatus& pathParameters)
{
   SCTPSocketMaster::MasterInstance.lock();
   const int ok = sctp_setPathStatus(assocID,
                                     (pathIndex < 0) ? sctp_getPrimary(assocID) : (short)pathIndex,
                                     (SCTP_PathStatus*)&pathParameters);
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok == 0);
}


// ###### Set send buffer size ##############################################
bool SCTPSocket::setSendBuffer(const size_t size)
{
   bool ok = true;
   SCTPSocketMaster::MasterInstance.lock();
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.begin();
   if(iterator != ConnectionlessAssociationList.end()) {
      SCTPAssociation* association = iterator->second;
      if(association->setSendBuffer(size) == false) {
         ok = false;
      }
      iterator++;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok);
}


// ###### Set receive buffer size ###########################################
bool SCTPSocket::setReceiveBuffer(const size_t size)
{
   bool ok = true;
   SCTPSocketMaster::MasterInstance.lock();
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.begin();
   if(iterator != ConnectionlessAssociationList.end()) {
      SCTPAssociation* association = iterator->second;
      if(association->setReceiveBuffer(size) == false) {
         ok = false;
      }
      iterator++;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok);
}


// ###### Set traffic class #################################################
bool SCTPSocket::setTrafficClass(const card8 trafficClass,
                                 const int   streamID)
{
   bool ok = true;
   SCTPSocketMaster::MasterInstance.lock();
   DefaultTrafficClass = trafficClass;
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.begin();
   if(iterator != ConnectionlessAssociationList.end()) {
      SCTPAssociation* association = iterator->second;
      if(association->setTrafficClass(trafficClass,streamID) == false) {
         ok = false;
      }
      iterator++;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(ok);
}


// ###### Get primary address ###############################################
SocketAddress* SCTPSocket::getPrimaryAddress(const unsigned int assocID)
{
   SCTPSocketMaster::MasterInstance.lock();

   SocketAddress* address = NULL;
   const int index        = sctp_getPrimary(assocID);
   if(index >= 0) {
      SCTP_Path_Status pathStatus;
      const int result = sctp_getPathStatus(assocID,index,&pathStatus);
      if(result == 0) {
         address = SocketAddress::createSocketAddress(0,(char*)&pathStatus.destinationAddress);
      }
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(address);
}


// ###### Set primary address ###############################################
bool SCTPSocket::setPrimary(const unsigned int   assocID,
                            const SocketAddress& primary)
{
   SCTPSocketMaster::MasterInstance.lock();

   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            primary.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy).getData());
   SCTP_Path_Status pathStatus;
   int index  = 0;
   int result = sctp_getPathStatus(assocID,index,&pathStatus);
   bool ok    = false;
   while(result == 0) {
      if(!(strcasecmp((char*)&address,(char*)&pathStatus.destinationAddress))) {
         result = sctp_setPrimary(assocID,index);
         if(result == 0) {
            ok = true;
         }
         break;
      }

      index++;
      result = sctp_getPathStatus(assocID,index,&pathStatus);
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(ok);
}


// ###### Set peer primary address ##########################################
bool SCTPSocket::setPeerPrimary(const unsigned int   assocID,
                                const SocketAddress& primary)
{
   SCTPSocketMaster::MasterInstance.lock();
   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            primary.getAddressString().getData());
   const int result = sctp_setRemotePrimary(assocID,address);
   SCTPSocketMaster::MasterInstance.unlock();
   return(result == 0);
}


// ###### Add address #######################################################
bool SCTPSocket::addAddress(const unsigned int   assocID,
                            const SocketAddress& addAddress)
{
   if(assocID == 0) {
      bool ok = true;
      SCTPSocketMaster::MasterInstance.lock();
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         ConnectionlessAssociationList.begin();
      if(iterator != ConnectionlessAssociationList.end()) {
         SCTPAssociation* association = iterator->second;
         if(association->addAddress(addAddress) == false) {
            ok = false;
         }
         iterator++;
      }
      SCTPSocketMaster::MasterInstance.unlock();
      return(ok);
   }

   SCTPSocketMaster::MasterInstance.lock();
   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            addAddress.getAddressString().getData());
   const int result = sctp_addIPAddress(assocID,address,&CorrelationID);
   CorrelationID++;
   SCTPSocketMaster::MasterInstance.unlock();
   return(result == 0);
}


// ###### Delete address ####################################################
bool SCTPSocket::deleteAddress(const unsigned int   assocID,
                               const SocketAddress& delAddress)
{
   if(assocID == 0) {
      bool ok = true;
      SCTPSocketMaster::MasterInstance.lock();
      multimap<unsigned int, SCTPAssociation*>::iterator iterator =
         ConnectionlessAssociationList.begin();
      if(iterator != ConnectionlessAssociationList.end()) {
         SCTPAssociation* association = iterator->second;
         if(association->deleteAddress(delAddress) == false) {
            ok = false;
         }
         iterator++;
      }
      SCTPSocketMaster::MasterInstance.unlock();
      return(ok);
   }

   SCTPSocketMaster::MasterInstance.lock();
   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            delAddress.getAddressString().getData());
   const int result = sctp_deleteIPAddress(assocID,address,&CorrelationID);
   CorrelationID++;
   SCTPSocketMaster::MasterInstance.unlock();
   return(result == 0);
}


// ###### Check for necessity to auto-close associations ####################
void SCTPSocket::checkAutoClose()
{
   const card64 now = getMicroTime();
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      ConnectionlessAssociationList.begin();
   while(iterator != ConnectionlessAssociationList.end()) {

      SCTPAssociation* association = iterator->second;
#ifdef PRINT_AUTOCLOSE_CHECK
      cout << "AutoConnect: Check for AutoClose:" << endl
           << "   AssocID          = " << association->getID() << endl
           << "   UseCount         = " << association->UseCount << endl
           << "   LastUsage        = " << now - association->LastUsage << endl
           << "   AutoCloseTimeout = " << AutoCloseTimeout << endl;
#endif
      if((association->UseCount == 0) &&
        ((now - association->LastUsage > AutoCloseTimeout) ||
         (association->ShutdownCompleteNotification)       ||
         (association->CommunicationLostNotification))) {
#ifdef PRINT_AUTOCLOSE_TIMEOUT
         const unsigned int assocID = association->getID();
         cout << "AutoConnect: AutoClose of association #" << assocID << " due to ";
         if(association->LastUsage + AutoCloseTimeout < now) {
            cout << "timeout";
         }
         else if(association->ShutdownCompleteNotification) {
            cout << "shutdown complete";
         }
         else if(association->CommunicationLostNotification) {
            cout << "communication lost";
         }
         cout << "..." << endl;
#endif
         delete association;
#ifdef PRINT_AUTOCLOSE_TIMEOUT
         cout << "AutoConnect: AutoClose of association #" << assocID << " completed!" << endl;
#endif
         ConnectionlessAssociationList.erase(iterator);
         iterator = ConnectionlessAssociationList.begin();
      }
      else {
         iterator++;
      }
   }
}


// ###### AutoConnect maintenance ###########################################
void SCTPSocket::checkAutoConnect()
{
   if(Flags & SSF_AutoConnect) {
      SCTPSocketMaster::MasterInstance.lock();

      // ====== Check, if there are new incoming associations ===============
      const cardinal oldFlags = Flags;
      Flags |= SSF_Listening;
      SCTPAssociation* association = accept(NULL,false);
      while(association != NULL) {
#ifdef PRINT_NEW_ASSOCIATIONS
         cout << "AutoConnect: New incoming association #" << association->getID() << "..." << endl;
#endif
         ConnectionlessAssociationList.insert(pair<unsigned int, SCTPAssociation*>(association->getID(),association));
         association = accept(NULL,false);
      }
      Flags = oldFlags;

      SCTPSocketMaster::MasterInstance.unlock();
   }
}


}
