/*
 *  $Id: sctpsocketmaster.cc,v 1.37 2002/01/07 15:22:43 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: SCTP Socket Master
 *
 */



#include "tdsystem.h"
#include "sctpsocketmaster.h"


#include <signal.h>


namespace Coral {


/*
#define PRINT_NOTIFICATIONS
#define PRINT_ARRIVENOTIFICATION
// #define PRINT_GC
// #define PRINT_ASSOC_USECOUNT
*/

// Do not show a warning on initialization failure.
#define NO_INITFAIL_WARNING



// ###### Constructor #######################################################
SCTPSocketMaster::SCTPSocketMaster()
   : Thread("SCTPSocketMaster")
{
   if(InitializationResult == -1000) {
      Callbacks.dataArriveNotif          = &dataArriveNotif;
      Callbacks.sendFailureNotif         = &sendFailureNotif;
      Callbacks.networkStatusChangeNotif = &networkStatusChangeNotif;
      Callbacks.communicationUpNotif     = &communicationUpNotif;
      Callbacks.communicationLostNotif   = &communicationLostNotif;
      Callbacks.communicationErrorNotif  = &communicationErrorNotif;
      Callbacks.restartNotif             = &restartNotif;
      Callbacks.shutdownCompleteNotif    = &shutdownCompleteNotif;
      Callbacks.queueStatusChangeNotif   = &queueStatusChangeNotif;
      Callbacks.asconfStatusNotif        = &asconfStatusNotif;
      InitializationResult = sctp_initLibrary();
      if(InitializationResult == 0) {
         enableOOTBHandling(false);
         LastGarbageCollection = getMicroTime();

         // The thread may not be started here, since
         // its static initializers may *not* be initialized!
      }
      else {
#ifndef NO_INITFAIL_WARNING
         cerr << "ERROR: SCTP Library initialization failed!" << endl;
         if(getuid() != 0) {
            cerr << "       You need root permissions to use the SCTP Library!" << endl;
         }
#endif
      }
   }
   else {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPSocketMaster::SCTPSocketMaster() - "
              "Do not try to initialice SCTPSocketMaster singleton twice!" << endl;
#endif
   }
}


// ###### Destructor ########################################################
SCTPSocketMaster::~SCTPSocketMaster()
{
#ifdef PRINT_GC
   cout << "garbageCollection: SocketMaster destructor..." << endl;
#endif

   // ====== Stop master thread =============================================
   lock();
   cancel();
   PThread = 0;
   unlock();

   // ====== Do garbage collection for associations =========================
   lock();
   multimap<unsigned int, unsigned short>::iterator iterator = ClosingAssociations.begin();
   while(iterator != ClosingAssociations.end()) {
      // associationGarbageCollection(iterator->first,true) may not be called
      // here, since sctp_abort() directly calls communicationLostNotification(),
      // which itself calls associationGarbageCollection(iterator->first,true)
      // again. Therefore, sctp_abort() is used here.
      sctp_abort(iterator->first);
      iterator = ClosingAssociations.begin();
   }
   unlock();

   // ====== Do garbage collection for sockets ==============================
   socketGarbageCollection();

#ifdef PRINT_GC
   cout << "garbageCollection: SocketMaster destructor completed!" << endl;
#endif
}


// ###### SCTP event loop thread ############################################
void SCTPSocketMaster::run()
{
   for(;;) {
      card64 now   = getMicroTime();
      card64 msecs = (LastGarbageCollection + GarbageCollectionInterval - now) / 1000;

      MasterInstance.lock();
      const int tid = sctp_startTimer((unsigned int)msecs,timerCallback,NULL,NULL);
      MasterInstance.unlock();

      sctp_extendedEventLoop(lock,unlock,(void*)this);

      MasterInstance.lock();
      sctp_stopTimer(tid);
      MasterInstance.unlock();

      now = getMicroTime();
      if(now - LastGarbageCollection >= GarbageCollectionInterval) {
         socketGarbageCollection();
      }
   }
}


// ###### SCTP timer callback ###############################################
void SCTPSocketMaster::timerCallback(unsigned int tid,
                                     void*        param1,
                                     void*        param2)
{
}


// ###### Enable or disable OOTB handling ###################################
void SCTPSocketMaster::enableOOTBHandling(const bool enable)
{
   MasterInstance.lock();
   SCTP_Library_Parameters parameters;
   parameters.sendOotbAborts = 0;
#ifndef DISABLE_WARNINGS
   if(sctp_setLibraryParameters(&parameters) != SCTP_SUCCESS) {
      cerr << "WARNING: SCTPSocketMaster::enableOOTBHandling() - Setting of SCTP Library parameters failed!" << endl;
   }
#endif
   MasterInstance.unlock();
}


// ###### Add AssociationID to be deleted ###################################
void SCTPSocketMaster::delayedDeleteAssociation(const unsigned short instanceID,
                                                const unsigned int   assocID)
{
#ifdef PRINT_GC
   cout << "delayedDeleteAssociation: A=" << assocID << " I=" << instanceID << endl;
#endif
   ClosingAssociations.insert(pair<unsigned int, unsigned short>(assocID,instanceID));
}


// ###### Add InstanceID to be deleted ######################################
void SCTPSocketMaster::delayedDeleteSocket(const unsigned short instanceID)
{
#ifdef PRINT_GC
   cout << "delayedDeleteSocket: I=" << instanceID << endl;
#endif
   ClosingSockets.insert(instanceID);
}


// ###### Try to delete instance ############################################
void SCTPSocketMaster::socketGarbageCollection()
{
#ifdef PRINT_GC
   cout << "Socket garbage collection..." << endl;
#endif
   MasterInstance.lock();
   LastGarbageCollection = getMicroTime();

   // ====== Try to auto-close connectionless associations ==================
   multimap<unsigned short, SCTPSocket*>::iterator socketIterator = SocketList.begin();
   while(socketIterator != SCTPSocketMaster::SocketList.end()) {
      SCTPSocket* socket = socketIterator->second;
      socket->checkAutoClose();
      socketIterator++;
   }

   // ====== Try to delete sockets already removed sockets using sctplib ====
   set<unsigned short>::iterator iterator = ClosingSockets.begin();
   while(iterator != ClosingSockets.end()) {
      const unsigned short instanceID = *iterator;

      bool used = false;
      multimap<unsigned int, unsigned short>::iterator assocIterator = ClosingAssociations.begin();
      while(assocIterator != ClosingAssociations.end()) {
         if(instanceID == assocIterator->second) {
            used = true;
            break;
         }
         assocIterator++;
      }

      if(!used) {
#ifdef PRINT_GC
         cout << "socketGarbageCollection: Removing instance #" << instanceID << "." << endl;
#endif
         iterator++;
         ClosingSockets.erase(instanceID);
         sctp_unregisterInstance(instanceID);
      }
      else {
         iterator++;
      }
   }

   MasterInstance.unlock();
#ifdef PRINT_GC
   cout << "Socket garbage collection completed in "
        << getMicroTime() - LastGarbageCollection << " s" << endl;
#endif
}


// ###### Try to delete association and instance ############################
bool SCTPSocketMaster::associationGarbageCollection(const unsigned int assocID,
                                                    const bool         sendAbort)
{
   // ====== Delayed removal ================================================
   multimap<unsigned int, unsigned short>::iterator iterator = ClosingAssociations.find(assocID);
   if(iterator != ClosingAssociations.end()) {
      const unsigned short instanceID = iterator->second;
#ifdef PRINT_GC
      cout << "associationGarbageCollection: Removing association #" << assocID << "." << endl;
#endif

      // ====== Delete association ==========================================
      if(sendAbort) {
         sctp_abort(assocID);
      }
      sctp_deleteAssociation(assocID);
      ClosingAssociations.erase(iterator);

      // ====== Check, if instance can be deleted ===========================
      if(ClosingSockets.find(instanceID) != ClosingSockets.end()) {
         // ====== Check, if instance is still in use =======================
         bool deleteInstance = true;
         iterator = ClosingAssociations.begin();
         while(iterator != ClosingAssociations.end()) {
            if(iterator->second == instanceID) {
#ifdef PRINT_GC
               cout << "associationGarbageCollection: Instance #" << instanceID
                    << " is still in use -> no unregistering." << endl;
#endif
               deleteInstance = false;
               break;
            }
            iterator++;
         }

         // ====== Delete instance ==========================================
         if(deleteInstance) {
#ifdef PRINT_GC
            cout << "associationGarbageCollection: Removing instance #" << instanceID << "." << endl;
#endif
            ClosingSockets.erase(instanceID);
            sctp_unregisterInstance(instanceID);
         }
      }

      return(true);
   }
   return(false);
}


// ###### SCTP data arrive notification callback ############################
void SCTPSocketMaster::dataArriveNotif(unsigned int assocID,
                                       unsigned int streamID,
                                       unsigned int length,
                                       unsigned int protoID,
                                       unsigned int unordered,
                                       void*        ulpDataPtr)
{
#ifdef PRINT_ARRIVENOTIFICATION
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d S%02d: Data Arrive Notification, Length = %d",
               assocID,streamID,length);
   cerr << str << endl;
#endif


   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,streamID);
      sctp_data_arrive* sda = &notification.Content.sn_data_arrive;
      sda->sda_type          = SCTP_DATA_ARRIVE;
      sda->sda_flags         = (unordered == 1) ? SCTP_ARRIVE_UNORDERED : 0;
      sda->sda_length        = sizeof(sctp_data_arrive);
      sda->sda_assoc_id      = assocID;
      sda->sda_stream        = streamID;
      sda->sda_ppid          = protoID;
      sda->sda_bytes_arrived = length;
      addNotification(socket,assocID,notification);
   }
}


// ###### SCTP send failure notification callback ###########################
void SCTPSocketMaster::sendFailureNotif(unsigned int   assocID,
                                        unsigned char* unsent_data,
                                        unsigned int   dataLength,
                                        unsigned int*  context,
                                        void*          dummy)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Send Failure Notification",assocID);
   cerr << str << endl;
#endif


   // ====== Generate "Send Failure" notification ===========================
   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,0);
      sctp_send_failed* ssf = &notification.Content.sn_send_failed;
      ssf->ssf_type     = SCTP_REMOTE_ERROR;
      ssf->ssf_flags    = 0;
      ssf->ssf_length   = sizeof(sctp_send_failed);
      ssf->ssf_error    = 0;
      ssf->ssf_assoc_id = assocID;
      ssf->ssf_info.sinfo_stream     = 0;
      ssf->ssf_info.sinfo_ssn        = 0;
      ssf->ssf_info.sinfo_flags      = 0;
      ssf->ssf_info.sinfo_ppid       = 0;
      ssf->ssf_info.sinfo_context    = 0;
      ssf->ssf_info.sinfo_timetolive = 0;
      ssf->ssf_info.sinfo_assoc_id   = assocID;
      // ??? Sufficient space within SCTPNotification ???
      // memcpy((char*)&ssf->ssf_data,(char*)unsent_data,dataLength);
      addNotification(socket,assocID,notification);
   }
}


// ###### SCTP network status change notification callback ##################
void SCTPSocketMaster::networkStatusChangeNotif(unsigned int   assocID,
                                                short          destAddrIndex,
                                                unsigned short newState,
                                                void*          ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Network Status Change",assocID);
   cerr << str << endl;
#endif


   // ====== Select new primary path, if it has become inactive ============
   SCTP_AssociationStatus assocStatus;
   SCTP_PathStatus        pathStatus;
   if((newState == SCTP_PATH_UNREACHABLE) && (destAddrIndex == sctp_getPrimary(assocID))) {
      unsigned short         pathID;
      sctp_getAssocStatus(assocID,&assocStatus);
      for(pathID = 0;pathID < assocStatus.numberOfAddresses;pathID++) {
         sctp_getPathStatus(assocID,pathID,&pathStatus);
         if(pathStatus.state == SCTP_PATH_OK) {
            break;
         }
      }
      if(pathID < assocStatus.numberOfAddresses) {
         sctp_setPrimary(assocID,pathID);
      }
   }

   // ====== Get new destination address ====================================
   SocketAddress* destination = NULL;
   if(pathStatus.state == SCTP_PATH_OK) {
      destination = SocketAddress::createSocketAddress(
                                      SocketAddress::PF_HidePort,
                                      (char*)&pathStatus.destinationAddress);
   }

   // ====== Generate "Network Status Change" notification ==================
   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,0);
      sctp_paddr_change* spc = &notification.Content.sn_paddr_change;
      spc->spc_type     = SCTP_PEER_ADDR_CHANGE;
      spc->spc_flags    = 0;
      spc->spc_error    = 0;
      spc->spc_length   = sizeof(sctp_paddr_change);
      if(destination) {
         spc->spc_state  = ADDRESS_MADE_PRIM;
         destination->getSystemAddress((sockaddr*)&spc->spc_aaddr,sizeof(sockaddr_storage));
      }
      else {
         spc->spc_state  = ADDRESS_UNREACHABLE;
      }
      addNotification(socket,assocID,notification);
      if(destination) {
         delete destination;
      }
   }
}


// ###### SCTP communication up notification callback #######################
void* SCTPSocketMaster::communicationUpNotif(unsigned int   assocID,
                                             unsigned short status,
                                             int            noOfDestinations,
                                             unsigned short noOfInStreams,
                                             unsigned short noOfOutStreams,
                                             void*          dummy)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Up Notification",assocID);
   cerr << str << endl;
#endif

   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,0);

      // ====== Successfull associate() execution ===========================
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID);
      if(association != NULL) {
         association->CommunicationUpNotification = true;
         association->EstablishCondition.broadcast();
      }
      // ====== Incoming connection =========================================
      else {
         association = new SCTPAssociation(socket,assocID,socket->NotificationFlags);
         if(association != NULL) {
            association->CommunicationUpNotification = true;
            SCTPSocket::IncomingConnection* newConnection = new SCTPSocket::IncomingConnection;
            if(newConnection != NULL) {
               newConnection->NextConnection = NULL;
               newConnection->Association    = association;
               newConnection->Notification   = notification;

               if(socket->ConnectionRequests == NULL) {
                  socket->ConnectionRequests = newConnection;
               }
               else {
                  SCTPSocket::IncomingConnection* c = socket->ConnectionRequests;
                  while(c->NextConnection != NULL) {
                     c = c->NextConnection;
                  }
                  c->NextConnection = newConnection;
               }
               socket->EstablishCondition.broadcast();
            }
         }
      }


      // ====== Generate "Communication Up" notification ====================
      if(association != NULL) {
         sctp_assoc_change* sac = &notification.Content.sn_assoc_change;
         sac->sac_type   = SCTP_ASSOC_CHANGE;
         sac->sac_flags  = 0;
         sac->sac_length = sizeof(sctp_assoc_change);
         sac->sac_state  = COMMUNICATION_UP;
         sac->sac_error  = 0;
         sac->sac_outbound_streams = noOfOutStreams;
         sac->sac_inbound_streams  = noOfInStreams;
         sac->sac_assoc_id         = assocID;
         addNotification(socket,assocID,notification);
      }
   }

   return(NULL);
}


// ###### SCTP communication lost notification callback #####################
void SCTPSocketMaster::communicationLostNotif(unsigned int   assocID,
                                              unsigned short status,
                                              void*          ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Lost Notification, Status=%d",assocID,status);
   cerr << str << endl;
#endif


   // ====== Delayed removal ================================================
   if(associationGarbageCollection(assocID,false)) {
      return;
   }


   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         // ====== Notify waiting main thread ===============================
         association->CommunicationLostNotification = true;
         association->ShutdownCompleteNotification  = true;
         association->ShutdownCompleteCondition.broadcast();


         // ====== Generate "Communication Lost" notification ===============
         SCTPNotification notification;
         initNotification(notification);
         sctp_assoc_change* sac = &notification.Content.sn_assoc_change;
         sac->sac_type             = SCTP_ASSOC_CHANGE;
         sac->sac_flags            = 0;
         sac->sac_length           = sizeof(sctp_assoc_change);
         sac->sac_state            = COMMUNICATION_LOST;
         sac->sac_error            = 0;
         sac->sac_outbound_streams = 0;
         sac->sac_inbound_streams  = 0;
         sac->sac_assoc_id         = assocID;
         addNotification(socket,assocID,notification);

         // If association() is waiting for connection establishment,
         // send signal abort.
         association->EstablishCondition.broadcast();
         association->ReadyForTransmit.broadcast();
      }
   }
}


// ###### SCTP communication error notification callback ####################
void SCTPSocketMaster::communicationErrorNotif(unsigned int   assocID,
                                               unsigned short status,
                                               void*          dummy)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Error Notification, Status=%d",assocID,status);
   cerr << str << endl;
#endif


   // ====== Generate "Communication Error" notification ====================
   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,0);
      sctp_remote_error* sre = &notification.Content.sn_remote_error;
      sre->sre_type      = SCTP_REMOTE_ERROR;
      sre->sre_flags     = 0;
      sre->sre_length    = sizeof(sctp_remote_error);
      sre->sre_error     = 0;
      sre->sre_assoc_id  = assocID;
      addNotification(socket,assocID,notification);
   }
}


// ###### SCTP restart notification callback ################################
void SCTPSocketMaster::restartNotif(unsigned int assocID, void* ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Restart Notification",assocID);
   cerr << str << endl;
#endif


   // ====== Generate "Restart" notification ==================================
   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPNotification notification;
      initNotification(notification,assocID,0);
      sctp_assoc_change* sac = &notification.Content.sn_assoc_change;
      sac->sac_type             = SCTP_ASSOC_CHANGE;
      sac->sac_flags            = 0;
      sac->sac_length           = sizeof(sctp_assoc_change);
      sac->sac_state            = RESTART;
      sac->sac_error            = 0;
      SCTP_Association_Status status;
      if(sctp_getAssocStatus(assocID,&status) == 0) {
         sac->sac_outbound_streams = status.outStreams;
         sac->sac_inbound_streams  = status.inStreams;
      }
      else {
#ifndef DISABLE_WARNINGS
         cerr << "WARNING: SCTPSocketMaster::restartNotif() - sctp_getAssocStatus() failed!" << endl;
#endif
         sac->sac_outbound_streams = 1;
         sac->sac_inbound_streams  = 1;
      }
      sac->sac_assoc_id = assocID;
      addNotification(socket,assocID,notification);
   }
}


// ###### SCTP shutdown complete notification callback ######################
void SCTPSocketMaster::shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Shutdown Complete Notification",assocID);
   cerr << str << endl;
#endif

   // ====== Delayed removal ================================================
   if(associationGarbageCollection(assocID,false)) {
      return;
   }

   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         // ====== Notify waiting main thread ===============================
         association->ShutdownCompleteNotification = true;
         association->ShutdownCompleteCondition.broadcast();
         association->ReadyForTransmit.broadcast();

         // ====== Generate "Shutdown Complete" notification ================
         SCTPNotification notification;
         initNotification(notification);
         sctp_shutdown_event* sse = &notification.Content.sn_shutdown_event;
         sse->sse_type            = SCTP_SHUTDOWN_EVENT;
         sse->sse_flags           = 0;
         sse->sse_length          = sizeof(sctp_shutdown_event);
         sse->sse_assoc_id        = assocID;
         addNotification(socket,assocID,notification);
      }
   }
}


// ###### SCTP queue status change notification callback ######################
void SCTPSocketMaster::queueStatusChangeNotif(unsigned int assocID,
                                              int          queueType,
                                              int          queueIdentifier,
                                              int          queueLength,
                                              void*        ulpData)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Queue Status Change Notification, Qid=%d Qtype=%d Qlength=%d",
               assocID,queueIdentifier,queueType,queueLength);
   cerr << str << endl;
#endif

   SCTPSocket* socket = getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         association->ReadyForTransmit.broadcast();
      }
   }
}


// ###### SCTP asconf status change notification callback #####################
void SCTPSocketMaster::asconfStatusNotif(unsigned int assocID,
                                         unsigned int correlationID,
                                         int          result,
                                         void*        request,
                                         void*        ulpData)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: ASConf Status Change Notification, Cid=%d result=%d",
               assocID,correlationID,result);
   cerr << str << endl;
#endif
}


// ###### User callback #####################################################
void SCTPSocketMaster::userCallback(int       fileDescriptor,
                                    short int eventMask,
                                    void*     userData)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "F%04d: User Callback, mask=$%x",
               fileDescriptor,eventMask);
   cerr << str << endl;
#endif
   UserSocketNotification* usn = (UserSocketNotification*)userData;
   if(userData != NULL) {
      usn->Mask |= eventMask;
      usn->UpdateCondition.broadcast();
      sctp_unregisterUserCallback(usn->FileDescriptor);
   }
}


// ###### Get socket for given association ID ###############################
SCTPSocket* SCTPSocketMaster::getSocketForAssociationID(const unsigned int assocID)
{
   unsigned short instanceID = 0;
   if(sctp_getInstanceID(assocID,&instanceID) == 0) {
      if(instanceID != 0) {
         multimap<unsigned short, SCTPSocket*>::iterator iterator =
            SocketList.find(instanceID);
         if(iterator != SocketList.end()) {
            return(iterator->second);
         }
      }
#ifdef DEBUG
      else {
        cerr << "WARNING: SCTPSocketMaster::getSocketForAssociationID() - "
                "Instance ID for association ID #" << assocID << " is zero!" << endl;
      }
#endif
   }
#ifdef DEBUG
   else {
      cerr << "WARNING: SCTPSocketMaster::getSocketForAssociationID() - "
              "No instance ID for association ID #" << assocID << " found!" << endl;
   }
#endif
   return(NULL);
}


// ###### Clear SCTPNotification structure ##################################
void SCTPSocketMaster::initNotification(SCTPNotification& notification)
{
   notification.Content.sn_type = SCTP_UNDEFINED;
   notification.ContentPosition = 0;
   notification.RemotePort      = 0;
   notification.RemoteAddresses = 0;
   for(unsigned int i = 0;i < SCTP_MAX_NUM_ADDRESSES;i++) {
      notification.RemoteAddress[i][0] = 0x00;
   }
}


// ###### Initialize SCTPNotification structure #############################
bool SCTPSocketMaster::initNotification(SCTPNotification& notification,
                                        unsigned int      assocID,
                                        unsigned short    streamID)
{
   notification.Content.sn_type = SCTP_UNDEFINED;
   notification.ContentPosition = 0;
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(assocID,&status) == 0) {
      notification.RemotePort      = status.destPort;
      notification.RemoteAddresses = min((unsigned short)SCTP_MAX_NUM_ADDRESSES,
                                         status.numberOfAddresses);
      for(unsigned int i = 0;i < notification.RemoteAddresses;i++) {
         SCTP_Path_Status pathStatus;
         if(sctp_getPathStatus(assocID,i,&pathStatus) != 0) {
#ifndef DISABLE_WARNINGS
            cerr << "WARNING: SCTPSocketMaster::initNotification() - sctp_getPathStatus() failure!"
                 << endl;
#endif

         }
         else {
            memcpy((char*)&notification.RemoteAddress[i],
                   (char*)&pathStatus.destinationAddress,
                   sizeof(pathStatus.destinationAddress));
         }
      }
      return(true);
   }
#ifndef DISABLE_WARNINGS
         cerr << "WARNING: SCTPSocketMaster::initNotification() - sctp_getAssocStatus() failure!"
              << endl;
#endif
   return(false);
}


// ###### Add SCTPNotification structure to socket's queue ##################
void SCTPSocketMaster::addNotification(SCTPSocket*             socket,
                                       unsigned int            assocID,
                                       const SCTPNotification& notification)
{
   // ====== Get notification flags =========================================
   SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
   if(association == NULL) {
      // Association not found -> already closed.
      return;
   }
   const unsigned int notificationFlags = association->NotificationFlags;

   // ====== Check, if notification has to be added =========================
   const sctp_notification_header* header = &notification.Content.sn_notification_header;
   if((header->snh_type == SCTP_DATA_ARRIVE)    ||
      (header->snh_type == SCTP_SHUTDOWN_EVENT) ||
      ((header->snh_type == SCTP_ASSOC_CHANGE) && (notification.Content.sn_assoc_change.sac_state == COMMUNICATION_LOST)) ||
      ((header->snh_type == SCTP_ASSOC_CHANGE)     && (notificationFlags & SCTP_RECVASSOCEVNT))    ||
      ((header->snh_type == SCTP_PEER_ADDR_CHANGE) && (notificationFlags & SCTP_RECVPADDREVNT))    ||
      ((header->snh_type == SCTP_REMOTE_ERROR)     && (notificationFlags & SCTP_RECVPEERERR))      ||
      ((header->snh_type == SCTP_SEND_FAILED)      && (notificationFlags & SCTP_RECVSENDFAILEVNT)) ||
      ((header->snh_type == SCTP_SHUTDOWN_EVENT)   && (notificationFlags & SCTP_RECVSHUTDOWNEVNT))) {

      // ====== Add notification to global or association's queue ===========
#ifdef PRINT_ASSOC_USECOUNT
      cout << "AddNotification: UseCount increment for A" << association->getID() << ": "
           << association->UseCount << " -> ";
#endif
      association->UseCount++;
#ifdef PRINT_ASSOC_USECOUNT
      cout << association->UseCount << ". Notification Type = " << header->snh_type << endl;
#endif
      if(socket->Flags & SCTPSocket::SSF_GlobalQueue) {
         socket->GlobalQueue.addNotification(notification);
      }
      else {
         association->InQueue.addNotification(notification);
      }
   }
}


// ###### Add user socket notification ######################################
void SCTPSocketMaster::addUserSocketNotification(UserSocketNotification* usn)
{
   lock();
   usn->Mask = 0;
   const int result =
      sctp_registerUserCallback(usn->FileDescriptor,userCallback,(void*)usn);
   if(result < 0) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPSocketMaster::addUserSocketNotification() - sctp_registerUserCallback() failed!" << endl;
#endif
   }
   unlock();
}


// ###### Delete user socket notification ###################################
void SCTPSocketMaster::deleteUserSocketNotification(UserSocketNotification* usn)
{
   lock();
   sctp_unregisterUserCallback(usn->FileDescriptor);
   unlock();
}


// ###### Lock ##############################################################
void SCTPSocketMaster::lock()
{
   const cardinal oldstate = Thread::setCancelState(Thread::TCS_CancelDisabled);
   synchronized();
   LockLevel++;
   if(LockLevel == 1) {
      OldCancelState = oldstate;
   }
}


// ###### Unlock ############################################################
void SCTPSocketMaster::unlock()
{
#ifndef DISABLE_WARNINGS
   if(LockLevel == 0) {
      cerr << "INTERNAL ERROR: Too many calls to SCTPSocketMaster::unlock()!" << endl;
      ::exit(1);
   }
#endif
   LockLevel--;
   const cardinal lockLevel = LockLevel;
   unsynchronized();
   if(lockLevel == 0) {
      Thread::setCancelState(OldCancelState);
   }
}


}
