/*
 *  $Id: dhcpclient.cc,v 1.2 2001/11/29 12:25:22 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: DHCP Client
 *
 */


#include "tdsystem.h"
#include "dhcpclient.h"

#include <packetaddress.h>

#include <sys/ioctl.h>
#include <net/ethernet.h>
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <linux/if.h>


// Verbose printing of configuration information
#define VERBOSE



namespace Coral {


// ###### Static data #######################################################
DHCPClient::ReplyReaderThread         DHCPClient::ReplyReader;
set<DHCPClient::DHCPClientInterface*> DHCPClient::InterfaceSet;
Socket     DHCPClient::InputSocket;
Socket     DHCPClient::OutputSocket;
Randomizer DHCPClient::Random;
bool       DHCPClient::Ready       = false;
bool       DHCPClient::Initialized = false;

const InternetAddress DHCPClient::NullAddress("0.0.0.0:0");
const InternetAddress DHCPClient::BroadcastAddress("255.255.255.255:67");
const InternetAddress DHCPClient::LocalOutputAddress(67);
const InternetAddress DHCPClient::LocalInputAddress(68);


// ###### Constructor #######################################################
DHCPClient::ReplyReaderThread::ReplyReaderThread()
{
}


// ###### Destructor ########################################################
DHCPClient::ReplyReaderThread::~ReplyReaderThread()
{
}


// ###### Reply receiver thread's run() implementation ######################
void DHCPClient::ReplyReaderThread::run()
{
   DHCPConfiguration config;
   for(;;) {
      const cardinal type = Client->receiveReply(config);
      if((type == DHCPOFFER) || (type == DHCPACK)) {
         Client->synchronized();

         set<DHCPClientInterface*>::iterator iterator = InterfaceSet.begin();
         while(iterator != InterfaceSet.end()) {
            DHCPClientInterface* interface = *iterator;
            const card64         now       = getMicroTime();
            if(interface->TransactionID == config.TransactionID) {
               interface->Config = config;

               // ====== DHCP offer =========================================
               if((type == DHCPOFFER) &&
                  ((interface->State == DHCPClientInterface::DIS_FastDiscovery) ||
                   (interface->State == DHCPClientInterface::DIS_SlowDiscovery))) {
                  interface->State      = DHCPClientInterface::DIS_Requesting;
                  interface->Trial      = 0;
                  interface->NextAction = 0;  // Restart ASAP!
#ifdef VERBOSE
                     cout << "Offer for " << interface->Name << ":" << endl << interface->Config;
                     cout << interface->Name << ": Discovery -> Requesting" << endl;
#endif
                  Client->setNextActionAbs(0,min(Client->getInterval(0),interface->NextAction));
                  break;
               }

               // ====== DHCP acknowledgement ==================================
               else if((type == DHCPACK) &&
                       ((interface->State == DHCPClientInterface::DIS_Requesting) ||
                        (interface->State == DHCPClientInterface::DIS_Renewing))) {
                  const cardinal oldState = interface->State;
                  interface->State      = DHCPClientInterface::DIS_Up;
                  interface->NextAction = now + interface->UpTimeout;
#ifdef VERBOSE
                  cout << "Acknowledgement for " << interface->Name << ":" << endl << interface->Config;
                  if(oldState == DHCPClientInterface::DIS_Renewing) {
                     cout << interface->Name << ": Renewing -> Up" << endl;
                  }
                  else {
                     cout << interface->Name << ": Requesting -> Up" << endl;
                  }
#endif
                  if(Client->configureInterface(interface) == false) {
                     interface->State = DHCPClientInterface::DIS_Down;
                     interface->NextAction = now + interface->DownTimeout;
#ifdef VERBOSE
                     cout << interface->Name << ": Up -> Down" << endl;
#endif
                  }
                  Client->setNextActionAbs(0,min(Client->getInterval(0),interface->NextAction));
                  break;
               }

               // ====== DHCP negative acknowledgement ======================
               else if((type == DHCPNAK) &&
                       ((interface->State == DHCPClientInterface::DIS_Requesting) ||
                        (interface->State == DHCPClientInterface::DIS_Renewing))) {
                  interface->State      = DHCPClientInterface::DIS_Down;
                  interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
                  if(interface->State == DHCPClientInterface::DIS_Renewing) {
                     cout << interface->Name << ": Renewing -> Down" << endl;
                  }
                  else {
                     cout << interface->Name << ": Requesting -> Down" << endl;
                  }
#endif
                  Client->setNextActionAbs(0,min(Client->getInterval(0),interface->NextAction));
               }

            }
            iterator++;
         }

         Client->unsynchronized();
      }
   }
}


// ###### Constructor #######################################################
DHCPClient::DHCPClient()
{
   if(Initialized) {
      cerr << "ERROR: DHCPClient::DHCPClient() - DHCPClient singleton initialized twice!" << endl;
      return;
   }
   Initialized = true;

   if(OutputSocket.create(AF_PACKET,SOCK_PACKET,htons((short)ETH_P_ALL)) == false) {
      cerr << "ERROR: DHCPClient::DHCPClient() - Unable to create raw output socket!" << endl;
      return;
   }
   if(OutputSocket.setSoBroadcast(true) == false) {
      cerr << "ERROR: DHCPClient::DHCPClient() - Unable to enable socket's broadcast mode!" << endl;
      return;
   }

   if(InputSocket.create(Socket::IPv4,Socket::Datagram,0) == false) {
      cerr << "ERROR: DHCPClient::DHCPClient() - Unable to create input socket!" << endl;
   }
   if(InputSocket.setSoBroadcast(true) == false) {
      cerr << "ERROR: DHCPClient::DHCPClient() - Unable to enable socket's broadcast mode!" << endl;
      return;
   }
   if(InputSocket.bind(LocalInputAddress) == false) {
      cerr << "ERROR: DHCPClient::DHCPClient() - Unable to bind socket to " << LocalInputAddress << "!" << endl;
      return;
   }

   setNextActionAbs(0,0);
   setTimerCorrection(0,false);
   Ready = true;
}


// ###### Destructor ########################################################
DHCPClient::~DHCPClient()
{
}


// ###### Start DHCP client #################################################
bool DHCPClient::start(const char* name)
{
   if(Ready == true) {
      ReplyReader.Client = this;
      ReplyReader.start();
      return(SingleTimerThread::start());
   }
   return(false);
}


// ###### Stop DHCP client ##################################################
void* DHCPClient::stop()
{
   ReplyReader.stop();
   return(SingleTimerThread::stop());
}


// ###### Add interface #####################################################
bool DHCPClient::addInterface(const char*    name,
                              const card64   downTimeout,
                              const card64   fastDiscoveryTimeout,
                              const cardinal fastDiscoveryTrials,
                              const card64   slowDiscoveryTimeout,
                              const cardinal slowDiscoveryTrials,
                              const card64   requestTimeout,
                              const cardinal requestTrials,
                              const card64   upTimeout,
                              const card64   renewTimeout,
                              const cardinal renewTrials)
{
   DHCPClientInterface* interface = new DHCPClientInterface;
   if((interface != NULL) && (findInterface(name) == NULL)) {
      interface->State                = DHCPClientInterface::DIS_Down;
      interface->Trial                = 0;
      interface->NextAction           = 0;
      interface->TransactionID        = 0;

      interface->DownTimeout          = downTimeout;
      interface->FastDiscoveryTimeout = fastDiscoveryTimeout;
      interface->FastDiscoveryTrials  = fastDiscoveryTrials;
      interface->SlowDiscoveryTimeout = slowDiscoveryTimeout;
      interface->SlowDiscoveryTrials  = slowDiscoveryTrials;
      interface->RequestTimeout       = requestTimeout;
      interface->RequestTrials        = requestTrials;
      interface->UpTimeout            = upTimeout;
      interface->RenewTimeout         = renewTimeout;
      interface->RenewTrials          = renewTrials;
      interface->DesiredLeaseTime     = (cardinal)((2 * interface->UpTimeout) / (card64)1000000);

      snprintf((char*)&interface->Name,sizeof(interface->Name),"%s",name);
      synchronized();
      InterfaceSet.insert(interface);
      setNextActionAbs(0,0);
      unsynchronized();
      return(true);
   }
   return(false);
}


// ###### Find interface ####################################################
DHCPClient::DHCPClientInterface* DHCPClient::findInterface(const char* name)
{
   DHCPClientInterface* interface = NULL;
   synchronized();
   set<DHCPClientInterface*>::iterator iterator = InterfaceSet.begin();
   while(iterator != InterfaceSet.end()) {
      if(!(strcmp((*iterator)->Name,name))) {
         interface = *iterator;
         break;
      }
      iterator++;
   }
   unsynchronized();
   return(interface);
}


// ###### Delete interface ##################################################
bool DHCPClient::deleteInterface(const char* name)
{
   synchronized();
   DHCPClientInterface* interface = findInterface(name);
   if(interface != NULL) {
      InterfaceSet.erase(interface);
      delete interface;
   }
   unsynchronized();
   return(interface != NULL);
}


// ###### Start interface configuration by DHCP #############################
bool DHCPClient::startInterface(const char* name)
{
   synchronized();
   DHCPClientInterface* interface = findInterface(name);
   if(interface != NULL) {
      interface->State      = DHCPClientInterface::DIS_Down;
      interface->Trial      = 0;
      interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
      cout << interface->Name << ": DHCP configuration started!" << endl;
#endif
      setNextActionAbs(0,interface->NextAction);
   }
   unsynchronized();
   return(interface != NULL);;
}


// ###### Stop interface configuration by DHCP ##############################
bool DHCPClient::stopInterface(const char* name)
{
   synchronized();
   DHCPClientInterface* interface = findInterface(name);
   if(interface != NULL) {
      interface->State      = DHCPClientInterface::DIS_UserShutdown;
      interface->Trial      = 0;
      interface->NextAction = (card64)-1;
#ifdef VERBOSE
      cout << interface->Name << ": DHCP configuration stopped!" << endl;
#endif
      configureInterface(interface);
   }
   unsynchronized();
   return(interface != NULL);
}


// ###### DHCP client timer thread's timerEvent() implementation ############
void DHCPClient::timerEvent(const cardinal timer)
{
   card64 now  = getMicroTime();
   card64 next = now + IdleTimeout;

   synchronized();
   set<DHCPClientInterface*>::iterator iterator = InterfaceSet.begin();
   while(iterator != InterfaceSet.end()) {
      DHCPClientInterface* interface = *iterator;
      now = getMicroTime();
      if(interface->NextAction > now) {
         iterator++;
         continue;
      }

      switch(interface->State) {
         // ====== Shutdown by user =========================================
         case DHCPClientInterface::DIS_UserShutdown:
             interface->State      = DHCPClientInterface::DIS_Down;
             interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
             cout << interface->Name << ": UserShutdown -> Down" << endl;
#endif
          break;

         // ====== Interface is down ========================================
         case DHCPClientInterface::DIS_Down:
             interface->TransactionID = Random.random32();
             interface->State      = DHCPClientInterface::DIS_FastDiscovery;
             interface->Trial      = 0;
             interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
             cout << interface->Name << ": Down -> FastDiscovery" << endl;
#endif
             configureInterface(interface);
          break;

         // ====== Fast discovery mode ======================================
         case DHCPClientInterface::DIS_FastDiscovery:
             interface->Trial++;
             if(sendDiscovery(interface,interface->TransactionID) == true) {
                if(interface->Trial >= interface->FastDiscoveryTrials) {
                   interface->State      = DHCPClientInterface::DIS_SlowDiscovery;
                   interface->Trial      = 1;
                   interface->NextAction = now + interface->SlowDiscoveryTimeout;
#ifdef VERBOSE
                   cout << interface->Name << ": FastDiscovery -> SlowDiscovery" << endl;
#endif
                }
                else {
                   interface->NextAction = now + interface->FastDiscoveryTimeout;
#ifdef VERBOSE
                   cout << interface->Name << ": FastDiscovery, Trial " << interface->Trial << endl;
#endif
                }
             }
             else {
                interface->State      = DHCPClientInterface::DIS_Down;
                interface->NextAction = now + interface->DownTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": FastDiscovery -> Down" << endl;
#endif
                configureInterface(interface);
             }
          break;

         // ====== Slow discovery mode ======================================
         case DHCPClientInterface::DIS_SlowDiscovery:
             interface->Trial++;
             if((interface->Trial <= interface->SlowDiscoveryTrials) &&
                (sendDiscovery(interface,interface->TransactionID) == true)) {
                interface->NextAction = now + interface->SlowDiscoveryTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": SlowDiscovery, Trial " << interface->Trial << endl;
#endif
             }
             else {
                interface->State      = DHCPClientInterface::DIS_Down;
                interface->NextAction = now + interface->DownTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": SlowDiscovery -> Down" << endl;
#endif
                configureInterface(interface);
             }
          break;

         // ====== Requesting mode ==========================================
         case DHCPClientInterface::DIS_Requesting:
             interface->Trial++;
             if((interface->Trial <= interface->RequestTrials) &&
                (sendRequest(interface,interface->Config) == true)) {
                interface->NextAction = now + interface->RequestTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": Requesting, Trial " << interface->Trial << endl;
#endif
             }
             else {
                interface->State      = DHCPClientInterface::DIS_Down;
                interface->NextAction = now + interface->DownTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": Requesting -> Down" << endl;
#endif
                configureInterface(interface);
             }
          break;

         // ====== Interface is up ==========================================
         case DHCPClientInterface::DIS_Up:
             interface->State      = DHCPClientInterface::DIS_Renewing;
             interface->Trial      = 0;
             interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
             cout << interface->Name << ": Up -> Renewing" << endl;
#endif
          break;

         // ====== Renewing mode ============================================
         case DHCPClientInterface::DIS_Renewing:
             interface->Trial++;
             if((interface->Trial <= interface->RenewTrials) &&
                (sendRequest(interface,interface->Config) == true)) {
                interface->NextAction = now + interface->RenewTimeout;
#ifdef VERBOSE
                cout << interface->Name << ": Renewing, Trial " << interface->Trial << endl;
#endif
             }
             else {
                interface->State      = DHCPClientInterface::DIS_Down;
                interface->NextAction = 0;   // Restart ASAP!
#ifdef VERBOSE
                cout << interface->Name << ": Renewing -> Down" << endl;
#endif
                configureInterface(interface);
             }
          break;

      }

      next = min(next,interface->NextAction);
      iterator++;
   }
   unsynchronized();

   setNextActionAbs(0,next);
}


// ###### Send DHCP Discovery packet ########################################
bool DHCPClient::sendDiscovery(DHCPClientInterface* interface,
                               card32&              transactionID)
{
   // ====== Get hardware address ===========================================
   NetworkInterface networkInterface(interface->Name);
   sockaddr hardwareAddress;
   if(networkInterface.getHardwareAddress(hardwareAddress) == false) {
      return(false);
   }
   if(hardwareAddress.sa_family != HTYPE_ETHER) {
      return(false);
   }


   // ====== Prepare packet DHCP/UDP/IP/Ethernet ============================
   RawDHCPPacket packet;
   packet.Ethernet.ether_type = htons(ETHERTYPE_IP);
   for(cardinal i = 0; i < 6;i++) {
      packet.Ethernet.ether_shost[i] = hardwareAddress.sa_data[i];
   }
   packet.Ethernet.ether_dhost[0] = 0xff;
   packet.Ethernet.ether_dhost[1] = 0xff;
   packet.Ethernet.ether_dhost[2] = 0xff;
   packet.Ethernet.ether_dhost[3] = 0xff;
   packet.Ethernet.ether_dhost[4] = 0xff;
   packet.Ethernet.ether_dhost[5] = 0xff;

   packet.IP.ip_v   = 4;
   packet.IP.ip_hl  = 5;
   packet.IP.ip_len = htons(sizeof(packet.IP) + sizeof(packet.UDP) + sizeof(packet.DHCP));
   packet.IP.ip_id  = htons(getpid());
   packet.IP.ip_tos = IPTOS_PREC_NETCONTROL;
   packet.IP.ip_ttl = 1;
   packet.IP.ip_p   = IPPROTO_UDP;
   packet.IP.ip_off = 0;
   InternetAddress::setIPv4Address(NullAddress,&packet.IP.ip_src);
   InternetAddress::setIPv4Address(BroadcastAddress,&packet.IP.ip_dst);
   packet.IP.ip_sum = 0;
   packet.IP.ip_sum = InternetAddress::wrapChecksum(
                         InternetAddress::calculateChecksum((unsigned char *)&packet.IP,sizeof(packet.IP),0));

   packet.UDP.dest   = htons(BroadcastAddress.getPort());
   packet.UDP.source = htons(LocalOutputAddress.getPort());
   packet.UDP.len    = htons(sizeof(packet.UDP) + sizeof(packet.DHCP));
   packet.UDP.check  = 0;


   transactionID = Random.random32();
   packet.DHCP.op    = BOOTREQUEST;
   packet.DHCP.htype = HTYPE_ETHER;
   packet.DHCP.hlen  = sizeof(packet.Ethernet.ether_shost);
   packet.DHCP.hops  = 0;
   packet.DHCP.xid   = transactionID;
   packet.DHCP.secs  = 0;
   packet.DHCP.flags = htons((u_int16_t)BOOTP_BROADCAST);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.ciaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.yiaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.siaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.giaddr);
   memcpy((char*)&packet.DHCP.chaddr,
          (const char*)&packet.Ethernet.ether_shost,
          sizeof(packet.Ethernet.ether_shost));
   for(cardinal j = 0;j < sizeof(packet.DHCP.sname);j++) {
      packet.DHCP.sname[j] = 0x00;
   }
   for(cardinal j = 0;j < sizeof(packet.DHCP.file);j++) {
      packet.DHCP.file[j]  = 0x00;
   }

   cardinal i = sizeof(DHCP_OPTIONS_COOKIE) - 1;  // Do not count 0x00!
   memcpy((char*)&packet.DHCP.options,DHCP_OPTIONS_COOKIE,i);
   packet.DHCP.options[i++] = DHO_DHCP_MESSAGE_TYPE;
   packet.DHCP.options[i++] = 1;
   packet.DHCP.options[i++] = DHCPDISCOVER;

   packet.DHCP.options[i++] = DHO_DHCP_PARAMETER_REQUEST_LIST;
   packet.DHCP.options[i++] = 7;
   packet.DHCP.options[i++] = DHO_SUBNET_MASK;
   packet.DHCP.options[i++] = DHO_TIME_OFFSET;
   packet.DHCP.options[i++] = DHO_DOMAIN_NAME_SERVERS;
   packet.DHCP.options[i++] = DHO_DOMAIN_NAME;
   packet.DHCP.options[i++] = DHO_HOST_NAME;
   packet.DHCP.options[i++] = DHO_ROUTERS;
   packet.DHCP.options[i++] = DHO_BROADCAST_ADDRESS;

   packet.DHCP.options[i++] = DHO_DHCP_LEASE_TIME;
   packet.DHCP.options[i++] = 4;
   card32 leaseTime = translate32((card32)interface->DesiredLeaseTime);
   memcpy((char*)&packet.DHCP.options[i],(const char*)&leaseTime,sizeof(leaseTime));
   i+= 4;

   packet.DHCP.options[i++] = DHO_END;
   for(   ;i < sizeof(packet.DHCP.options);i++) {
      packet.DHCP.options[i] = 0x00;
   }


   // ====== Calculate checksum =============================================
   packet.UDP.check =
      InternetAddress::wrapChecksum(
         InternetAddress::calculateChecksum((unsigned char *)&packet.UDP, sizeof(packet.UDP),
            InternetAddress::calculateChecksum((unsigned char*)&packet.DHCP, sizeof(packet.DHCP),
               InternetAddress::calculateChecksum((unsigned char *)&packet.IP.ip_src,2 * sizeof(packet.IP.ip_src),
                  IPPROTO_UDP + (card32)ntohs(packet.UDP.len))))
      );


   // ====== Send packet ====================================================
   const PacketAddress packetAddress(interface->Name);
   const ssize_t result = OutputSocket.sendTo((void*)&packet, sizeof(packet),
                                              0, packetAddress);
   return(result == sizeof(packet));
}


// ###### Convert receiver packet to configuration ##########################
bool DHCPClient::packetToConfig(const dhcp_packet* packet,
                                const cardinal     length,
                                DHCPConfiguration& config)
{
   config.TransactionID = packet->xid;
   config.Address       = InternetAddress::getIPv4Address(packet->yiaddr);
   config.Broadcast     = NullAddress;
   config.SubnetMask    = NullAddress;
   config.Router        = NullAddress;
   config.NameServer[0] = NullAddress;
   config.NameServer[1] = NullAddress;
   config.NameServer[2] = NullAddress;
   config.NameServers   = 0;
   config.Host          = String();
   config.Domain        = String();
   config.LeaseTime     = 0;
   config.MessageType   = 0;
   config.ServerID      = InternetAddress();
   config.Server        = InternetAddress::getIPv4Address(packet->siaddr);
   config.Server.setPort(67);

   cardinal p = sizeof(DHCP_OPTIONS_COOKIE) - 1;
   while((packet->options[p] != DHO_END) &&
         ((unsigned long)&packet->options[p] - (unsigned long)packet < length)) {

      const cardinal option = packet->options[p];
      if((unsigned long)&packet->options[p + 1] - (unsigned long)packet >= length) {
         break;
      }
      const cardinal size = packet->options[++p];
      if((unsigned long)&packet->options[p + size] - (unsigned long)packet >= length) {
         break;
      }
      p++;

      switch(option) {
         case DHO_DHCP_MESSAGE_TYPE:
            if(size >= 1) {
               config.MessageType = packet->options[p];
            }
          break;
         case DHO_SUBNET_MASK:
             if(size >= 4) {
                config.SubnetMask = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p]));
             }
          break;
         case DHO_DOMAIN_NAME_SERVERS:
             if(size >= 4) {
                config.NameServer[0] = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p]));
                config.NameServers = 1;
                if(size >= 8) {
                   config.NameServer[1] = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p + 4]));
                   config.NameServers = 2;
                   if(size >= 12) {
                      config.NameServer[2] = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p + 8]));
                      config.NameServers = 3;
                   }
                }
             }
          break;
         case DHO_DHCP_SERVER_IDENTIFIER:
             if(size >= 4) {
                config.ServerID = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p]));
             }
          break;
         case DHO_ROUTERS:
             if(size >= 4) {
                config.Router = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p]));
             }
          break;
         case DHO_BROADCAST_ADDRESS:
             if(size >= 4) {
                config.Broadcast = InternetAddress::getIPv4Address(*((in_addr*)&packet->options[p]));
             }
          break;
         case DHO_DHCP_LEASE_TIME:
             if(size >= 4) {
                config.LeaseTime = translate32(*((card32*)&packet->options[p]));
             }
          break;
         case DHO_HOST_NAME:
             config.Host = String((const char*)&packet->options[p],size);
          break;
         case DHO_DOMAIN_NAME:
             config.Domain = String((const char*)&packet->options[p],size);
          break;
      }
      p += size;
   }

   if((config.Server.isValid())         &&
      (config.ServerID.isValid())       &&
      (config.Router.isValid())         &&
      (config.Address.isValid())        &&
      (config.Broadcast.isValid())      &&
      (config.SubnetMask.isValid())     &&
      (config.NameServers > 0)          &&
      (config.Router.isValid())) {
      return(true);
   }
   return(false);
}


// ###### Send DHCP request #################################################
bool DHCPClient::sendRequest(DHCPClientInterface* interface,
                             DHCPConfiguration&   config)
{
   // ====== Get hardware address ===========================================
   NetworkInterface networkInterface(interface->Name);
   sockaddr hardwareAddress;
   if(networkInterface.getHardwareAddress(hardwareAddress) == false) {
      return(false);
   }
   if(hardwareAddress.sa_family != HTYPE_ETHER) {
      return(false);
   }


   // ====== Prepare packet DHCP/UDP/IP/Ethernet ============================
   RawDHCPPacket packet;
   packet.Ethernet.ether_type = htons(ETHERTYPE_IP);
   for(cardinal i = 0; i < 6;i++) {
      packet.Ethernet.ether_shost[i] = hardwareAddress.sa_data[i];
   }
   packet.Ethernet.ether_dhost[0] = 0xff;
   packet.Ethernet.ether_dhost[1] = 0xff;
   packet.Ethernet.ether_dhost[2] = 0xff;
   packet.Ethernet.ether_dhost[3] = 0xff;
   packet.Ethernet.ether_dhost[4] = 0xff;
   packet.Ethernet.ether_dhost[5] = 0xff;

   packet.IP.ip_v   = 4;
   packet.IP.ip_hl  = 5;
   packet.IP.ip_len = htons(sizeof(packet.IP) + sizeof(packet.UDP) + sizeof(packet.DHCP));
   packet.IP.ip_id  = htons(getpid());
   packet.IP.ip_tos = IPTOS_PREC_NETCONTROL;
   packet.IP.ip_ttl = 1;
   packet.IP.ip_p   = IPPROTO_UDP;
   packet.IP.ip_off = 0;
   InternetAddress::setIPv4Address(NullAddress,&packet.IP.ip_src);
   InternetAddress::setIPv4Address(BroadcastAddress,&packet.IP.ip_dst);
   packet.IP.ip_sum = 0;
   packet.IP.ip_sum = InternetAddress::wrapChecksum(
                         InternetAddress::calculateChecksum((unsigned char *)&packet.IP,sizeof(packet.IP),0));

   packet.UDP.dest   = htons(BroadcastAddress.getPort());
   packet.UDP.source = htons(LocalOutputAddress.getPort());
   packet.UDP.len    = htons(sizeof(packet.UDP) + sizeof(packet.DHCP));
   packet.UDP.check  = 0;

   packet.DHCP.op    = BOOTREQUEST;
   packet.DHCP.htype = HTYPE_ETHER;
   packet.DHCP.hlen  = sizeof(packet.Ethernet.ether_shost);
   packet.DHCP.hops  = 0;
   packet.DHCP.xid   = config.TransactionID;
   packet.DHCP.secs  = 0;
   packet.DHCP.flags = htons((u_int16_t)BOOTP_BROADCAST);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.ciaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.yiaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.siaddr);
   InternetAddress::setIPv4Address(NullAddress,&packet.DHCP.giaddr);
   memcpy((char*)&packet.DHCP.chaddr,
          (const char*)&packet.Ethernet.ether_shost,
          sizeof(packet.Ethernet.ether_shost));

   memset((char*)&packet.DHCP.sname,0,sizeof(packet.DHCP.sname));
   memset((char*)&packet.DHCP.file,0,sizeof(packet.DHCP.file));


   cardinal i = sizeof(DHCP_OPTIONS_COOKIE) - 1;  // Do not count 0x00!
   memcpy((char*)&packet.DHCP.options,DHCP_OPTIONS_COOKIE,i);

   packet.DHCP.options[i++] = DHO_DHCP_MESSAGE_TYPE;
   packet.DHCP.options[i++] = 1;
   packet.DHCP.options[i++] = DHCPREQUEST;

   packet.DHCP.options[i++] = DHO_DHCP_SERVER_IDENTIFIER;
   packet.DHCP.options[i++] = 4;
   InternetAddress::setIPv4Address(config.ServerID,(in_addr*)&packet.DHCP.options[i]);
   i += 4;

   packet.DHCP.options[i++] = DHO_DHCP_REQUESTED_ADDRESS;
   packet.DHCP.options[i++] = 4;
   InternetAddress::setIPv4Address(config.Address,(in_addr*)&packet.DHCP.options[i]);
   i += 4;

   packet.DHCP.options[i++] = DHO_DHCP_LEASE_TIME;
   packet.DHCP.options[i++] = 4;
   card32 leaseTime = translate32((card32)interface->DesiredLeaseTime);
   memcpy((char*)&packet.DHCP.options[i],(const char*)&leaseTime,sizeof(leaseTime));
   i+= 4;

   packet.DHCP.options[i++] = DHO_DHCP_PARAMETER_REQUEST_LIST;
   packet.DHCP.options[i++] = 7;
   packet.DHCP.options[i++] = DHO_SUBNET_MASK;
   packet.DHCP.options[i++] = DHO_TIME_OFFSET;
   packet.DHCP.options[i++] = DHO_DOMAIN_NAME_SERVERS;
   packet.DHCP.options[i++] = DHO_DOMAIN_NAME;
   packet.DHCP.options[i++] = DHO_HOST_NAME;
   packet.DHCP.options[i++] = DHO_ROUTERS;
   packet.DHCP.options[i++] = DHO_BROADCAST_ADDRESS;

   packet.DHCP.options[i++] = DHO_END;
   for(   ;i < sizeof(packet.DHCP.options);i++) {
      packet.DHCP.options[i] = 0x00;
   }


   // ====== Calculate checksum =============================================
   packet.UDP.check =
      InternetAddress::wrapChecksum(
         InternetAddress::calculateChecksum((unsigned char *)&packet.UDP, sizeof(packet.UDP),
            InternetAddress::calculateChecksum((unsigned char*)&packet.DHCP, sizeof(packet.DHCP),
               InternetAddress::calculateChecksum((unsigned char *)&packet.IP.ip_src,2 * sizeof(packet.IP.ip_src),
                  IPPROTO_UDP + (card32)ntohs(packet.UDP.len))))
      );


   // ====== Send packet ====================================================
   const PacketAddress packetAddress(interface->Name);
   const ssize_t result = OutputSocket.sendTo((void*)&packet, sizeof(packet),
                                              0, packetAddress);
   return(result == sizeof(packet));
}


// ###### Receive DHCP Reply ################################################
cardinal DHCPClient::receiveReply(DHCPConfiguration& config)
{
   char buffer [DHCP_MTU_MAX];
   const ssize_t result = InputSocket.receive((char*)&buffer,sizeof(buffer));
   if(result >= (ssize_t)BOOTP_MIN_LEN) {
      dhcp_packet* offer = (dhcp_packet*)&buffer;
      if((offer->op == BOOTREPLY)      &&
         (offer->htype == HTYPE_ETHER) &&
         (offer->hlen == 6)            &&
         (!strncmp((const char*)&offer->options,DHCP_OPTIONS_COOKIE,sizeof(DHCP_OPTIONS_COOKIE) - 1))) {
         if(packetToConfig(offer,(cardinal)result,config) == true) {
            return(config.MessageType);
         }
      }
   }
   return(0);
}


// ###### Configure interface ###############################################
bool DHCPClient::configureInterface(DHCPClientInterface* interface)
{
   bool ok;
#ifdef VERBOSE
   printTimeStamp(cout);
   cout << interface->Name << " ";
#endif

   NetworkInterface networkCard(interface->Name);
   switch(interface->State) {
      case DHCPClientInterface::DIS_Up:
      case DHCPClientInterface::DIS_Renewing: {
         const bool ok1 = networkCard.addAddress(NetworkInterface::AT_Address,interface->Config.Address,0);
         const bool ok2 = networkCard.addAddress(NetworkInterface::AT_Broadcast,interface->Config.Broadcast,0);
         const bool ok3 = networkCard.addAddress(NetworkInterface::AT_Netmask,interface->Config.SubnetMask,0);
         const bool ok4 = networkCard.up();
         ok  = (ok1 && ok2 && ok3 && ok4);
#ifdef VERBOSE
         if(ok) {
            cout << "is up!" << endl;
         }
         else {
            cout << "- configuration failed!" << endl;
         }
#endif
       }
      break;

     case DHCPClientInterface::DIS_Down:
     case DHCPClientInterface::DIS_UserShutdown:
        ok = networkCard.down();
#ifdef VERBOSE
        if(ok) {
           cout << "is down!" << endl;
        }
        else {
           cout << "- shutdown failed!" << endl;
        }
#endif
      break;

     default: {
        const bool ok1 = networkCard.addAddress(NetworkInterface::AT_Address,NullAddress,0);
        const bool ok2 = networkCard.addAddress(NetworkInterface::AT_Broadcast,NullAddress,0);
        const bool ok3 = networkCard.addAddress(NetworkInterface::AT_Netmask,NullAddress,0);
        const bool ok4 = networkCard.up();
        ok  = (ok1 && ok2 && ok3 && ok4);
#ifdef VERBOSE
        if(ok) {
           cout << "- pre-configuration failed!" << endl;
        }
        else {
           cout << "is pre-configured!" << endl;
        }
#endif
       }
      break;
   }
   return(ok);
}


// ###### Get configuration #################################################
bool DHCPClient::getConfiguration(const char*        name,
                                  DHCPConfiguration& config)
{
   bool ok = false;

   synchronized();
   DHCPClientInterface* interface = findInterface(name);
   if((interface->State == DHCPClientInterface::DIS_Up) ||
      (interface->State == DHCPClientInterface::DIS_Renewing)) {
      config = interface->Config;
      ok = true;
   }
   unsynchronized();

   return(ok);
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const DHCPConfiguration& config)
{
   os << "   Address:           " << config.Address.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy|SocketAddress::PF_HidePort) << endl;
   os << "   Host:              " << config.Host    << endl;
   os << "   Domain:            " << config.Domain  << endl;
   os << "   Subnet Mask:       " << config.SubnetMask.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy|SocketAddress::PF_HidePort) << endl;
   os << "   Broadcast Address: " << config.Broadcast.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy|SocketAddress::PF_HidePort) << endl;
   os << "   Name Server:       ";
   for(cardinal i = 0;i < config.NameServers;i++) {
      if(i > 0) {
         os << "                      ";
      }
      os << config.NameServer[i].getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy|SocketAddress::PF_HidePort) << endl;
   }
   os << "   Router:            " << config.Router.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_Legacy|SocketAddress::PF_HidePort) << endl;
   char str[128];
   snprintf((char*)&str,sizeof(str),"$%08x",config.TransactionID);
   os << "   DHCP Transact.ID:  " << str << endl;
   return(os);
}


}
