#include <tdsystem.h>
#include <wirelessnetworkinterface.h>
#include <dhcpclient.h>
#include <gnuplot.h>
#include <breakdetector.h>


using namespace Coral;


struct Interface
{
   char                      Name[IFNAMSIZ + 1];
   char                      Title[256];
   WirelessNetworkInterface* NetworkCard;
   DHCPConfiguration         Config;
   double                    DownThreshold;
   double                    UpThreshold;
   double                    Alpha;
   double                    LinkQuality;
   integer                   LinkSignal;
   integer                   LinkNoise;
   bool                      Up;
   bool                      WasWireless;
   bool                      IsConfigured;
};



int main(int argc, char** argv)
{
   set<Interface*> interfaceSet;
   GNUPlotScript   script;
   GNUPlotData     data;

   if(argc < 2) {
      cout << "Usage: " << argv[0] << " [Interface1{,DownTh{,UpTh{,Alpha}}}] {Interface2{,DownTh{,UpTh{,Alpha}}}} ... {-prefix=name prefix}" << endl;
      exit(1);
   }
   for(integer i = 1;i < argc;i++) {
      if(argv[i][0] == '-') {
         if(!(strncmp(argv[i],"-prefix=",8))) {
            if(script.create(&argv[i][8]) == false) {
               cerr << "ERROR: Unable to create GNUPlot script file!" << endl;
               exit(1);
            }
            if(data.create(&argv[i][8]) == false) {
               cerr << "ERROR: Unable to create GNUPlot data file!" << endl;
               exit(1);
            }
         }
         else {
            cerr << "ERROR: Bad argument " << argv[i] << "!" << endl;
            exit(1);
         }
      }
      else {
         Interface* interface = new Interface;
         if(interface == NULL) {
            cerr << "ERROR: Out of memory!" << endl;
            exit(1);
         }

         interface->UpThreshold   = 25.0;
         interface->DownThreshold = 15.0;
         interface->Alpha         = 0.5;
         interface->LinkQuality   = -1000.0;
         const int result = sscanf(argv[i],"%[^,],%lf,%lf,%lf",
                                   (char*)&interface->Name,
                                   &interface->DownThreshold,
                                   &interface->UpThreshold,
                                   &interface->Alpha);
         if(result >= 1) {
            interface->NetworkCard = new WirelessNetworkInterface(interface->Name);
            if(interface->NetworkCard == NULL) {
               cerr << "ERROR: Out of memory!" << endl;
               exit(1);
            }
            if(interface->NetworkCard->isWireless()) {
               snprintf((char*)&interface->Title, sizeof(interface->Title),
                        "%s Wireless", interface->Name);
               interface->WasWireless = true;
            }
            else {
               snprintf((char*)&interface->Title, sizeof(interface->Title),
                        "%s Wired", interface->Name);
            }
            if(interface->UpThreshold > 100.0) {
               interface->UpThreshold = 100.0;
            }
            if(interface->UpThreshold < 0.0) {
               interface->UpThreshold = 0.0;
            }
            if(interface->DownThreshold > 100.0) {
               interface->DownThreshold = 100.0;
            }
            if(interface->DownThreshold < 0.0) {
               interface->DownThreshold = 0.0;
            }
            if(interface->DownThreshold > interface->UpThreshold) {
               interface->DownThreshold = interface->UpThreshold;
            }
            if(interface->Alpha > 0.99) {
               interface->Alpha = 0.99;
            }
            if(interface->Alpha < 0.01) {
               interface->Alpha = 0.01;
            }
            interface->Up = true;
            interfaceSet.insert(interface);
         }
         else {
            cerr << "ERROR: Bad interface argument \"" << argv[i] << "\"!" << endl;
            exit(1);
         }
      }
   }

   DHCPClient dhcpClient;
   set<Interface*>::iterator iterator = interfaceSet.begin();
   while(iterator != interfaceSet.end()) {
      if(dhcpClient.addInterface((*iterator)->Name) == false) {
         cerr << "ERROR: Unable to add device " << (*iterator)->Name << ". Added twice?" << endl;
         exit(1);
      }
      iterator++;
   }
   if(interfaceSet.begin() == interfaceSet.end()) {
      cerr << "ERROR: No interface(s) given!" << endl;
      exit(1);
   }


   if(data.good()) {
      cardinal column = 2;
      set<Interface*>::iterator iterator = interfaceSet.begin();
      while(iterator != interfaceSet.end()) {
         script.plot("Link Quality Statistics",
                     (*iterator)->Title,
                     "Time [s]",
                     "Link Quality [%]",
                     NULL,
                     false,
                     data,
                     column);
         column += 4;
         iterator++;
      }
      script.done();

      column = 3;
      iterator = interfaceSet.begin();
      while(iterator != interfaceSet.end()) {
         script.plot("Signal Level Statistics",
                     (*iterator)->Title,
                     "Time [s]",
                     "Signal Level [dBm]",
                     NULL,
                     false,
                     data,
                     column);
         column += 4;
         iterator++;
      }
      script.done();

      column = 4;
      iterator = interfaceSet.begin();
      while(iterator != interfaceSet.end()) {
         script.plot("Noise Level Statistics",
                     (*iterator)->Title,
                     "Time [s]",
                     "Noise Level [dBm]",
                     NULL,
                     false,
                     data,
                     column);
         column += 4;
         iterator++;
      }
      script.done();

      column = 5;
      iterator = interfaceSet.begin();
      while(iterator != interfaceSet.end()) {
         script.plot("Status Statistics",
                     (*iterator)->Title,
                     "Time [s]",
                     "Status [0=down, 1=config, 2=up]",
                     NULL,
                     false,
                     data,
                     column,
                     "steps",
                     0,0, -0.5, 2.5);
         column += 4;
         iterator++;
      }
      script.done();
   }


   if(dhcpClient.start() == false) {
      cerr << "ERROR: Unable to start DHCP client!" << endl;
      exit(1);
   }


   cout << "Interface Control - Copyright (C) 2001 Thomas Dreibholz" << endl;
   cout << "-------------------------------------------------------" << endl;
   cout << endl;



   InternetAddress routerAddress;
   Interface*      routerInterface = NULL;
   const card64    start           = getMicroTime();

   installBreakDetector();
   while(!breakDetected()) {
      Interface* bestInterface  = NULL;
      Interface* worstInterface = NULL;
      double bestQuality  = -1000000.0;
      double worstQuality = 1000000.0;

      printTimeStamp(cout);
      cout << endl;

      data << (getMicroTime() - start) / 1000000.0 << " ";

      iterator = interfaceSet.begin();
      while(iterator != interfaceSet.end()) {
         Interface* interface                        = *iterator;
         WirelessNetworkInterface* wirelessInterface = interface->NetworkCard;


         interface->LinkSignal = -102;
         interface->LinkNoise  = -101;
         if(wirelessInterface->isWireless()) {
            WirelessLinkStatus status;
            if(wirelessInterface->getLinkStatus(status) == true) {
               if(interface->LinkQuality >= 0.0) {
                  interface->LinkQuality =
                     (interface->Alpha * interface->LinkQuality) +
                     ((1.0 - interface->Alpha) *
                        ((double)status.LinkQuality * 100.0) / (double)status.MaxLinkQuality);
                  interface->LinkSignal = status.Signal - 0x100;
                  interface->LinkNoise  = status.Noise  - 0x100;
               }
               else {
                  interface->LinkQuality =
                     ((double)status.LinkQuality * 100.0) / (double)status.MaxLinkQuality;
               }
            }
            else {
               interface->LinkQuality = 0.0;
            }
         }
         else {
            if((wirelessInterface->isInstalled()) && (!interface->WasWireless)) {
               // Assume that wired network card is running.
               interface->LinkQuality = 100.0;
            }
            else {
               interface->LinkQuality = 0.0;
            }
         }


         if((interface->LinkQuality < interface->DownThreshold) && (interface->Up)) {
            dhcpClient.stopInterface(interface->Name);
            interface->Up = false;
            printTimeStamp(cout);
            cout << interface->Name << ": DHCP configuration deactivedted!" << endl;
         }
         else if((interface->LinkQuality >= interface->UpThreshold) && (!interface->Up)) {
            dhcpClient.startInterface(interface->Name);
            interface->Up = true;
            printTimeStamp(cout);
            cout << interface->Name << ": DHCP configuration activedted!" << endl;
         }


         if(interface->Up) {
            interface->IsConfigured = dhcpClient.getConfiguration(interface->Name,interface->Config);
            if(interface->IsConfigured) {
               if(interface->LinkQuality > bestQuality) {
                  bestQuality   = interface->LinkQuality;
                  bestInterface = interface;
               }
               if(interface->LinkQuality < worstQuality) {
                  worstQuality   = interface->LinkQuality;
                  worstInterface = interface;
               }
            }
         }


         char str[256];
         snprintf((char*)&str,sizeof(str),"   %5s:  Quality=%3d.%02d %%   DownTh=%3d.%02d %%   UpTh=%3d.%02d %%  Alpha=%1.3f  => ",
                  interface->Name,
                  (cardinal)interface->LinkQuality, (cardinal)(interface->LinkQuality * 10000) % 100,
                  (cardinal)interface->DownThreshold, (cardinal)(interface->DownThreshold * 10000) % 100,
                  (cardinal)interface->UpThreshold, (cardinal)(interface->UpThreshold * 10000) % 100,
                  interface->Alpha);
         cout << str;
         if(interface->Up) {
            if(interface->IsConfigured) {
               cout << "up, " << interface->Config.Address.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy);
            }
            else {
               cout << "configuring";
            }
         }
         else {
            cout << "down";
         }
         cout << endl;


         cardinal status = 0;
         if(interface->Up) {
            status = 1;
            if(interface->IsConfigured) {
               status = 2;
            }
         }
         data << " "
              << interface->LinkQuality << " "
              << interface->LinkSignal  << " "
              << interface->LinkNoise   << " "
              << status;


         iterator++;
      }


      bool updateRouter = false;
      if(bestInterface != NULL) {
         if((routerAddress   != bestInterface->Config.Router) ||
            (routerInterface != bestInterface)) {
            routerAddress   = bestInterface->Config.Router;
            routerInterface = bestInterface;
            updateRouter    = true;
         }
      }
      else {
         routerAddress   = InternetAddress();
         routerInterface = NULL;
         updateRouter    = false;   // No update necessary!
      }

      if(updateRouter) {
         cout << "Setting default gateway to " << routerAddress.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy) << "." << endl;

         system("/sbin/route del default >/dev/null 2>/dev/null");

         char str[256];
         snprintf((char*)&str,sizeof(str),"/sbin/route add default gw %s >/dev/null 2>/dev/null",
                  routerAddress.getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy).getData());
         if(system(str) != 0) {
            printTimeStamp(cerr);
            cerr << "ERROR: Unable to execute /sbin/route!" << endl;
         }
      }


      data << endl;


      cout << endl;
      Thread::delay(1000000);
   }


   cout << "Terminated!" << endl;
   return(0);
}
