/*
 *  $Id: networkinterface.h,v 1.1 2001/11/28 11:46:49 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Network Interface
 *
 */


#ifndef NETWORKINTERFACE_H
#define NETWORKINTERFACE_H


#include "tdsystem.h"
#include "socketaddress.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#if (SYSTEM == OS_Linux)
#include <linux/if.h>
#else
#include <net/if.h>
#endif



namespace Coral {



/**
 * This class implements an interface to network devices (e.g. eth0, lo).
 *
 * @short   Network Interface
 * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
 * @version 1.0
 */
class NetworkInterface
{
  // ====== Constructor/Destructor ==========================================
   public:
   /**
     * Constructor.
     */
   NetworkInterface(const char* name);

   /**
     * Destructor.
     */
   ~NetworkInterface();


   // ====== Interface configuration ========================================
   /**
     * Check, if interface is installed.
     *
     * @return true, if interface is installed; false otherwise.
     */
   bool isInstalled();

   /**
     * Get interface's name.
     *
     * @return Interface name.
     */
   const char* getName() const;

   /**
     * Get interface's flags.
     *
     * @param flags Reference to write flags to.
     * @return true, if flags have been obtained; false otherwise.
     */
   bool getFlags(unsigned int& flags);

   /**
     * Set interface's flags.
     *
     * @param flags Flags.
     * @return true, if flags have been set; false otherwise.
     */
   bool setFlags(const unsigned int flags);

   /**
     * Get interface's MTU.
     *
     * @param mtu Reference to write MTU to.
     * @return true, if MTU have been obtained; false otherwise.
     */
   bool getMTU(unsigned int& mtu);

   /**
     * Set interface's MTU.
     *
     * @param MTU Flags.
     * @return true, if MTU have been set; false otherwise.
     */
   bool setMTU(const unsigned int mtu);

   /**
     * Get interface's hardware address.
     *
     * @param mtu Reference to write hardware address to.
     * @return true, if hardware address have been obtained; false otherwise.
     */
   bool getHardwareAddress(sockaddr& address);

   /**
     * Set interface's hardware address.
     *
     * @param hardware address Flags.
     * @return true, if hardware address have been set; false otherwise.
     */
   bool setHardwareAddress(const sockaddr& address);

   /**
     * Set interface's up.
     *
     * @return true for success; false otherwise.
     */
   bool up();

   /**
     * Set interface's down.
     *
     * @return true for success; false otherwise.
     */
   bool down();


   enum AddressType {
      AT_Address     = 0,
      AT_Netmask     = 1,
      AT_Broadcast   = 2,
      AT_Destination = 3
   };


   /**
     * Add address to interface. For IPv4, this will replace the IPv4 address.
     *
     * @param addrType Address type (AT_Address, AT_Netmask, AT_Broadcast, AT_Destination).
     * @param address Address.
     * @param prefixLength Prefix length (IPv6 only).
     * @return true for success; false otherwise.
     */
   bool addAddress(const AddressType    addrType,
                   const SocketAddress& address,
                   const cardinal       prefixLength = 0);

   /**
     * Delete address from interface.
     *
     * @param addrType Address type (AT_Address, AT_Netmask, AT_Broadcast, AT_Destination).
     * @param address Address.
     * @param prefixLength Prefix length (IPv6 only).
     * @return true for success; false otherwise.
     */
   bool deleteAddress(const AddressType    addrType,
                      const SocketAddress& address,
                      const cardinal       prefixLength = 0);

   /**
     * Get interface's address.
     *
     * @param addrType Address type (AT_Address, AT_Netmask, AT_Broadcast, AT_Destination).
     * @param afFamily Address family (e.g. AF_INET).
     * @param index Address number (IPv6 only).
     * @param prefixLength Pointer to store prefix length to (IPv6 only).
     * @return Pointer to address.
     */
   SocketAddress* getAddress(const AddressType addrType,
                             const int         afFamily,
                             const cardinal    index        = 0,
                             cardinal*         prefixLength = NULL);


   // ====== Protected data =================================================
   protected:
   /**
     * Interface name.
     */
   char Name[IFNAMSIZ + 1];

   /**
     * Socket for interface control.
     */
   static int ControlSocket;


   // ====== Private data ===================================================
   private:
   struct in6_ifreq {
      in6_addr     ifr6_addr;
      card32       ifr6_prefixlen;
      unsigned int ifr6_ifindex;
   };
};


}


#endif
