/*
 *  $Id: wirelessnetworkinterface.h,v 1.1 2001/11/28 11:46:49 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Wireless Network Interface
 *
 */


#ifndef WIRELESSNETWORKINTERFACE_H
#define WIRELESSNETWORKINTERFACE_H


#include "tdsystem.h"
#include "networkinterface.h"


#if (SYSTEM == OS_Linux)
#include <linux/wireless.h>
#else
#warning WirelessNetworkInterface is currently not implemented for non-Linux systems!
typedef int iw_param;
typedef int iw_range;
#endif



namespace Coral {



struct WirelessLinkStatus
{
   cardinal Status;
   cardinal MaxLinkQuality;
   cardinal LinkQuality;
   integer  MaxSignal;
   integer  Signal;
   integer  MaxNoise;
   integer  Noise;
   cardinal DiscardNWID;
   cardinal DiscardCrypt;
   cardinal DiscardMisc;
};

ostream& operator<<(ostream& os, const WirelessLinkStatus& status);



struct WirelessConfig
{
  bool HasNWID;
  bool HasFrequency;
  bool HasSensitivity;
  bool HasKey;
  bool HasESSID;
  bool HasNickname;
  bool HasAPAddress;
  bool HasBitrate;
  bool HasRTSThreshold;
  bool HasFragmentationThreshold;
  bool HasOperationMode;
  bool HasPowerManagement;
  bool HasTransmitPower;
  bool HasRetry;
  bool HasStatistics;
  bool HasRange;

  char      Name[IFNAMSIZ];
  iw_param	 NWID;
  double    Frequency;
  iw_param	 Sensitivity;
  char	     Key[IW_ENCODING_TOKEN_MAX];
  cardinal  KeySize;
  cardinal  KeyFlags;
  cardinal  ESSID_On;
  char      ESSID[IW_ESSID_MAX_SIZE + 1];
  sockaddr  APAddress;
  char      Nickname[IW_ESSID_MAX_SIZE + 1];
  iw_param  Bitrate;
  iw_param  RTSThreshold;
  iw_param  FragmentationThreshold;
  cardinal  OperationMode;
  iw_param  PowerManagement;
  iw_param  TransmitPower;
  iw_param  Retry;

  iw_range  Range;
};

ostream& operator<<(ostream& os, const WirelessConfig& config);



/**
 * This class implements an interface to wireless network devices.
 *
 * @short   Wireless Network Interface
 * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
 * @version 1.0
 */
class WirelessNetworkInterface : public NetworkInterface
{
   // ====== Constructor/Destructor ==========================================
   public:
   /**
     * Constructor.
     */
   WirelessNetworkInterface(const char* name);

   /**
     * Destructor.
     */
   ~WirelessNetworkInterface();


   // ====== Configuration and link status ==================================
   /**
     * Check, if interface is wireless.
     *
     * @return true, if interface is wireless; false otherwise.
     */
   bool isWireless();

   /**
     * Get wireless configuration.
     *
     * @param config Reference to store configuration to.
     * @return true for success; false otherwise.
     */
   bool getConfig(WirelessConfig& config);

   /**
     * Get wireless link status.
     *
     * @param config Reference to store link status to.
     * @return true for success; false otherwise.
     */
   bool getLinkStatus(WirelessLinkStatus& status);
};


}


#endif
